Package: ExactTree
Type: Package
Title: Exact Tree
Version: 0.1.1
Authors@R: c(person(given = "Juan", family = "Claramunt Gonzalez", role =
        c("aut", "cre","cph"), email = "j.claramunt.gonzalez@fsw.leidenuniv.nl"),
        person("Bart Jan","van Os",role=c("aut"),email="trees@4os.eu"),
        person(given = "Elise", family = "Dusseldorp", role = c("aut"), email
        = "elise.dusseldorp@fsw.leidenuniv.nl"))
Maintainer: Juan Claramunt Gonzalez <j.claramunt.gonzalez@fsw.leidenuniv.nl>
Description: Grows optimally global trees based on the algorithm defined in "An exact dynamic programming algorithm for regression and classification trees" (2011). It is possible to obtain both classification and regression trees depending on the measurement level of the outcome variable. The algorithm is based on the dynamic programming principle and guarantees that the resulting tree is optimal with respect to the chosen impurity measure. The package also includes a function to visualize the resulting trees, a function that summarizes the tree with its splitting information and leaf information, and a predict function that provides estimates for a new dataset given a model fit. 
Depends: R (>= 3.0.2), partykit, pracma, stats, grid, utils, graphics,
        formula.tools
Imports: gridtext, DescTools, methods, rpart
License: GPL-3
Encoding: UTF-8
Author: Juan Claramunt Gonzalez [aut, cre, cph],
  Bart Jan van Os [aut],
  Elise Dusseldorp [aut]
NeedsCompilation: yes
RoxygenNote: 7.3.2
Packaged: 2026-07-29 12:23:52 UTC; jclaramunt
Repository: CRAN
Date/Publication: 2026-08-07 10:30:14 UTC

# Generated by roxygen2: do not edit by hand

S3method(as.party,ETree)
S3method(plot,ETree)
S3method(predict,ETree)
S3method(prune,ETree)
S3method(summary,ETree)
export(ETree)
export(ETree.control)
export(SelectVar)
importFrom(DescTools,BinToDec)
importFrom(DescTools,DecToBin)
importFrom(formula.tools,lhs.vars)
importFrom(formula.tools,rhs.vars)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(grid,gpar)
importFrom(grid,grid.draw)
importFrom(grid,grid.layout)
importFrom(grid,grid.lines)
importFrom(grid,grid.newpage)
importFrom(grid,grid.points)
importFrom(grid,grid.polygon)
importFrom(grid,grid.rect)
importFrom(grid,grid.text)
importFrom(grid,grid.yaxis)
importFrom(grid,popViewport)
importFrom(grid,pushViewport)
importFrom(grid,unit)
importFrom(grid,upViewport)
importFrom(grid,viewport)
importFrom(gridtext,richtext_grob)
importFrom(methods,is)
importFrom(partykit,as.party)
importFrom(partykit,character_split)
importFrom(partykit,id_node)
importFrom(partykit,is.terminal)
importFrom(partykit,kids_node)
importFrom(partykit,node_party)
importFrom(partykit,nodeids)
importFrom(partykit,party)
importFrom(partykit,partynode)
importFrom(partykit,partysplit)
importFrom(partykit,plot.party)
importFrom(partykit,split_node)
importFrom(pracma,Reshape)
importFrom(pracma,fprintf)
importFrom(pracma,histc)
importFrom(pracma,isempty)
importFrom(pracma,randperm)
importFrom(pracma,repmat)
importFrom(pracma,strcmp)
importFrom(rpart,prune)
importFrom(stats,complete.cases)
importFrom(stats,model.frame)
useDynLib(ExactTree, .registration=TRUE)

CombineTrees2<-function(H1,TT1,H2,TT2,maxSize,FirstSplit){
#%and enforce split
H<-rep(2.3E10,maxSize)
Tl<-list() #=cell(3,1);
TT<-list() #cell(maxSize,1);
availSize<- dim(H1)[1]

for(ms in 2:maxSize){
# %msm=ms-min(maxSize,availSize);
# %H1
# %H1(1+msm:ms-(1+msm))
# %H2
# %H2(ms-(1+msm):-1:1+msm)
# %HN=H1(1+msm:ms-(1+msm))+H2(ms-msm:-1:msm);
  HN<-H1[1:(ms-1)]+H2[seq(from=ms-1,to=1,by=-1)]

  mH<-min(HN)
  mi<-which.min(HN)

  if(mH<H[ms]){
    H[ms:maxSize]<-mH
    Tl[[1]]<-FirstSplit
    Tl[[2]]<-TT1[mi]
    Tl[[3]]<-TT2[ms-mi]
    TT[[ms]]<-Tl
  }
}

H[1]<-0

return(H=H,TT=TT)

}

SetupRiskMeasure<-function(Y,measure,Prior,LossM){
#if undefined setup default Prior and LossM (Measure=3)

if(measure==3){
  if(is.null(Prior)||(length(Prior)==0)){
    for(y in 1:(NCOL(Y))){
      Prior[[y]]<-getDefPrior(matrix(Y[,y], nrow = NROW(Y), ncol = 1))
    }
  }

  if(is.null(LossM)||(length(LossM)==0)){
    for(y in 1:(NCOL(Y))){
      LossM[[y]]<-getDefLossM(max(Y[,y]))
    }
  }

  ClassSizes <- list()
  for(y in 1:(NCOL(Y))){
    ClassSizes[[y]] <- getFreqY(Y[,y])
  }

}else{
  ClassSizes <- list()
}


return(list(Prior=Prior, LossM=LossM, ClassSizes=ClassSizes))


}

GetPMisclassification<-function(Y){
nc<-NCOL(Y)
y<-matrix(0,nrow = nc, ncol = 1)
Y<-matrix(Y, nrow = NROW(Y), ncol = nc)
for(d in 1:nc){
  hf<-0
  O<-sort(Y[,d])#O should be a vector
  while(length(O)>0){
    c<-min(O)
    I<-which(O>c)
    f<-length(O)-length(I)
    if(f>hf){
      hf<-f
      y[d]<-c
    }
    O<-O[I]
  }
}

return(y)

}

getIndex<-function(prevNode=1, side="left", node,Tree,Index, keepnode=NULL){


if(Tree[node,1]!=0){

  #compute index
  if(node!=1){

    indexPrev<-ifelse(side=="left", 2*prevNode-1, 2*prevNode)
    Index[[2*(node)-1]]<-Index[[indexPrev]][Index[[2*(node)-1]]]
    Index[[2*(node)]]<-Index[[indexPrev]][Index[[2*(node)]]]
  }



  #Reiterations
  leftSide<-getIndex(node, side = "left", node + Tree[node,3],Tree,Index, keepnode)
  Index<-leftSide$Index

  rightSide<-getIndex(node, side = "right", node + Tree[node,4],Tree,Index, keepnode)
  Index<-rightSide$Index


  keepnode<-c(leftSide$keep,rightSide$keep)


}else{
  if(side=="left"){
    keepnode<-c(keepnode, 2*prevNode-1)
  }else{
    keepnode<-c(keepnode, 2*prevNode)
  }

  out<-list(Index=Index, keep=keepnode)
  return(out) #stop("Final node")
}





out<-list(Index=Index, keep=keepnode)
return(out)







}

ProcessTree<-function(Y,X,Tv,node,defCheck=NULL, measure, Prior, LossM, ClassSizes){
#% build full Tv table given incomplete table produced in Fortran
#% NOTE: nodes are defined in FORTRAN with relative positions


if(is.null(defCheck)){
  check<-0
}else{
  check<-defCheck
}

nc<-NCOL(Tv)
nr<-NROW(Tv)
if(nr==1 || nc==1){
Tv<-matrix(Tv, nrow = nr, ncol = nc)
}
Tv<-cbind(Tv,matrix(0,nrow = nr, ncol = (2 + 2*NCOL(X)))) #[,7+2*NCOL(X)]<-0
Tree <- Tv
Indices<-list()

ProcessT<-ProcessSubTree(Y,X,node,Tree, check, measure, Prior, LossM, ClassSizes,Indices) #Updates the tree
Tree<-ProcessT$Tree
Tv <- Tree

Indices<-ProcessT$Indices



return(list(Tv=Tv,Indices=Indices))

}

SplitSpace<-function(Y,X,p,i){
  #Splits the Measurement space and dependent Variables according to split i in predictor p
  I1<-which(X[,p]<=i)
  Y1<-Y[I1,]
  X1<-X[I1,]
  I2<-which(X[,p]>i)
  #%I=setdiff(1:size(X,1),I);
  Y2<-Y[I2,]
  X2<-X[I2,]


  return(list(Y1=Y1,X1=X1,Y2=Y2,X2=X2,I1=I1,I2=I2))

}

! Subroutines for timer functions, USING PORTLIB
MODULE Timer
! Function    : Date and time functions for DF6, most useful under Windows NT
! Version     : 1.0,  1999
! Author      : Bart Jan van Os
! Note        :
USE OptTreeCallBacks, ONLY: dp

PRIVATE

PUBLIC InitTimer,SetTimer,GetTimer,GetElapsedTime,GetSpecifiedElapsedTime,&
       WriteTimeFor,WriteElapsedTimeFor,WriteSpecifiedTimeFor,&
       WriteSpecifiedElapsedTimeFor,WriteDateAndTime,dp,sp

! Portable replacement for REAL(4)/REAL*4 notation.
INTEGER, PARAMETER :: sp = SELECTED_REAL_KIND(6, 37)

REAL Total,StartTimeA,StartTimeB, LastTimeA, LastTimeB

CONTAINS

SUBROUTINE InitTimer
! Function    : Initialize Timer, Total + Elapsed time. DOES FUNCTION IN
! Parameter block.
! Local variables.
! Program Start

   Total=0.0
   LastTimeA=0.0
   StartTimeA=0.0
   LastTimeB=0.0
   StartTimeB=0.0

END SUBROUTINE  InitTimer


SUBROUTINE SetTimer (CurrentTime)
!USE PORTLIB
! Function    : Initialize elapsed time. DOES FUNCTION IN WIN95
!               CurrentTime is allways the total elapsed time since prog
! Parameter block.
REAL(dp)          CurrentTime
! Local variables.
REAL(sp)          TA(2)
! Program Start

   Total = Total+DTIME(TA)
   CurrentTime=Total

   LastTimeA=LastTimeA+TA(1)
   StartTimeA=LastTimeA
   LastTimeB=LastTimeB+TA(2)
   StartTimeB=LastTimeB


END SUBROUTINE SetTimer



SUBROUTINE GetTimer (CurrentTime)
!USE PORTLIB
! Function    : Get elapsed time. DOES FUNCTION IN WIN95
!               CurrentTime is allways the total elapsed time since prog
! Parameter block.
REAL(dp)          CurrentTime
! Local variables
REAL(sp)          TA(2)
! Program Start

   Total = Total+DTIME(TA)
   CurrentTime=Total
   LastTimeA=LastTimeA+TA(1)
   LastTimeB=LastTimeB+TA(2)

END SUBROUTINE  GetTimer



SUBROUTINE GetElapsedTime (Elapsed)
! Function    : Get elapsed time.
!               Gives the elapsed time since last call to SetTime
! Parameter block.
REAL(dp)          Elapsed
! Local variables
! Program Start

   Elapsed = (LastTimeA-StartTimeA) + (LastTimeB-StartTimeB)

END SUBROUTINE  GetElapsedTime



SUBROUTINE GetSpecifiedElapsedTime (Elaps1,Elaps2)
! Function    : Get elapsed time.
!               Gives the elapsed time since last call to SetTime
!               specified in user and privelidged code time
! Parameter block.
REAL(dp)          Elaps1,Elaps2
! Local variables
! Program Start

   Elaps1 = LastTimeA-StartTimeA
   Elaps2 = LastTimeB-StartTimeB

END SUBROUTINE  GetSpecifiedElapsedTime



SUBROUTINE WriteTimeFor (Name,ElapsedTime,OutUnit)
! Function    : Print the elpased time for Name
! Parameter block.
INTEGER           OutUnit
REAL(dp)          ElapsedTime
CHARACTER(LEN=10)      Name
! Program Start

   !WRITE (OutUnit,'(1X,A10," needed ",F10.3," seconds of CPU time.")') &
    !  Name,ElapsedTime

Name=Name
ElapsedTime=ElapsedTime
OutUnit=OutUnit

END SUBROUTINE  WriteTimeFor



SUBROUTINE WriteElapsedTimeFor (Name,OutUnit)
! Function    : Print the elapsed time for Name
! Parameter block.
INTEGER           OutUnit
REAL(dp)           ElapsedTime
CHARACTER(LEN=10)      Name
! Program Start

Name=Name
OutUnit=OutUnit

   CALL GetElapsedTime (ElapsedTime)
   !CALL WriteTimeFor (Name,ElapsedTime,OutUnit)

END SUBROUTINE  WriteElapsedTimeFor




SUBROUTINE WriteSpecifiedTimeFor (Name,Elaps1,Elaps2,OutUnit)
! Function    : Print an elpased time for Name, specified over
!               user and system time.
! Parameter block.
INTEGER           OutUnit
REAL(dp)           Elaps1,Elaps2
CHARACTER(LEN=10)      Name
! Program Start

   !WRITE (OutUnit,'(1X,A10," needed ", F10.3, " seconds of CPU time.")') &
  !    Name,Elaps1+Elaps2
   !WRITE (OutUnit,'(1X,"(", F10.3, " seconds of user time",'//&
    !               '" and ",F10.3, " seconds of system time.)")') &
    !  Elaps1, Elaps2

Name=Name
Elaps1=Elaps1
Elaps2=Elaps2
OutUnit=OutUnit


END SUBROUTINE  WriteSpecifiedTimeFor



SUBROUTINE WriteSpecifiedElapsedTimeFor (Name,OutUnit)
! Function    : Print total elapsed time for Name, specified
!               over user and system time.
! Parameter block.
INTEGER           OutUnit
CHARACTER(LEN=10)      Name
! Local variables.
REAL(dp)           Elaps1,Elaps2
! Program Start

OutUnit=OutUnit
Name=Name

   CALL GetSpecifiedElapsedTime (Elaps1,Elaps2)
   !CALL WriteSpecifiedTimeFor (Name,Elaps1,Elaps2,OutUnit)

END SUBROUTINE  WriteSpecifiedElapsedTimeFor



SUBROUTINE WriteDateAndTime(OutUnit,Title)
! Function    :
! Parameter block.
INTEGER OutUnit
CHARACTER(*) Title
! Local variables.
CHARACTER (12) Dates,Times
! BEGIN ROUTINE

OutUnit=OutUnit
Title=Title

   CALL DATE_AND_TIME (Dates,Times)
   !WRITE (OutUnit,*) TRIM(Title),' ',Dates(7:8),'-',Dates(5:6),'-',Dates(1:4),&
   !    '    ',Times(1:2),':',Times(3:4),':',Times(5:6)

END SUBROUTINE  WriteDateAndTime



END MODULE Timer

transformResults<-function(TAll,NtermNodes,X,Y, Desc, CVdata, origData){


  if(is.list(TAll)){
    Tree<-TAll[[NtermNodes]]
  }else{
    Tree<-TAll
  }

  #MISSING WHEN NtermNodes=1






  # #Here I need to process the tree MISSING
  # Tree<-as.matrix(Tree)
  # if(Tree[1,1] > 0){
  #   #tables are available, recalculate everything in MATLAB and
  #   #show the results for each tree
  #   for(t in 1:NTrees){
  #     #do postprocessing stuff in MATLAB
  #     Tv <- TAll[[t]]
  #     Tv <- ProcessTree(Y,X,Tv,1,1) #last argument: boolean check that signals check of fortran output
  #     TAllList[[t]] <- Tv
  #     fprintf(file=fid,fmt='\n%s','POST-PROCESSED RESULTS FOR TREE OF ')
  #     if(maxSize==0){
  #       fprintf(file=fid,fmt='MAX DEPTH %2i\n',(maxLevel+1)-(NTrees-t))
  #     }else{
  #       fprintf(file=fid,fmt='MAX SIZE %2i MAX DEPTH %2i\n',maxSize-(NTrees-t),maxLevel+1)
  #     }
  #     #DisplayTreeResults(fid,Tv,Y,X,['DisplayNodes,DisplayMSpace']);
  #     #DisplayTreeResults(fid,Tv,Y,X,'DisplayNodes') #MISSING, DIMENSIONAL ERROR
  #   }
  #   Tree<-as.matrix(TAllList[[NTrees]])
  # }



  tNodes<-which(Tree[,1]==0)

  list_results<-list()

  list_results$li<-data.frame(Node=tNodes,N=Tree[tNodes,6],y=Tree[tNodes,7],h=Tree[tNodes,5], row.names = paste0("Leaf ",1:length(tNodes)))


  m<-length(Desc)/2

  if(NtermNodes > 1){
    sNodes<-which(Tree[,1]>0)
    split_vars<-c()
    split_points<-c()
    for (i in 1:length(sNodes)) {
      j<-sNodes[i]
      split_vars[i]<-Desc[[m+Tree[j,1]]]
      #Split point different if the variable is categorical or not. Change 16/03/26
      if(is.factor(origData[,split_vars[i]])){
        split_points[i]<-as.character(Desc[[Tree[j,1]]][(Tree[j,2])])#as.character added on 08/04/2026
      }else{
        split_points[i]<-Desc[[Tree[j,1]]][(Tree[j,2])+1] #Change 10/08/22
      }
    }


    list_results$si<-data.frame(Parent_node=sNodes,Child_nodes_Y=Tree[sNodes,3],Child_node_N=Tree[sNodes,4],Splitting_variable=split_vars,Split_point=split_points,
                                row.names = paste0("Split ",1:length(sNodes)))
  }else{
    list_results$si<-NULL
  }

  list_results$var.names<-unlist(c(paste0("Y",1:NCOL(Y)),Desc[(m+1):(2*m)]))

  list_results$data<-as.data.frame(cbind(Y,X))
  colnames(list_results$data)<-list_results$var.names

  list_results$NtermNodes<-NtermNodes

  list_results$CVOutput<-CVdata

  class(list_results)<-"ETree"

  return(list_results)
}

#' Control Parameters for ETree Algorithm
#'
#' Various parameters that control aspects of the \dQuote{ETree} algorithm.
#'
#' @param measure
#'        0 = minimize residual Sum of Squares (continuous Y)
#'        1 = minimize Misclassification Rate        (discrete Y)
#'        3 = minimize the Risk Sum P(A)*R(A)        (discrete Y)
#'            Measure = scalar or Measure=c(3,c(Prior,LossM))
#' @param maxsize
#'        if null: only maxdepth restrictions are applied
#'        if defined: the maximum number of terminal nodes for the best tree
#' @param maxdepth
#'        The maximum number of layers in a tree, NOT INCLUDING THE LAYER 0.
#'        maxdepth=0 means a tree with just one node (the root node),
#'        maxdepth=1 means one split and two terminal nodes, etc.
#' @param minbucket defines the minimum number of observations in a terminal node
#'        if the number of observations in a node is equal to this minimum
#'        the node becomes a possible terminal node; if the number of observations in a node
#'        is smaller, the node is set illegal and disregarded.
#' @param minheterogeneity defines the situation when a node A is considered Very Homogeneous
#'        and is not allowed to be split further, i.e. the node A becomes a possible terminal node
#'        (|A| = number of objects in node)
#' @param ncv specify N for N-fold cross-validation
#'           if NCV = 0, no N-fold cross-validation
#'           if NCV<0, abs(N)-fold cross-validation ONLY (no tree estimation)
#'           if NULL (default), NCV=0
#' @param cvVector <vector> process N-fold cross-validation according to the classes defined by <vector>
#'           vector is assumed to assign each observational unit to one of
#'           N-classes, number 1,...,N
#' @param heterogeneityonly (TRUE,FALSE)
#'        Algorithm will only give Heterogeneity/Impurity not a tree table
#'        => setting to false will give treetable as well
#' @param BoundH optional bound for best tree value, for BranchandBound
#' @param branchandbound (TRUE,FALSE)
#'        Algorithm will use branch and bound rules to speed up
#'        => for maxsize tree only, h only
#'        => make sure you specify boundH
#' @param alltreesizes (FALSE,TRUE)
#'        algorithm will compute multiple trees of size<=size restriction
#' @param sortedpsplits (TRUE,FALSE)
#'        Algorithm will try to speed up use a sorting logic.
#'        => suitable for continuous predictors with many categories
#' @param lookaheadheuristic Specify treedepth=the number of levels the heuristic will lookahead.
#'        When growing a tree with lookahead search (heurist approach),
#'        this specifies the maximum depth with respect to which
#'        any local split is optimized. When setting this depth = 1, this
#'        ammounts to convential tree growing.
#'
#' @return A list containing the options for the function ETree.
#'
#' @seealso \code{\link{ETree}}
#' @examples
#' ETree.control(measure=0, maxsize = 6, maxdepth = 4, minbucket = 5, ncv=10, alltreesizes = TRUE)
#'
#' @export
ETree.control<-function(measure=NULL, maxsize=4, maxdepth=2, minbucket=10, minheterogeneity=0.05,
                        ncv=10, cvVector=NULL, heterogeneityonly=NULL, BoundH=NULL, branchandbound=NULL, alltreesizes=FALSE, sortedpsplits=NULL, lookaheadheuristic=NULL) {





  if(is.null(ncv)+is.null(cvVector)<1){
    stop("Use only one argument between ncv and cvVector.")
  }



  AlgoType<-1
  CVVec<-c()



  if(!is.null(measure)){
    Measure<-measure
  }else{
    Measure<-0
  }


  if(!is.null(maxsize)){
    MaxSize<-maxsize
  }else{
    MaxSize<-0
  }


  if(!is.null(maxdepth)){
    MaxDepth<-maxdepth+1
  }else{
    MaxDepth<-0
  }



  if(!is.null(minbucket)){
    MinBucket<-minbucket
  }else{
    MinBucket<-10
  }




  if(!is.null(minheterogeneity)){
    MinH<-minheterogeneity
  }else{
    MinH<-0.05
  }



  if(!is.null(heterogeneityonly)){
    if(heterogeneityonly){
      AlgoType <- bitset(AlgoType,1,0)
    }else{
      AlgoType <- bitset(AlgoType,1,1)
    }
  }



  if(!is.null(BoundH)){
    BoundH<-BoundH
  }else{
    BoundH<-0
  }



  if(!is.null(branchandbound)){
    if(branchandbound){
      AlgoType <- bitset(AlgoType,2,1)
    }else{
      AlgoType <- bitset(AlgoType,2,0)
    }
  }



  if(!is.null(alltreesizes)){
    if(alltreesizes){
      AlgoType <- bitset(AlgoType,3,1)
    }else{
      AlgoType <- bitset(AlgoType,3,0)
    }
  }




  if(!is.null(sortedpsplits)){
    if(sortedpsplits){
      AlgoType <- bitset(AlgoType,4,1)
    }else{
      AlgoType <- bitset(AlgoType,4,0)
    }
  }



  if(!is.null(lookaheadheuristic)){
    AlgoType <- bitset(AlgoType,5,1)
    LookAheadDepth <- lookaheadheuristic
  }else{
    LookAheadDepth <- 0
  }





  if((MaxSize==0) && (MaxDepth==0)){
    stop('You have to specify at least the maximum size or the maximum depth.')
  }




  return(list(AlgoType=AlgoType, Measure=Measure, MaxSize=MaxSize, MaxDepth=MaxDepth, MinBucket=MinBucket, MinH=MinH,
       ncv=ncv, cvVector=cvVector, BoundH=BoundH, LookAheadDepth=LookAheadDepth))
}

#' Predictions for new data with a ETree object
#'
#' Predicts for (new) subjects the outcome variable based on a fitted
#'   \code{ETree} object.
#'
#' @param object an object of the class \dQuote{ETree}.
#' @param newdata a data frame with data on new subjects for whom predictions should be made.
#'   The data frame should contain at least the variables used in the splits of the fitted tree.
#'   It is not necessary to include the treatment variable.
#' @param type character string denoting the type of predicted object to be returned. The default is
#'   set to \code{type="pred"}: a vector with predicted treatment subgroup classes per subject
#'   is returned. If set to \code{"matrix"}, a matrix is returned with the leaf and
#'   corresponding node of the tree to which a subject is assigned.
#' @param depth If alltreesizes was set to TRUE in ETree, you need to specify the depth of the tree
#'   you want to use in the predict function. This parameter should be equal to the number of the tree
#'   you want to use in the TAll output of ETree. If NULL, the largest tree is used.
#' @param \dots optional additional arguments.
#'
#' @return One of the following objects is returned depending on output type specified
#'   in the function:
#'
#'   If \code{type="pred"}:
#'   vector of the predicted outcome for every individual in the data set. Returns NA
#'   for subjects with missing values on one or more of the splitting variables.
#'
#'   If \code{type="matrix"}:
#'   a matrix with predicted locations of subjects within the fitted tree. The leaf numbers are
#'   in the first column and the corresponding node numbers in the second column. Returns NA
#'   for subjects with missing values on one or more of the splitting variables.
#'
#'   If \code{type="prob"}:
#'   a matrix with probabilities.
#'
#' @seealso \code{\link{ETree}}, \code{\link{ETree.control}}
#'
#' @examples
#' \donttest{
#'   data(iris)
#'   trainingIris<-iris[1:100,]
#'   # Fit an Exact Tree model
#'   controlEtree <- ETree.control(measure=0, maxsize = 4, maxdepth = 3,
#'   minbucket = 5, ncv=5, alltreesizes = FALSE)
#'   tree<-ETree(Species ~ Sepal.Length + Sepal.Width + Petal.Length + Petal.Width,
#'   control= controlEtree, data = trainingIris)
#'   testIris<-iris[101:150,]
#'   predictions<-predict(tree, newdata=testIris, type="pred", depth=1)
#' }
#'
#'
#' @export
predict.ETree<-function(object, newdata = NULL, type = 'pred', depth=NULL, ...){


  if(is.null(depth)){
    depth<-length(object$TAll)
  }

  Desc<-object$Desc
  Data<-object$Data
  Yvarnames<-object$Yvarnames
  Xvarnames<-object$Xvarnames

  if(length(object$Sizes)!=0){
    Sizes<-object$Sizes[[depth]]
    Prob<-object$ProbLeaves[[depth]]
  }

  object<-object$Transf_Trees[[depth]]

  parent_nodes<-object$si[,"Parent_node"]
  li<-object$li


  #Transform categorical variables of the test dataset to numeric. The split points are numeric.
  #Change 01/07/2026. Do not transform to numeric the categorical variables.
  #Do the same transformation as in SelectVar.
  isFactor<-c()
  for (i in 1:NCOL(newdata)) {
    isFactor[i]<-ifelse(is.factor(newdata[,i]), TRUE, FALSE)
  }

  whichFactor<-which(isFactor==TRUE)

  for (i in whichFactor) {
    tryCatch({
      newdata[,i]<- as.numeric(newdata[,i])
    }, warning = function(w) {
      newdata[,i]<- as.numeric(as.factor(newdata[,i]))
    }, error = function(e) {
      levels(newdata[,i])<-1:length(levels(newdata[,i]))
      newdata[,i]<- as.numeric(as.factor(newdata[,i]))
    })
  }



  if(type=="matrix") {
    predY<-matrix(rep(0,2*NROW(newdata)),NROW(newdata),2)

    for(i in 1:NROW(newdata)){

      # if(is.na(newdata[i,])){
      #
      #   predY[i,1]<-NA
      #   predY[i,2]<-NA
      #
      # }else{

      newroot<-1
      while(newroot %in% parent_nodes){

        oldroot<-newroot
        split_num<-which(object$si[,"Parent_node"]==oldroot)
        splt_var<-object$si[split_num,"Splitting_variable"]

        if(is.na(newdata[i,splt_var])){
          newroot<-NA
        }else{
          splt_point<-object$si[split_num,"Split_point"]

          if(is.numeric(splt_point)){

            if(newdata[i,splt_var]<splt_point){ #Condition for numeric Change 10/08/2022 Old: if(newdata[i,splt_var]<splt_point){
              newroot<-object$si[split_num,"Child_nodes_Y"]
            }else{
              newroot<-object$si[split_num,"Child_node_N"]
            }

          }else{

            #New 26/03/2026
            Desc_index1 <- which(Xvarnames==splt_var)
            index_splt_point <- which(Desc[[Desc_index1]]==splt_point)
            selected_categories <- Desc[[Desc_index1]][1:index_splt_point]

            if(newdata[i,splt_var] %in% selected_categories){ #Check this one!!!!!! (10/08/2022). Condition for characters. Maybe this is not needed if categorical variables are transformed into numbers
              newroot<-object$si[split_num,"Child_nodes_Y"]
            }else{
              newroot<-object$si[split_num,"Child_node_N"]
            }


            # Old
            # if(newdata[i,splt_var]==splt_point){ #Check this one!!!!!! (10/08/2022). Condition for characters. Maybe this is not needed if categorical variables are transformed into numbers
            #   newroot<-object$si[split_num,"Child_nodes_Y"]
            # }else{
            #   newroot<-object$si[split_num,"Child_node_N"]
            # }

          }

        }
        # if(is.na(newdata[i,splt_var])){
        #   newroot<-NA
        # }else{
        #   splt_point<-object$si[split_num,"Split_point"]
        #   if(newdata[i,splt_var]<splt_point){ #Check this condition: Change 10/08/2022 Old: if(newdata[i,splt_var]<=splt_point){
        #     newroot<-object$si[split_num,"Child_nodes_Y"]
        #   }else{
        #     newroot<-object$si[split_num,"Child_node_N"]
        #   }
        #
        # }
      }

      if(is.na(newroot)){
        predY[i,1]<-NA
        predY[i,2]<-NA
      }else{
        leaf<-which(li$Node==newroot)
        predY[i,1]<-leaf
        predY[i,2]<-newroot
        #predh[i]<-li[leaf,"h"]
      }

      #}

    }
    colnames(predY)<-c("Leaf","Node")

  }else if(type=="pred"){
    predY<-rep(0,NROW(newdata))

    for(i in 1:NROW(newdata)){


      newroot<-1
      while(newroot %in% parent_nodes){

        oldroot<-newroot
        split_num<-which(object$si[,"Parent_node"]==oldroot)
        splt_var<-object$si[split_num,"Splitting_variable"]

        if(is.na(newdata[i,splt_var])){
          newroot<-NA
        }else{
          splt_point<-object$si[split_num,"Split_point"]

          if(is.numeric(splt_point)){

            if(newdata[i,splt_var]<splt_point){ #Condition for numeric Change 10/08/2022 Old: if(newdata[i,splt_var]<splt_point){
              newroot<-object$si[split_num,"Child_nodes_Y"]
            }else{
              newroot<-object$si[split_num,"Child_node_N"]
            }

          }else{

            #New 26/03/2026
            Desc_index1 <- which(Xvarnames==splt_var)
            index_splt_point <- which(Desc[[Desc_index1]]==splt_point)
            selected_categories <- Desc[[Desc_index1]][1:index_splt_point]

            if(newdata[i,splt_var] %in% selected_categories){ #Check this one!!!!!! (10/08/2022). Condition for characters. Maybe this is not needed if categorical variables are transformed into numbers
              newroot<-object$si[split_num,"Child_nodes_Y"]
            }else{
              newroot<-object$si[split_num,"Child_node_N"]
            }


            # Old
            # if(newdata[i,splt_var]==splt_point){ #Check this one!!!!!! (10/08/2022). Condition for characters. Maybe this is not needed if categorical variables are transformed into numbers
            #   newroot<-object$si[split_num,"Child_nodes_Y"]
            # }else{
            #   newroot<-object$si[split_num,"Child_node_N"]
            # }

          }

        }

      }


      if(is.na(newroot)){
        predY[i]<-NA

      }else{
        if(is.numeric(Data[,Yvarnames])){
          leaf<-which(li$Node==newroot)
          predY[i]<-li[leaf,"y"]
          #predh[i]<-li[leaf,"h"]
        }else if(is.factor(Data[,Yvarnames])){
          # #New 26/03/2026
          # Desc_index1 <- which(Xvarnames==splt_var)
          # index_splt_point <- which(Desc[[Desc_index1]]==splt_point)
          # selected_categories <- Desc[[Desc_index1]][1:index_splt_point]
          #
          # if(!is.factor(newdata[,splt_var])){
          #
          #   if(newdata[i,splt_var] < splt_point){ #Check this one!!!!!! (10/08/2022). Condition for characters. Maybe this is not needed if categorical variables are transformed into numbers
          #     newroot<-object$si[split_num,"Child_nodes_Y"]
          #   }else{
          #     newroot<-object$si[split_num,"Child_node_N"]
          #   }
          #
          # }else{
          #
          #   if(newdata[i,splt_var] %in% selected_categories){ #Check this one!!!!!! (10/08/2022). Condition for characters. Maybe this is not needed if categorical variables are transformed into numbers
          #     newroot<-object$si[split_num,"Child_nodes_Y"]
          #   }else{
          #     newroot<-object$si[split_num,"Child_node_N"]
          #   }
          #
          # }

          levelsY<-levels(Data[,Yvarnames])
          leaf<-which(li$Node==newroot)
          predY[i]<-levelsY[li[leaf,"y"]]
        }else{
          stop("Y must be numeric or factor")
        }
      }

    }
    names(predY) <- 1:NROW(newdata)[1]

  }else if(type=="prob"){

    categories <- levels(Data[,Yvarnames])
    ncat <- length(categories)

    predY<-matrix(rep(0,ncat*NROW(newdata)),NROW(newdata),ncat)

    for(i in 1:NROW(newdata)){


      newroot<-1
      while(newroot %in% parent_nodes){

        oldroot<-newroot
        split_num<-which(object$si[,"Parent_node"]==oldroot)
        splt_var<-object$si[split_num,"Splitting_variable"]

        if(is.na(newdata[i,splt_var])){
          newroot<-NA
        }else{
          splt_point<-object$si[split_num,"Split_point"]

          if(is.numeric(splt_point)){

            if(newdata[i,splt_var]<splt_point){ #Condition for numeric Change 10/08/2022 Old: if(newdata[i,splt_var]<splt_point){
              newroot<-object$si[split_num,"Child_nodes_Y"]
            }else{
              newroot<-object$si[split_num,"Child_node_N"]
            }

          }else{

            #New 26/03/2026
            Desc_index1 <- which(Xvarnames==splt_var)
            index_splt_point <- which(Desc[[Desc_index1]]==splt_point)
            selected_categories <- Desc[[Desc_index1]][1:index_splt_point]

            if(newdata[i,splt_var] %in% selected_categories){ #Check this one!!!!!! (10/08/2022). Condition for characters. Maybe this is not needed if categorical variables are transformed into numbers
              newroot<-object$si[split_num,"Child_nodes_Y"]
            }else{
              newroot<-object$si[split_num,"Child_node_N"]
            }

            # if(newdata[i,splt_var]==splt_point){ #Check this one!!!!!! (10/08/2022). Condition for characters. Maybe this is not needed if categorical variables are transformed into numbers
            #   newroot<-object$si[split_num,"Child_nodes_Y"]
            # }else{
            #   newroot<-object$si[split_num,"Child_node_N"]
            # }

          }

        }

      }


      if(is.na(newroot)){
        predY[i]<-NA

      }else{
        if(is.factor(Data[,Yvarnames])){
          leaf<-which(li$Node==newroot)
          if(is.list(Prob)){
            predY[i,]<-Prob[[leaf]]
          }else{
            predY[i,]<-Prob
          }

        }else{
          stop("Y must be a factor")
        }
      }

    }
    colnames(predY) <- categories



  }





  return(predY)



}



# for(i in 1:NROW(newdata)){
#
#   newroot<-1
#   while(newroot %in% parent_nodes){
#
#     oldroot<-newroot
#     splt_var<-object$si[oldroot,"Splitting_variable"]
#     splt_point<-object$si[oldroot,"Split_point"]
#     if(newdata[i,splt_var]>=splt_point){ #Check this condition
#       newroot<-object$si[oldroot,"Child_nodes_Y"]
#     }else{
#       newroot<-object$si[oldroot,"Child_nodes_N"]
#     }
#
#   }
#
#
#
#   leaf<-which(li$Node==newroot)
#   predY[i]<-li[leaf,"y"]
#   #predh[i]<-li[leaf,"h"]
#
#
# }


# #Create variable to be returned
# if(type=="matrix") {
#   nodemat <- matrix(0, nrow=dim(newdata)[1], ncol=2)
#   for(i in 1:length(nnum)){
#     nodemat[index[which(Gmat[,i]==1)],1] <- which(object$li[,1]==nnum[i])
#     nodemat[index[which(Gmat[,i]==1)],2] <- nnum[i]
#   }
#   if(nmis!= 0){
#     nodemat[naindex,c(1:2)] <- NA
#   }
#   colnames(nodemat) <- c("Leaf", "Node")
#   #rownames(nodemat) <- as.numeric(rownames(ytxna)) # give subjects numbers of original dataset
#   return(nodemat)
# }
#
# if(type=="pred"){
#   classmat <- numeric(dim(newdata)[1])
#   for(i in 1:length(nnum)) {
#     classmat[index[which(Gmat[,i]==1)]] <- object$li[which(nnum[i]==object$li[,1]),10]
#   }
#   if(nmis!= 0){
#     classmat[naindex] <- NA
#   }
#   names(classmat) <- 1:dim(newdata)[1]
#   return(classmat)
# }

missingDataDealer<-function(data, XSelected, YSelected){

  Y<-data[,YSelected]
  X<-data[,XSelected]

  #Missing data in the outcome
  if(any(is.na(Y))){
    warning("There are missing values on your outcome variable. These cases are removed")
    index<-is.na(Y)
    data<-data[!index,]
  }

  #if(method == "Delete"){
  if(any(is.na(X))){
    warning("There are missing values on the predictors. The cases with missing values are removed. If you want those cases to be included in the model, deal with the missing values (e.g., imputation) before running Exact Trees.")
    indexNA<-!complete.cases(X)
    data<-data[!indexNA,]
  }
  #}

  # if(method == "Category"){
  #
  #
  #   nvars <- NCOL(X)
  #   for (i in 1:nvars) {
  #     if(any(is.na(X[i,]))){
  #       warning("There are missing values on your predictors variable.")
  #
  #       #How do we want to deal with them?
  #
  #
  #       #If X is numeric
  #
  #
  #
  #
  #       #If X is ordinal
  #
  #
  #
  #
  #       #If X is nominal
  #
  #
  #     }#End if
  #   }#End for
  #
  # }#End Category

  return(data)


}

OptTreeGate<-function(defY,X,XType,defMeasure,defMaxSize,defMaxDepth,defMinNodeSize,defMinH,AlgoType=NULL,CVVec=NULL,BoundH=NULL,LookAheadDepth, verbose = TRUE){ #,Description
  # % OLD INTERFACE, RETAINED FOR BCAKWARDS COMPATIBILITY
  #  OptTreeGate(Y,X,Measure,MaxSize,MaxDepth,MinNodeSize,MinH,AlgoType,CVVec)
  # defY<-Y
  # defMeasure<-Measure
  # defMaxSize<-MaxSize
  # defMaxDepth<-MaxDepth
  # defMinNodeSize<-MinNodeSize
  # defMinH<-MinH
  # % USE OptimalTree INSTEAD
  # %
  # %
  # % Find the optimal classification or regression tree
  # % minimizing an additive global objective function
  # % This version gateways to a Fortran dll
  # %
  # % X is expected to have ordered, consecutive categories
  # %   Measure:
  # %     0 = minimize residual Sum of Squared Error (continuous Y)
  # %     1 = minimize Misclassification Rate        (discrete Y)
  # %     3 = minimize the Risk Sum P(A)*R(A)        (discrete Y)
  # %         Measure = scalar or Measure={3,[Prior,[LossM]]}
  # %     if zero: only MaxDepth rstrictions are applied
  # %     if defined: the maximum number of terminal nodes for the best tree
  # %   MaxDepth:
  # %		The maximum number of layers in a tree, MaxDepth=1 means a tree with just
  # %     one node, MaxDepth=2 means one split and two terminal nodes, etc.
  # %     MaxDepth-1 is the maximum number of consecutive splits in the tree.
  # %     This is related to the maximum order of interactions in the tree.
  # %   MinNodeSize:
  # %     defines the minimum number of observations in a terminal node
  # %     if the number of observations in a node is equal to this minimum
  # %     the node becomes a possible terminal node; if the number of observations in a node
  # %     is smaller, the node is set illigal and disregarded.
  # % 	 MinH:
  # %		defines the situation when a node A is considered Very Homogeneous
  # %     and is not allowed to be split further, i.e. the node A becomes a possible terminal node
  # %     (|A| = number of objects in node)
  # %		Measure = 0
  # %			(h/|A|)<minH*(TotalH/n);
  # %			=> minH defines a proportion of the Mean SSE of the unpredicted Y (TotalH/n).
  # %			If the actual MSSE in a node is smaller than this proportion
  # %        of the original MSSE, the node is considered Very Homogeneous
  # %		Measure = 1
  # %			(h/|A|)<minH;
  # %			=> minH is the minimum proportion of misclassification.
  # %        If the actual proportion is smaller, the node is considered Very Homogeneous
  # %		Measure = 3
  # %			h<minH*TotalH;
  # %			=> minH is the minimum proportion of Risk of the unpredicted Y Risk.
  # %        If the actual proportion is smaller, the node is considered Very Homogeneous
  # %   AlgoType: OPTIONAL
  # %        0  = h only;
  # %        1  = h+tree table
  # %        2  = branch and bound, MaxSize only, h only, no other options
  # %        +4 = results for all trees <= size restrictions
  # %        +8 = sorted predictor splitting (faster for continous predictors?)
  # %        +16= heuristic search
  # %        default = 1
  # %        NOTE: currently, not all option combinations are implemented:
  # %           4 can not be combined with 8
  # %           8 can only be used with a MaxDepthTree
  # %     CVVec: OPTIONAL
  # %        specify a vector that assigns all observations to N classes for N-fold cross-validation
  # %        if NCV = 0, no N-fold cross-validation
  # %        if NCV<0, abs(N)-fold cross-validation ONLY (no tree estimation)
  # %        default=0
  # %     BoundH: OPTIONAL
  # %        optional bound for best tree value, when using AlgoType == 2
  # % example:
  # %     [h,T]=OptTreeGate(Y,X,0,0,3,8,0.05);
  # %     gives a SSE tree for continuous dependent variables Y
  # %     with no size constraint, MaxDepth=3, MinNodeSize=8, MinH=0.05
  # %
  # % preprocessing:
  # %	 examples of reading data: ReadDataFile?.m
  # %   then use
  # %     [Y,X,Desc]=SelectVar(Data,Names,[11],[4 6 8 9 10]);
  # %   to create the neccesary discretized Predictors and dependent variable(s)
  # %   where Data and Names are the arrays created with `ReadDataFile?.m'
  # %   and the first (index) vector selects the dependent variables (=columns of Data),
  # %   the second (index) vector selects the predictor variables (=columns of Data).
  # %
  # % postprocessing:
  # %   use
  # %     genTexTree(T,Desc)
  # %   to produce a TeX tree (if neccesary, do twice)



  #define globals
  #global LogFileName Description LookAheadDepth m n measure maxLevel maxSize minNodeSize minH somethingLarge nS nN depth Prior LossM ClassSizes SWHandle IsInterrupted CurrentCV;


  # LookAheadDepth<-globalenv()$LookAheadDepth
  # LogFileName<-globalenv()$LogFileName
  # Description<-globalenv()$Description

  if(!exists('XType') || is.null(XType)){
    XType<-rep(0, NCOL(X))

  }



  if(!exists('LookAheadDepth') || is.null(LookAheadDepth)){
    LookAheadDepth <- 0
  }


  # SWHandle<<-c()
  # IsInterrupted <<- FALSE
  CurrentCV<-0
  somethingLarge<-2E10
  Depth_var<-0
  nS<-0
  nN<-0
  m<-dim(X)[2]
  n<-dim(X)[1]

  if((defMaxDepth==0) && (defMaxSize>0)){
    defMaxDepth <- defMaxSize
  }

  if(defMaxSize>(2^(defMaxDepth-1))){
    defMaxSize<-2^(defMaxDepth-1) # Prevent redundant heap
    warning('The MaxSize has been reduced because it exceeds the actual Depth constraint')

    # print('Warning: The MaxSize has been reduced because it exceeds the actual Depth constraint')
  }

  maxSize<-defMaxSize
  maxLevel<-defMaxDepth-1
  minNodeSize<-defMinNodeSize
  minH<-defMinH

  #% Start LogFile

  #%check whether defMeasure includes a Prior and Loss (Measure=3)
  SMeasOut<-SetupMeasure(defMeasure)
  measure<-SMeasOut$measure
  measureCB<-SMeasOut$measureCB
  Prior<-SMeasOut$Prior
  LossM<-SMeasOut$LossM

  #init Y: autorecode for discrete data (measure>1)
  Y <- SetupY(defY,measure)

  #if undefined setup default Prior and LossM (Measure=3)
  SRMeasOut<-SetupRiskMeasure(Y,measure,Prior,LossM)
  Prior<-SRMeasOut$Prior
  LossM<-SRMeasOut$LossM
  ClassSizes<-SRMeasOut$ClassSizes

  UP<-UnravelCells(Prior) #UP = Unraveled Prior
  UL<-UnravelCells(LossM) #UP = Unraveled Loss matrix

  #%setup globals for `Heterogeneity' and `VeryHomogeneous' callbacks
  #SetupCallBacks(Y,X,measureCB,measure, Prior, LossM, ClassSizes)

  # get optional arguments
  if(is.null(BoundH)){
    BoundH<-GetHeterogeneity(Y=Y, measure=measure, Prior=Prior, LossM=LossM, ClassSizes=ClassSizes)
  }

  if(is.null(AlgoType)){
        AlgoType <- 1
  }


  #Define NCV, the number of cross-validations
  if(is.null(CVVec)){
    NCV<-0
  }else{
    NCV<-max(CVVec)
  }


  tic<-Sys.time()

  #SWHandle <<- OptTreeStatus_('Init',0,varargin=list(max(X)-1,LogFileName,Description,NCV))#MISSING!!! DIFFICULT FUNCTION

  #%call Fortran function
  OptTreeFout<-OptTreeF(Y,X,XType,measureCB,defMaxSize,defMaxDepth,defMinNodeSize,defMinH,
                        AlgoType,BoundH,Prior=UP, LossM=UL,LookAheadDepth)

  h<-OptTreeFout$h
  Tree<-OptTreeFout$Tv
  hAll<-matrix(OptTreeFout$hAll,nrow = 1)
  TAll<-OptTreeFout$TAll


  #output table with h for all trees
  NTrees <- NCOL(hAll)#dim(hAll)[2]
  if(verbose){
    fprintf(fmt='\n%s\n','APPARENT ERROR')
    fprintf(fmt='%s\n','Size      Error     Rel.Error')
  }
  maxH<-GetHeterogeneity(Y=Y, measure=measure, Prior=Prior, LossM=LossM, ClassSizes=ClassSizes)
  for(t in 1:NTrees){
    Size <- MaxEndNodes(t,NTrees,defMaxSize,defMaxDepth)
    if(verbose){
      fprintf(fmt='%i %13.4f %13.4f\n',Size,hAll[t],hAll[t]/maxH)
    }
  }

  #return
  #T(1,1)

  Tree<-as.matrix(Tree)
  TAllList<-list()
  IndicesList<-list()

  if(Tree[1,1] > 0){
    #tables are available, recalculate everything in MATLAB and
    #show the results for each tree

    for(t in 1:NTrees){
      #do postprocessing stuff in MATLAB
      Tv <- TAll[,,t]
      ProcessT <- ProcessTree(Y,X,Tv,1,1,measure=measure, Prior=Prior, LossM=LossM, ClassSizes=ClassSizes) #5th argument: boolean check that signals check of fortran output
      Tv <- ProcessT$Tv
      IndicesList[[t]]<-ProcessT$Indices
      TAllList[[t]] <- Tv
      # if(verbose){
      #   fprintf(fmt='\n%s','POST-PROCESSED RESULTS FOR TREE OF ')
      #   if(maxSize==0){
      #     fprintf(fmt='MAX DEPTH %2i\n',(maxLevel)-(NTrees-t)) #Old: fprintf(fmt='MAX DEPTH %2i\n',(maxLevel+1)-(NTrees-t))
      #   }else{
      #     fprintf(fmt='MAX SIZE %2i MAX DEPTH %2i\n',maxSize-(NTrees-t),maxLevel) #Old: fprintf(fmt='MAX SIZE %2i MAX DEPTH %2i\n',maxSize-(NTrees-t),maxLevel+1)
      #   }
      # }
      #DisplayTreeResults(fid,Tv,Y,X,['DisplayNodes,DisplayMSpace']);
      #DTR<-DisplayTreeResults(defT=Tv,Y=Y,X=X,DisplayOptions='DisplayNodes', m=m, measure=measure, Prior= Prior, LossM= LossM, ClassSizes=ClassSizes)
    }
    Tree<-as.matrix(TAllList[[NTrees]])
  }



  IsInterrupted<-FALSE

  CVoutput<-c()

  if((!is.null(CVVec))){ #&&(!IsInterrupted)){
    if(measure!=0){
      CVoutput<-NULL
      warning('No CV results are returned. CROSSVALIDATED ERROR IS NOT YET CALCULATED FOR CLASSIFICATION TREES (MEASURE=1 or 3)')
    }else{
      CVoutput<-CVOptTreeF(CVVec,Y,X,XType,measureCB,defMaxSize,defMaxDepth,defMinNodeSize,defMinH,AlgoType,BoundH,UP,UL, IsInterrupted, LookAheadDepth, measure, Prior, LossM, ClassSizes)
    }
  }




  time<-Sys.time()-tic

  if(verbose){
    cat(paste0("Time: ",time))
  }

  return(list(h=h,Tree=Tree,hAll=hAll,TAll=TAll,TAll_list=TAllList, CVOutput=CVoutput, Indices=IndicesList))

}

#' This function preprocess the data for the Optimal Trees function.
#'
#'@param formula a description of the model to be fit. The format is \code{Y1 + \dots + Yn ~  X1 + \dots + Xn},
#'   where the variable before ~ represents the dependent variables and the variables after
#'   the ~ are the independent variables.
#' @param data Dataset to be analyzed. If data contains ordinal variables, order the factor levels
#' before using this function. Otherwise, the factors will be ordered alphabetically and the results will not be correct.
#' @param Names names of the variables in the data. If empty, it uses the column names of Data
#' @param YSelected index corresponding to the first Selected column.
#' @param XSelected Indices corresponding to the selected columns that will form X
#' @param map [ori vars] sets for each variable whether the tree will depict
#'                       thresholds based upon original values (var number), or recoded values (0).
#' @param original [TRUE/FALSE] sets for each variable whether the tree will depict
#'                       thresholds based upon original values (true), or recoded values (false).
#' @param round [factors] autorecoding based upon rounded categories; the categories will
#'                       be round after multiplying with the factor, and divided by the
#'                       factor after rounding. Set round=0 for variables that are not
#'                       rounded (e.g. discretized variables).
#' @param discretize [ncat] optimal discretization while minimizing SS within
#'                       a group, set ncat=0 for each variable that is to
#'                       be left unaltered
#' @param verbose logical, if TRUE, prints information about the progress of the algorithm.
#'
#' @details The function selects the Y and X variables according to the inputs and also returns
#' Desc, a list containing variable cutting points and variable names. These 3 elements are
#' required by the OptimalTrees function.
#'
#' @return Returns the following 3 elements:
#'  \item{Y}{dataset containing the Y variables.}
#'  \item{X}{dataset containing the X variables.}
#'  \item{Desc}{List containing the variable names and their corresponding cutting points.}
#'
#'
#' @seealso \code{\link{summary.ETree}}, \code{\link{ETree}},
#'   \code{\link{plot.ETree}}, \code{\link{predict.ETree}}
#'
#' @examples
#' data(mtcars)
#' dataSelection<-SelectVar( mpg ~ cyl + hp + wt, data = mtcars, discretize= c(0, 10, 10))
#'
#' @keywords tree
#' @keywords exact
#'
#' @importFrom pracma histc repmat
#' @importFrom formula.tools lhs.vars rhs.vars
#' @importFrom stats model.frame complete.cases
#'
#' @export
SelectVar<-function(formula=NULL, data, Names=NULL, YSelected=NULL, XSelected=NULL, map=NULL, original=NULL, round=NULL, discretize=NULL, verbose = TRUE){



  if(is.null(formula)&(is.null(YSelected) | is.null(XSelected))){
    stop("Fill in the formula or YSelected and XSelected")
  }


  if(!is.null(formula)){

    Lside<- lhs.vars(formula)
    Rside<- rhs.vars(formula, data=data)
    YSelected <- which(colnames(data) %in% Lside)
    XSelected <- which(colnames(data) %in% Rside)
  }

  isFactor<-c() #Change 11/08/2022
  for (i in 1:length(XSelected)) {
    isFactor[i]<-ifelse(is.factor(data[,XSelected[i]]), TRUE, FALSE)
  }

  XType<-as.integer(isFactor)


  if(is.null(map)+is.null(original)<1){
    stop("Use only one argument between map and original.")
  }



  if(length(round)==1){
    round<-rep(round,length(XSelected))
  }

  if(length(discretize)==1){
    discretize<-rep(discretize,length(XSelected))
  }



  Mapped <- FALSE
  Rounded <- FALSE
  RFactor<-rep(0, length(XSelected))
  Discretized <- FALSE
  XMap<-rep(0, length(XSelected))
  NCat<-rep(0, length(XSelected))
  if(is.null(Names)){
    Names <- colnames(data)
  }




  if (!is.null(map)){
    XMap <- map
    Mapped <- TRUE
  }

  if(!is.null(original)){
    Mapped <-TRUE
    XMap <- XSelected
    XMap[which(original==FALSE)]<-0
  }

  if(!is.null(round)){
    Rounded <- TRUE
    RFactor <- round
  }


  if(!is.null(discretize)){
    Discretized <- TRUE
    NCat <- discretize
  }


  DataOrig<-data

  #Deal with missing values
  data<-missingDataDealer(data, XSelected, YSelected)



  # start processing of data
  n<-dim(data)[1]
  m<-length(YSelected) #dim(YSelected)[2]
  Y<-matrix(0,nrow = n, ncol = m)

  for(y in 1:m){
    Y[,y]<-data[,YSelected[y]]
  }


  m<-length(XSelected) #dim(XSelected)[2]
  X<-matrix(0,nrow = n, ncol = m)
  Desc<-list() #matrix(nrow = m,ncol = 2)#cell(m,2); #it is more a list with 2 dimensions
  length(Desc)<-2*m




  for(x in 1:m){
    V<-data[,XSelected[x]] #predictor is selected

    if(isFactor[x]){

      #New change 17/03/2026: Order the factor based on the mean of Y.  #New change if factor 11/08/2022
      # Note: The algorithm is not ready for this.
      # mean_cat<-rep(0, length(levels(data[,XSelected[x]])))
      # for (cat in 1:length(levels(data[,XSelected[x]]))) {
      #   mean_cat[cat] <- mean(Y[data[,XSelected[x]]==(levels(data[,XSelected[x]])[cat])])
      # }
      # ordCat<-order(mean_cat)



      #X and Desc

      #Check 28/05/2026
      Desc[[x]]<- suppressWarnings(matrix(as.numeric(levels(data[,XSelected[x]])), ncol = 1))

      if(all(is.na(Desc[[x]]))){
        Desc[[x]]<- matrix(as.numeric(1:length(levels(data[,XSelected[x]]))), ncol = 1)
      }


      #Desc[[x]]<- matrix(levels(data[,XSelected[x]])[ordCat], ncol = 1)

      # New: Desc[[x]]<- matrix(num_levels, ncol = 1)
      Desc[[m+x]]<-Names[XSelected[x]]

      # old: X[,x]<- as.numeric(data[,XSelected[x]])

      #Not ready yet:
      #levels(data[,XSelected[x]])<-ordCat
      #X[,x]<- as.numeric(as.factor(data[,XSelected[x]]))

      #Not all can be transformed to numeric
      #If it cannot be transformed to numeric, we will use the order of the categories
      #Use a try-catch to fix it
      tryCatch({
        X[,x]<- as.numeric(data[,XSelected[x]])
      }, warning = function(w) {
        X[,x]<- as.numeric(as.factor(data[,XSelected[x]]))
      }, error = function(e) {
        levels(data[,XSelected[x]])<-1:length(levels(data[,XSelected[x]]))
        X[,x]<- as.numeric(as.factor(data[,XSelected[x]]))
      })




    }else{ #if not factor

      if(Rounded & RFactor[x]>0){
        V<-round(V*RFactor[x])/RFactor[x]
      }
      if(Discretized & NCat[x]>1){
        if(verbose){
          print(paste0('optimal discretization, N= ',n,' NCat= ',NCat[x]))
        }
        if(NCat[x]==2){
          DPOPout<-DPOP(V,3, verbose=verbose)
        }else{
          DPOPout<-DPOP(V,NCat[x], verbose=verbose)
        }

        H<-DPOPout$H
        P<-DPOPout$P
        I<-DPOPout$I
        measure<-DPOPout$measure
        Heap<-DPOPout$Heap
        if(NCat[x]==2){
          X[,x]<-GetOP(I,P[2,1])
        }else{
          X[,x]<-GetOP(I,P[NCat[x],])
        }
        M<-GetCodeMap(X[,x],data[,XSelected[x]])
      }else{
        AutoRecOut<-AutoRecode(V)
        X[,x]<-AutoRecOut$O
        M<-AutoRecOut$M
      }
      if(Mapped & (XMap[x]>0)){
        Desc[[x]]<-GetCodeMap(X[,x],data[,XMap[x]])
        Desc[[m+x]]<-Names[XMap[x]]
      }else{
        Desc[[x]]<-M
        Desc[[m+x]]<-Names[XSelected[x]]
      }

    } #end if not factor


  } #end for each predictor



  #show selection

  if(verbose){

    for(y in 1:NCOL(YSelected)){
      if(m==1){
        print(paste0('Y = ',Names[YSelected[y]]))
      }else{
        print(paste0('Y', y,' = ',Names[YSelected[y]]))
      }
    }

    for(x in 1:m){
      # if(isFactor[x]){
      #   Line <- paste0('X',x,' = ',Names[XSelected[x]],' NCat = ',length(levels(data[,XSelected[x]])))
      # }else{
      #   Line <- paste0('X',x,' = ',Names[XSelected[x]],' NCat = ',max(X[,x]))
      # }
      Line <- paste0('X',x,' = ',Names[XSelected[x]],' NCat = ',max(X[,x]))

      if (Mapped & (XMap[x]>0)){
        Line <- paste0(Line, ', Mapped on ',Names[XMap[x]])
      }
      if(Rounded & (RFactor[x]>0)){
        Line <- paste0(Line,', Rounded, factor = ',RFactor[x])
      }
      if(Discretized & (NCat[x]>0)){
        Line <- paste0(Line,', Optimally Discretized, max cat = ',NCat[x])
      }
      print(Line)
    }
  }

  colnames(Y)<-Names[YSelected]
  colnames(X)<-Names[XSelected]

  return(list(Y=Y, X=X, XType = XType, Desc=Desc, DataOrig=DataOrig))

}

computeProb<-function(Y,Indices){

  dimY<-dim(table(Y))
  categoriesY<-names(table(Y))
  Sizes<-list()
  Prob<-list()

  Sizes[[1]]<-table(Y)
  names(Sizes[[1]])<-NULL
  Prob[[1]]<-as.numeric(as.vector(Sizes[[1]]))/NROW(Y)

  for (i in 2:length(Indices)) {
    Ynode<-as.factor(Y[Indices[[i-1]],])
    if(length(Ynode)!=0){
      if(dim(table(Ynode))<dimY){#not all categories in the node
        Sizes[[i]]<-table(factor(Ynode,categoriesY))
      }else{#normal case
        Sizes[[i]]<-table(Ynode)
      }
      names(Sizes[[i]])<-NULL
      Prob[[i]]<-as.numeric(as.vector(Sizes[[i]]))/length(Ynode)
    }
  }


  return(list(Sizes=Sizes, Prob=Prob))

}

#' Main function of the package. It performs the Exact Tree method.
#'
#' @param formula a description of the model to be fit. The format is \code{Y ~  X1 + \dots + Xn},
#'   where the variable before ~ represents the dependent variable and the variables after
#'   the ~ are the independent variables.
#' @param data Dataset to be analyzed. Note: If data contains ordinal variables, transform them to numerical
#' before using this function. Otherwise, they will be treated as nominal variables.
#' @param map [ori vars] sets for each variable whether the tree will depict
#'                       thresholds based upon original values (var number), or recoded values (0).
#' @param original [TRUE/FALSE] sets for each variable whether the tree will depict
#'                       thresholds based upon original values (true), or recoded values (false).
#' @param round [factors] autorecoding based upon rounded categories; the categories will
#'                       be round after multiplying with the factor, and devided by the
#'                       factor after rounding
#' @param discretize [ncat] optimal discretization while minimizing SS within
#'                       a group, indicate with a number how many categories;set ncat=0 for each variable that is to
#'                       be left unaltered
#' @param selV list containing the output obtained using \code{\link{SelectVar}}.
#' Use this list instead of the previous inputs if you want to perform SelectVars outside ETree.
#' @param control a list with control parameters as returned by \code{\link{ETree.control}}.
#' @param verbose logical, if TRUE, prints information about the progress of the algorithm.
#'
#'
#'
#' @details The function results in a global tree (for given maxsize and maxdepth) and transforms the output to obtain summary information and plot the tree.
#'
#'
#' @return Returns the following 6 elements:
#'  \item{h}{contains the objective function value (fit) of the best tree.}
#'  \item{Tree}{contains the largest Tree table.}
#'  \item{hAll}{(OPTIONAL) contains the objective function values for all trees, if more than one is requested}
#'  \item{TAll}{(OPTIONAL) contains Tree tables for all trees, if more than one is requested.}
#'  \item{Transf_Trees}{contains an object of class ExactTree that can be used in \code{\link{plot.ETree}} and \code{\link{summary.ETree}}}
#'  \item{CVOutput}{Cross validation results for all the requested trees.}
#'
#'
#' @seealso \code{\link{summary.ETree}}, \code{\link{SelectVar}},
#'   \code{\link{plot.ETree}},\code{\link{ETree.control}},
#'   \code{\link{predict.ETree}},\code{\link{prune.ETree}}
#'
#' @examples
#' \donttest{
#'   data(iris)
#'   # Fit an Exact Tree model
#'   controlEtree <- ETree.control(measure=0, maxsize = 4, maxdepth = 3,
#'   minbucket = 5, ncv=5, alltreesizes = FALSE)
#'   ETree(Species ~ Sepal.Length + Sepal.Width + Petal.Length + Petal.Width,
#'   control= controlEtree, data = iris)
#' }
#'
#'
#' @keywords tree
#' @keywords exact
#'
#' @importFrom pracma histc repmat randperm Reshape strcmp isempty fprintf
#' @importFrom DescTools DecToBin BinToDec
#'
#' @export
#' @useDynLib ExactTree, .registration=TRUE
ETree<-function(formula=NULL,data=NULL, map=NULL, original=NULL, round=NULL, discretize=NULL, selV=NULL, control=NULL, verbose=TRUE){

  # library(matlab)
  # library(pracma)


  #We need either formula and data or Y,X and Desc
  if(is.null(selV)){#!is.null(formula)){
    if(is.null(data)){
      stop("Data is missing.")
    }
    if(is.null(formula)){
      stop("formula is missing.")
    }
    outSelVar<-SelectVar(formula=formula,data=data,map=map, original=original, round=round, discretize=discretize, verbose=verbose)
    Y<-outSelVar$Y
    X<-outSelVar$X
    Desc<-outSelVar$Desc
    Data<-outSelVar$DataOrig
    XType<-outSelVar$XType
    Yvarnames<-colnames(Y)
    Xvarnames<-colnames(X)

  }else{

    # if(!is.null(data)){
    #   stop("formula is missing.")
    # }

    if(is.null(selV$Y) | is.null(selV$X) | is.null(selV$Desc) | is.null(selV$DataOrig)){
      stop("If formula is missing, the function requires Y, X, DataOrig and Desc. These variables are obtained using SelectVar()")
    }

    Y<-selV$Y
    X<-selV$X
    Desc<-selV$Desc
    Data<-selV$DataOrig
    XType<-selV$XType

    Yvarnames<-colnames(Y)
    Xvarnames<-colnames(X)

  }


  AlgoType<-1
  CVVec<-c()

  if(is.null(control)){
    stop("Control is missing.")
  }else{
    AlgoType<-control$AlgoType
    Measure<-control$Measure
    MaxSize<-control$MaxSize
    MaxDepth<-control$MaxDepth
    MinBucket<-control$MinBucket
    MinH<-control$MinH
    ncv<-control$ncv
    cvVector<-control$cvVector
    #heterogeneityonly<-control$heterogeneityonly
    BoundH<-control$BoundH
    #branchandbound<-control$branchandbound
    #alltreesizes<-control$alltreesizes
    #sortedpsplits<-control$sortedpsplits
    #lookaheadheuristic<-control$lookaheadheuristic
    LookAheadDepth<-control$LookAheadDepth
  }



  if(!is.null(ncv) && ncv>=1){
    NCV<-ncv
    n<-dim(X)[1]
    CVVec<-matrix(0, nrow = n, ncol = 1)
    P<-randperm(n)
    bs<-trunc(n/NCV) #fix(n/NCV);
    for(c in 1:NCV){
      if(c<NCV){
        Pt<-P[(1+(c-1)*bs):(c*bs)]
      }else{
        Pt<-P[(1+(c-1)*bs):n] #pick up remainder
      }
      CVVec[Pt]<-c
    }
  }else if(!is.null(cvVector)){
    CVVec <- cvVector
    NCV <- max(CVVec)
    f <- Freq(CVVec)#Freq(CVVec,1);
    if(verbose){
      print(paste0("Sizes of cross-validation classes: ", t(f)))#disp ( ['Sizes of cross-validation classes: ',num2str(t(f))] )
    }
    if(min(Freq(CVVec))==0){
      warning("Misspecification of CV vector: empty class(es) detected")
      #print("Warning: misspecification of CV vector: empty class(es) detected")
    }
  }


  argout	<- list()


  if(BoundH==0){
    argout<-OptTreeGate(defY=Y,X=X,XType=XType,defMeasure=Measure,defMaxSize=MaxSize,defMaxDepth=MaxDepth,defMinNodeSize=MinBucket,defMinH=MinH,AlgoType=AlgoType,CVVec=CVVec,
                        LookAheadDepth=LookAheadDepth, verbose = verbose)
  }else{
    argout<-OptTreeGate(defY=Y,X=X,XType=XType,defMeasure=Measure,defMaxSize=MaxSize,defMaxDepth=MaxDepth,defMinNodeSize=MinBucket,defMinH=MinH,AlgoType=AlgoType,CVVec=CVVec,
                        BoundH=BoundH,LookAheadDepth=LookAheadDepth, verbose = verbose)
  }



  Ntrees<-ifelse(is.na(dim(argout$TAll)[3]),1,dim(argout$TAll)[3])



  #argout$Indices
  FinalIndex<-list()
  FinalSizes<-list()
  FinalProb<-list()
  ProbLeaves<-list()


  if(Ntrees>1){

    for (i in 2:Ntrees) {


      IndexList<-getIndex(prevNode=1, side="left", node=1, Tree=argout$TAll[,,i], Index=argout$Indices[[i]], keepnode = NULL)

      FinalIndex[[i]]<-IndexList$Index[IndexList$keep]


      #Need to fix it for multiple trees
      if(Measure!=0){
        SizesProb<-computeProb(Y,IndexList$Index)#computeProb(Y,argout$Indices[[i]])
        FinalSizes[[i]]<-SizesProb$Sizes
        FinalProb[[i]]<-SizesProb$Prob
        ProbLeaves[[i]]<-FinalProb[[i]][IndexList$keep+1]
      }


    }
    FinalIndex[[1]]<-1:NROW(Y)

    if(Measure!=0){
      FinalSizes[[1]]<-table(Y)
      names(FinalSizes[[1]])<-NULL
      FinalProb[[1]]<-as.numeric(as.vector(FinalSizes[[1]]))/NROW(Y)
      ProbLeaves[[1]]<-FinalProb[[1]]
    }


  }else{

    IndexList<-getIndex(prevNode=1, side="left", node=1, Tree=argout$TAll[,,1], Index=argout$Indices[[1]], keepnode = NULL)
    FinalIndex[[1]]<-IndexList$Index[IndexList$keep]

    if(Measure!=0){
      SizesProb<-computeProb(Y,IndexList$Index)#computeProb(Y,argout$Indices[[1]])
      FinalSizes[[1]]<-SizesProb$Sizes
      FinalProb[[1]]<-SizesProb$Prob
      ProbLeaves[[1]]<-FinalProb[[1]][IndexList$keep+1]
    }

  }



  #evaluation function

  Evaluation<-evalFunc(Y=Y,Indices=FinalIndex,TAll=argout$TAll_list, measure=Measure, NTrees=Ntrees)




  Transf_Tree<-list()
  if(Ntrees==1){

    Tree<-matrix(argout$TAll, nrow = NROW(argout$TAll), ncol = NCOL(argout$TAll))

    NTN<-sum(Tree[,1]==0)

    if(Measure>0 || is.null(CVVec)){
      CVdata<-NULL
    }else{
      CVout<-argout$CVOutput
      CVdata<-matrix(c(CVout$Size,CVout$Error,CVout$Rel_error,CVout$SE,CVout$SE2,CVout$CHECK),nrow = 1)
      colnames(CVdata)<-c("Size","Error","Rel. Error", "SE", "SE2", "CHECK")
    }


    Transf_Tree[[1]]<- transformResults(TAll=argout$Tree,NtermNodes=NTN,X=X,Y=Y, Desc=Desc, CVdata=CVdata, origData = data[,colnames(X)])



  }else{

    for (i in 1:Ntrees) {

      if(Measure>0 || is.null(CVVec)){
        CVdata<-NULL
      }else{
        CVout<-argout$CVOutput
        CVdata<-matrix(c(CVout$Size[i],CVout$Error[i],CVout$Rel_error[i],CVout$SE[i],CVout$SE2[i],CVout$CHECK[i]),nrow = 1)
        colnames(CVdata)<-c("Size","Error","Rel. Error", "SE", "SE2", "CHECK")
      }
      Transf_Tree[[i]]<-transformResults(TAll=argout$TAll_list,NtermNodes=i,X=X,Y=Y, Desc=Desc, CVdata=CVdata, origData = data[,colnames(X)])
    }

  }




  #class(Transf_Tree)<-"ETree"

  Output<-list(h=argout$h,Tree=argout$Tree,hAll=argout$hAll,TAll=argout$TAll_list, Transf_Trees=Transf_Tree, CVOutput=argout$CVOutput, Index=FinalIndex, Sizes=FinalSizes, Prob=FinalProb, ProbLeaves=ProbLeaves, Evaluation=Evaluation, Desc=Desc, Data=Data, Yvarnames=Yvarnames, Xvarnames=Xvarnames)

  class(Output)<-"ETree"

  return(Output)
}

DPOP<-function(Y,k,defMeasure=NULL, Prior=NULL, LossM=NULL, ClassSizes=NULL, verbose = TRUE){
# % DP for rank-order constrained partitioning n objects described by Y, into k groups
# % using the heterogeneity measure indicated by measure
# % H(2:k) returns the overall heterogeneity values for the partition into 2..k clusters
# % where H(1) returns the overall heterogeneity
# % P returns all partitions, for 2 up to k clusters
# % I returns the sort order, on which the splits are defined
if (is.null(defMeasure)<3){
  defMeasure <- 0
}


n<-NROW(Y)
m<-NCOL(Y)

if(m>1){
  warning('The function only works for one dimensional dependent variable Y')
  #print('only for one dimensional dependent variable Y')
  return()
}


#measure<-globalenv()$measure
#Heap<-globalenv()$Heap
#SplitHeap<-globalenv()$SplitHeap

#tic;
#cpu1=cputime;
time<-proc.time()

measure<-defMeasure
O<-sort(Y)
I<-order(Y)
Heap<-rep(0,n)
SplitHeap<-matrix(0, nrow = n, ncol= (k-2) )
H<-c()
#Global Heterogeneity
H[1]<-GetHeterogeneity(Y=O, measure=measure, Prior=Prior, LossM=LossM, ClassSizes=ClassSizes)

#Compute all heterogeneity values for ordered subsets of Y (=O)
Hetero<-matrix(0, nrow = n, ncol=n)
if(measure==0){ #speedup for minimizing SS
  SO<-c(0,cumsum(O))
  SSO<-c(0,cumsum(O^2))

  for(s in 1:(n-1)){ #%split s= split between object s and s+1
    for(ss in (s+1):n){
      ni<-(ss+1-s)
      Hetero[s,ss]<-(SSO[ss+1]-SSO[s])-((SO[ss+1]-SO[s])^2/ni) #raw score formula using cum sums
    }
    Heap[s]<-Hetero[1,s]
  }
}else if(measure==1000){ #measure ==0 vectorized no additional speedup in MATLAB 7
  SO<-c(0,cumsum(O))
  SSO<-c(0,cumsum(O^2))
  for(s in 1:(n-1)){ #%split s= split between object s and s+1
    Hetero[s,(s+1):n]<-(SSO[(s+2):(n+1)]-SSO[s])-((SO[(s+2):(n+1)]-SO[s])^2/(2:(1+n-s)))
    Heap[s]<-Hetero[1,s]
  }
}else{
  for(s in 1:(n-1)){ #%split s= split between object s and s+1
    for(ss in (s+1):n){
      Hetero[s,ss]=GetHeterogeneity(Y=O[s:ss], measure=measure, Prior=Prior, LossM=LossM, ClassSizes=ClassSizes)
      #Hetero(s,ss)=sum((O(s:ss)-mean(O(s:ss))).^2);
    }
    Heap[s]<-Hetero[1,s]
  }
}

#cpu2=cputime;
time2<-proc.time()
#%disp(['time for first phase ',num2str(cpu2-cpu1)]);
#%toc

#%Do all, but last stage (stage=ki)
for(ki in 2:(k-1)){
  for(e in seq(n,ki,-1)){
    minH<-Heap[ki-1]+Hetero[ki,e]
    minS<-ki-1
    for(s in ki:(e-1)){ #split s= split between object s and s+1
      newH<-Heap[s]+Hetero[s+1,e]
      if(newH<minH){
        minH<-newH
        minS<-s
      }
    }
    Heap[e]<-minH
    SplitHeap[e,ki-1]<-minS
  }
  H[ki]<-Heap[n]
}

#%Do Last Stage
minH<-Heap[k-1]+Hetero[k,n]
minS<-k-1
for(s in k:(n-1)){ #%split s= split between object s and s+1
  newH<-Heap[s]+Hetero[s+1,n]
  if(newH<minH){
    minH<-newH
    minS<-s
  }
}
H[k]<-minH

time3<-proc.time()-time #cpu2=cputime;
if(verbose){
  print(paste0('elapsed time ', time3[3]))
  print(paste0('cpu time ', time3[2]))
}

#%Get optimal partitions
P<- matrix(0,nrow = k, ncol = (k-1))
for(ki in 2:(k-1)){
  P[ki,]<-ResolveSplits(n,ki,P[ki,1:(k-1)],SplitHeap)
}
P[k,k-1]<-minS
P[k,]<-ResolveSplits(minS,k-1,P[k,1:(k-1)],SplitHeap)

return(list(H=H,P=P,I=I, measure=measure, Heap=Heap))

}


ResolveSplits<-function(n,k,SP,SplitHeap){
#%Resolve optimal partition from split heap
#global SplitHeap;
#SplitHeap<-globalenv()$SplitHeap
SP[k-1]<-SplitHeap[n,k-1]
if(k>2){
  SP<-ResolveSplits(SP[k-1],k-1,SP,SplitHeap)
}

return(SP)
}

evalFunc<-function(Y,Indices,TAll, measure, NTrees){


  N<-NROW(Y)
  alpha<-0.25 #1 is another option

  criterion<-c()

  if(measure==0){ #Y continuous


    for(i in 1:NTrees){
      Tree<-TAll[[i]]
      TerminalNodes<-which(Tree[,1]==0)

      if(NCOL(Y)==1 || NROW(Y)==1){ #Y is 1-dimensional
        ordY<-c()
        predY<-c()
        for(j in 1:length(TerminalNodes)){
          ordY<-c(ordY,Y[Indices[[i]][[j]]])
          TermNode<-TerminalNodes[j]
          predY<-c(predY,rep(Tree[TermNode,7],Tree[TermNode,6]))
        }

        SSq<-sum((ordY-predY)^2)
        loss<-N*log(SSq)
        complexity<-alpha*4*(length(TerminalNodes)+1)*log(N)

        criterion[i]<-loss+complexity


      }else{ #Y is multidimensional

        criterion<-NULL
        warning("Evaluation function is not implemented for multiple Y variables yet")
        #MISSING

      }




    }



  }else{ #Y discrete



    for(i in 1:NTrees){
      Tree<-TAll[[i]]
      TerminalNodes<-which(Tree[,1]==0)

      if(NCOL(Y)==1 || NROW(Y)==1){ #Y is 1-dimensional

        ordY<-c()
        predY<-c()
        for(j in 1:length(TerminalNodes)){
          ordY<-c(ordY,Y[Indices[[i]][[j]]])
          TermNode<-TerminalNodes[j]
          predY<-c(predY,rep(Tree[TermNode,7],Tree[TermNode,6]))
        }

        Indicator<-ordY==predY

        loss<-2*sum(Indicator)
        complexity<-alpha*(length(TerminalNodes))*log(N)

        criterion[i]<-loss+complexity

      }else{ #Y is multidimensional


        criterion<-NULL
        warning("Evaluation function is not implemented for multiple Y variables yet")
        #Missing


      }

    }




  }


  return(criterion)


}

getFreqY<-function(Y){
#gives frequencies of classes in Y, Y should be autorecoded (i.e. 1,...)
#create edges
#library(pracma)
minY<-min(Y)
maxY<-max(Y)
edges<-minY:maxY
f<-histc(Y,edges)$cnt

return(f)
}

bitset<-function(A,bit,V){
#library(DescTools)

  Abit<-DecToBin(A)

  lengthA<-nchar(Abit)

if(lengthA<bit){
  extraLength<-bit-lengthA-1
  for (i in 1:extraLength) {
    Abit<-paste0("0",Abit)
  }

  if(V==0){
    out<-A
  }else{
    bit_return<-paste0(1,Abit)
    out<-BinToDec(bit_return)
  }


}else{
  if(V==0){
    bit_return<-paste0(substr(Abit,1,lengthA-bit),0,substr(Abit,lengthA-bit+2,lengthA))
    out<-BinToDec(bit_return)
  }else{
    bit_return<-paste0(substr(Abit,1,lengthA-bit),1,substr(Abit,lengthA-bit+2,lengthA))
    out<-BinToDec(bit_return)
  }
}


  return(out)
}

GetCVErrorAndSE<-function(Y,Ypred,aMeasure=NULL, measure, Prior, LossM, ClassSizes){
#global measure Prior LossM ClassSizes;

# measure<-globalenv()$measure
# Prior<-globalenv()$Prior
# LossM<-globalenv()$LossM
# ClassSizes<-globalenv()$ClassSizes

if(!is.null(aMeasure)){
  ActiveMeasure <- aMeasure
}else{
  ActiveMeasure <- measure
}

if(ActiveMeasure==0){
  GSSEASEout<-GetSSErrorAndSE(Y,Ypred)
  Error<-GSSEASEout$Error
  SE<-GSSEASEout$SE
  SE2<-GSSEASEout$SE2
}else{
  stop('Undefined Heterogeneity Measure for SE')
}

return(list(Error=Error,SE=SE,SE2=SE2))

}




GetSSErrorAndSE<-function(Y,Ypred){
#Give Sum of Squared error in prediction
#library(pracma)

n<-dim(Y)[1]
E<-Y-Ypred
Error<-sum(E*E)
tE<-Y-repmat(colMeans(Y),n,1)
ym<-dim(Y)[2]
SEd<-matrix(0,nrow=ym,ncol=2)
for(d in 1:ym){
  s2<-sum(tE[,d]^2)/n
  Rcv<-sum(E^2)/n #MISSING colsums or sum?
  REcv<-Rcv/s2
  s2_1<-sum(E[,d]^4)/n-Rcv^2#if Rcv is a vector, then problem. I assume it is a vector in line 37
  s2_2<-sum(tE[,d]^4)/n-s2^2
  s_12<-sum((E^2)*(tE^2))/n-Rcv*s2
  SEd[d,1]<-REcv*sqrt(( (s2_1/Rcv^2)-(2*s_12/(Rcv*s2))+(s2_2/s2^2) )/n)
  SEd[d,2]<-sqrt(sum(Center(E[,d]^2)^2))/sum(tE[,d]^2)
}
SE<-SEd[1,1]
SE2<-SEd[1,2]

return(list(Error=Error,SE=SE,SE2=SE2))
}




# GetMisclassificationErrorAndSE<-function(Y){#MISSING In matlab also doesnt exist Error and SE
#   h<-0
#   yn<-dim(Y)[1]
#   ym<-dim(Y)[2]
#   if(yn==1){
#     return()
#   }
#
#   for(d in 1:ym){
#     O<-sort(Y[,d])
#     hf<-0
#     s<-1
#     for(i in 2:yn){
#       if(O[i]!=O[i-1]){
#         hf<-max(hf,(i-s))
#         s<-i
#       }
#     }
#     if(O[yn]==O[yn-1]){
#       hf<-max(hf,(yn+1-s))
#     }
#     h<-h+(yn-hf)
#   }
#
# return(list(Error=Error,SE=SE))
# }



# GetExpectedRiskAndSE<-function(Y, Prior, LossM, ClassSizes){
# #global Prior LossM ClassSizes;
# #library(pracma)
#
# # Prior<-globalenv()$Prior
# # LossM<-globalenv()$LossM
# # ClassSizes<-globalenv()$ClassSizes
# h<-0
# for(d in 1:dim(Y)[2]){
#   nCat <- max(dim(ClassSizes[[d]]))
#   L<-LossM[[d]]
#   f<-histc(x=Y[,d],edges=1:nCat)
#   f<-as.vector(f) #f has different shape if Y has one row
#   E<-as.vector(Prior[[d]])*f/ClassSizes[[d]]
#   pa<-colSums(E)#colSums or sum?
#   piA<-E/pa
#   ra<-min(t(L)%*%piA)
#
#   h<-h+pa%*%ra
# }
#
# return(list(Error=Error,SE=SE))
# }

UnravelCells<-function(C){
#flats out all elements of all (nested) cells into one vector
#library(pracma)
Vector <- c()
if (!isempty(C)){
  Vector <- UnravelCellsR(C)
}

return(Vector)

}




UnravelCellsR<-function(C){#put all the lists in a column list.
#library(pracma)
C<-Reshape(C,prod(dim(C)),1)
n<-max(dim(C))
Vector <- c()
for(i in 1:n){
  if (is.list(C[[i]])){
    Vector <- c(Vector,UnravelCellsR(C[[i]])) #c or cbind? or rbind?
  }else{
    ci<-C[[i]]
    ci<-Reshape(ci,prod(dim(ci)),1)
    Vector <- c(Vector,t(ci)) #c or cbind? or rbind?
  }
}

return(Vector)

}

#' Visualisation of a Exact Tree
#'
#'
#' @param x transformed tree of class \code{ExactTree}.
#' @param TerminalNodes Number of terminal nodes of the tree to plot.
#' @param digits specified number of decimal places of the splitpoints in the graph
#'   (default is 2).
#' @param \dots additional arguments to be passed.
#'
#' @return A plot of the tree with the specified number of terminal nodes.
#'
#' @examples
#' \donttest{
#'   data(iris)
#'   # Fit an Exact Tree model
#'   controlEtree <- ETree.control(measure=0, maxsize = 4, maxdepth = 3,
#'   minbucket = 5, ncv=5, alltreesizes = FALSE)
#'   tree<-ETree(Species ~ Sepal.Length + Sepal.Width + Petal.Length + Petal.Width,
#'   control= controlEtree, data = iris)
#'   plot(tree, TerminalNodes=4)
#' }
#'
#' @references Torsten Hothorn and Achim Zeileis (2013). partykit: A Toolkit for
#'   Recursive Partytioning. R package version 0.1-5.
#'
#' @seealso \code{\link{ETree}},\code{\link{summary.ETree}}
#' @keywords plot
#' @keywords as.party
#'
#' @importFrom partykit as.party character_split id_node is.terminal kids_node node_party nodeids
#'   party partynode partysplit plot.party split_node
#' @importFrom graphics par plot
#' @importFrom grid gpar grid.layout grid.lines grid.points grid.polygon grid.rect grid.text
#'   grid.yaxis popViewport pushViewport unit upViewport viewport grid.newpage grid.draw
#' @importFrom gridtext richtext_grob
#' @importFrom methods is
#'
#' @export
plot.ETree<-function(x, TerminalNodes=NULL, digits=2, ...){

  object<-x

  origVar<-c(object$Yvarnames, object$Xvarnames)
  origData<-object$Data
  x<-object$Transf_Trees



  if(is.null(TerminalNodes)||(length(x)==1)){

    if(length(x)!=1){
      stop("Terminal nodes is NULL. Please, indicate the number of terminal nodes.")
    }

    TAll<-x[[1]]
    TerminalNodes<-NROW(TAll$li)

  }else{

    if(length(x)<TerminalNodes){
      stop("TerminalNodes is larger than the maximum number of terminal nodes.")
    }

    TAll<-x[[TerminalNodes]]

  }


  TAll$origData<-origData[,origVar]
  colnames(TAll$origData)<-TAll$var.names
  TAll$Desc<-object$Desc

    if(TerminalNodes==1){
      #Here the plot for the tree with only root node

      Tree_plot<-TAll
      node<-1

      text<-paste0("*N* = ", Tree_plot$li[node,2] ,"<br>*&#375;* = ", round(Tree_plot$li[node,3],digits = digits),"<br>*h* = ",round(Tree_plot$li[node,4],digits = digits))
      box_gp <- gpar(col = "black", fill = "#DDD8C2", lty = 1)


      plotNode1<-richtext_grob(
        text, hjust=0,
        padding = unit(c(6, 6, 4, 6), "pt"),
        r = unit(c(0, 2, 4, 8), "pt"),
        gp = gpar(), box_gp = box_gp
      )


      grid.newpage()
      grid.draw(plotNode1)


    }else{
      #x is an object of class "ETrees"
      if(is.numeric(TAll$si[,5])){
        TAll$si[,5] <- round(TAll$si[,5],digits=digits)
      }
      TAll$li[,3] <- round(TAll$li[,3],digits=digits)
      TAll$li[,4] <- round(TAll$li[,4],digits=digits)
      party.ExactTree <- as.party(TAll)
      plot(party.ExactTree, inner_panel= node_ETrees, terminal_panel=terminal_ETrees, ...)
      return(invisible(party.ExactTree))
    }
}





# x<-1
# y<-1
#
#
# grid.newpage()
# grid.rect(x = 0.45, y = 0.5,
#           width = 0.25, height = 0.25,
#           just = "left", hjust = NULL, vjust = NULL,
#           default.units = "npc", name = NULL,
#           gp=gpar(), draw = TRUE, vp = NULL)
# x_d <- 0.5
# y_d <- 0.5
# node<-which(Tree_plot[,1]==0)
# grid.text(paste0("N = ", round(Tree_plot[node,6],digits = digits) ,"\ny = ", round(Tree_plot[node,7],digits = digits),"\nh = ",round(Tree_plot[node,5],digits = digits)), x=x_d, y=y_d,
#           just="left",gp=gpar(fontsize=10, col="black"))
#
#
#
#
# grid.newpage()
# x <- stats::runif(20)
# y <- stats::runif(20)
# rot <- stats::runif(20, 0, 360)
# grid.text("SOMETHING NICE AND BIG", x=x, y=y, rot=rot,
#           gp=gpar(fontsize=20, col="grey"))
# grid.text("SOMETHING NICE AND BIG", x=x, y=y, rot=rot,
#           gp=gpar(fontsize=20), check=TRUE)
# grid.newpage()
# draw.text <- function(just, i, j) {
#   grid.text("ABCD", x=x[j], y=y[i], just=just)
#   grid.text(deparse(substitute(just)), x=x[j], y=y[i] + unit(2, "lines"),
#             gp=gpar(col="grey", fontsize=8))
# }
# x <- unit(1:4/5, "npc")
# y <- unit(1:4/5, "npc")
# grid.grill(h=y, v=x, gp=gpar(col="grey"))
# draw.text(c("bottom"), 1, 1)
# draw.text(c("left", "bottom"), 2, 1)
#}

# functions supporting main plot.ETrees() function

## function to make indices for node assignment of levels of a factor variable (used in to.party)
indexID <- function(nodeID, ETrees.out) {
  splitlevels <- unlist(strsplit(as.character(ETrees.out$si[ETrees.out$si[,1]==nodeID, 5]), split = ", "))
  everylevel <- ETrees.out$Desc[[(which(ETrees.out$var.names == ETrees.out$si[ETrees.out$si[,1]==nodeID, 4]))-1]]#We subtract one because Y is not in Desc. #Old: sapply(ETrees.out$origData[ETrees.out$si[ETrees.out$si[,1]==nodeID, 4]], FUN = function(p) sort(unique(p)))#sapply(ETrees.out$data[ETrees.out$si[ETrees.out$si[,1]==nodeID, 3]], FUN = function(p) sort(unique(p)))#This does not seem fine.


  index1 <- numeric(length(splitlevels)) # index factor levels used for split
  for (i in 1:length(index1)) {
    index1[i] <- which(everylevel == splitlevels[i])
  }

  #Temporary fix until different categories can be placed together as splitpoints.
  if(sapply(ETrees.out$origData[colnames(ETrees.out$origData)==ETrees.out$si[ETrees.out$si[,1]==nodeID,4]], is.factor) == TRUE){
    index1<-index1-1 #We take the previous category because we split with using < & >=
  }
  index1<-1:index1

  # indices of all levels not used in split
  index2 <- 1:length(everylevel)
  index2 <- index2[-index1]

  Ls <- integer(length(everylevel))
  Ls[index1] <- 1L; Ls[index2] <- 2L

  return(Ls)
}


## conversion function ##
to.party <- function(nodeID,ETrees.out, ...){
  if("ETree" %in% is(nodeID)){ETrees.out <- nodeID; nodeID <- 1L}
  if(is.null(ETrees.out$var.names)) ETrees.out$var.names <- colnames(ETrees.out$data)
  if(nodeID %in% ETrees.out$li[,1]) return(partynode(id=as.integer(nodeID),
                                                     info = ETrees.out$li[ETrees.out$li[,1]==nodeID,]
  ))

  if(nodeID %in% ETrees.out$si[,1]) {
    #If the variable used for splitting is numeric, use breaks.
    if(sapply(ETrees.out$origData[colnames(ETrees.out$origData)==ETrees.out$si[ETrees.out$si[,1]==nodeID,4]], is.factor) == FALSE){
      split_number<-which(ETrees.out$si[,1]==nodeID)
      return(partynode(id=as.integer(nodeID),
                       split =  partysplit(varid = as.integer(which(ETrees.out$var.names==ETrees.out$si[ETrees.out$si[,1]==nodeID,4])),
                                           breaks = as.numeric(ETrees.out$si[ETrees.out$si[,1]==nodeID,5]), right=FALSE), #right=FALSE added 10/08/2022
                       kids = lapply(c(ETrees.out$si[split_number,2],ETrees.out$si[split_number,3]),to.party,ETrees.out)#MISSING, check this line. Old line: lapply(c(nodeID*2,nodeID*2+1),to.party,ETrees.out)
      )
      )}

    #If the variable used for splitting is a factor, use index.
    if(sapply(ETrees.out$origData[colnames(ETrees.out$origData)==ETrees.out$si[ETrees.out$si[,1]==nodeID,4]], is.factor) == TRUE){
      split_number<-which(ETrees.out$si[,1]==nodeID)
      return(partynode(id=as.integer(nodeID),
                       split =  partysplit(varid = as.integer(which(ETrees.out$var.names==ETrees.out$si[ETrees.out$si[,1]==nodeID,4])),
                                           index = as.integer(indexID(nodeID=nodeID, ETrees.out=ETrees.out))),#,right=FALSE), #right=FALSE added 10/08/2022
                       kids = lapply(c(ETrees.out$si[split_number,2],ETrees.out$si[split_number,3]),to.party,ETrees.out))#MISSING: OLD CODE: lapply(c(nodeID*2,nodeID*2+1),to.party,ETrees.out))
      )}
  }
}



#' Transformation of a Exact Tree object to party object
#'
#'
#' @param obj tree of class \code{ETree}.
#' @param nodeID Node identification.
#' @param \dots additional arguments to be passed.
#' @returns object transformed to a \code{constparty} object.
#' @keywords as.party
#'
#' @importFrom partykit party
#'
#' @export
## conversion method
as.party.ETree <- function(obj, nodeID=1L, ...){
  ETrees.out <- obj
  ETreesNodes <- to.party(nodeID,ETrees.out,...)
  # browser() ##########
  # Old code:
  # party.object <- party(ETreesNodes,ETrees.out$data
  #                       , fitted=data.frame(
  #                         "(fitted)"=colnames(ETrees.out$nind)[
  #                           apply(ETrees.out$nind==1,1,which)],
  #                         "(response)"=ETrees.out$data[,1],
  #                         check.names=FALSE)
  #                       , terms=terms(as.formula(paste(
  #                         colnames(ETrees.out$data)[1],"~",
  #                         paste(colnames(ETrees.out$data)[-1],collapse="+"))
  #                       )),
  #                       ...
  # )

  party.object <- party(ETreesNodes, ETrees.out$origData, #ETrees.out,
                        ...
  )
  party.object$ni <- ETrees.out$li
  class(party.object) <- c("constparty",class(party.object))
  return(party.object)
}


## ETrees inner node
node_ETrees <- function(obj, id = TRUE, abbreviate = FALSE, fill = "white", gp = gpar())
{
  meta <- obj$data
  nam <- names(obj)

  extract_label <- function(node) {
    if(is.terminal(node)) return(rep.int("", 2))

    varlab <- character_split(split_node(node), meta)$name
    if(abbreviate > 0) varlab <- abbreviate(varlab, as.numeric(abbreviate))

    plab <- ""
    return(c(varlab, plab))
  }

  maxstr <- function(node) {
    lab <- extract_label(node)
    klab <- if(is.terminal(node)) "" else unlist(lapply(kids_node(node), maxstr))
    lab <- c(lab, klab)
    lab <- unlist(lapply(lab, function(x) strsplit(x, "\n")))
    return(lab[which.max(nchar(lab))])
  }

  nstr <- maxstr(node_party(obj))
  if(nchar(nstr) < 6) nstr <- "aAAAAa"

  ### panel function for the inner nodes
  rval <- function(node) {
    node_vp <- viewport(
      x = unit(0.5, "npc"),
      y = unit(0.5, "npc"),
      width = unit(1, "strwidth", nstr) * 1.3,
      height = unit(3, "lines"),
      name = paste("node_inner", id_node(node), sep = ""),
      gp = gp
    )
    pushViewport(node_vp)

    xell <- c(seq(0, 0.2, by = 0.01),
              seq(0.2, 0.8, by = 0.05),
              seq(0.8, 1, by = 0.01))
    yell <- sqrt(xell * (1-xell))

    lab <- extract_label(node)
    fill <- rep(fill, length.out = 2)

    grid.polygon(x = unit(c(xell, rev(xell)), "npc"),
                 y = unit(c(yell, -yell)+0.5, "npc"),
                 gp = gpar(fill = fill[1]))

    grid.text(lab[1], y = unit(1.5 + 0.5 * FALSE, "lines"))

    if(id) {
      nodeIDvp <- viewport(x = unit(0.5, "npc"), y = unit(1, "npc"),
                           width = max(unit(1, "lines"), unit(1.3, "strwidth", nam[id_node(node)])),
                           height = max(unit(1, "lines"), unit(1.3, "strheight", nam[id_node(node)])))
      pushViewport(nodeIDvp)
      popViewport()
    }
    upViewport()
  }
  return(rval)
}
class(node_ETrees) <- "grapcon_generator"


## ETrees terminal node plot ##
terminal_ETrees <- function(obj,
                            col = "black",
                            fill = "lightgray",
                            width = 0.5,
                            yscale = NULL,
                            ylines = 3,
                            cex = 0.5,
                            id = TRUE,
                            gp = gpar())
{
  ni <- obj$ni #leaf information
  ni.temp <- ni
  # d.CI <- data.frame(
  #   lower = ni[,8] - 1.96 * ni[,9],
  #   mean = ni[,8],
  #   upper = ni[,8] + 1.96 * ni[,9]
  # )
  #yscale <- c(0-max(abs(d.CI)),max(abs(d.CI))) + c(-0.1, 0.1) * max(abs(d.CI))
  rval <- function(node) { # core plotting function
    nid <- id_node(node)
    top_vp <- viewport(layout = grid.layout(nrow = 3, ncol = 3, # define viewport
                                            widths = unit(c(ylines, 1, 1),
                                                          c("lines", "null", "lines")),
                                            heights = unit(c(1, 1, 2), c("lines","null","lines"))),
                       width = unit(1, "npc"),
                       height = unit(1, "npc") - unit(2, "lines"),
                       name = paste("node_ETrees", nid, sep = ""),
                       gp = gp)

    pushViewport(top_vp)
    grid.rect(gp = gpar(fill = "transparent", col =0))

    ind2 <- nid == nodeids(obj,terminal=TRUE)

    ## main title
    bottom <- viewport(layout.pos.col=2, layout.pos.row=3)
    pushViewport(bottom)
    if(id){
      grid.text(sprintf("Leaf %s",
                        which(nid==nodeids(obj,terminal=TRUE)))
      )
    }else{grid.text("")}
    popViewport()

    plot <- viewport(layout.pos.col = 2, layout.pos.row = 2,
                     xscale = c(0, 1), #yscale = yscale,
                     name = paste("node_ETrees", nid, "plot",
                                  sep = ""))
    pushViewport(plot)

    xl <- 0.5 - width/4
    xr <- 0.5 + width/4



    ## box
    grid.rect(gp = gpar(fill = "#DDD8C2"))

    ## Text
    text<-paste0("N = ", ni[ind2,2] ,"\ny = ", ni[ind2,3],"\nh = ",ni[ind2,4])
    grid.text(label = text)

    ##refline
    # grid.lines(unit(c(0, 1), "npc"),
    #            unit(0, "native"), gp = gpar(col = col,lwd=unit(width/4,"npc"),lty="dashed"))
    # grid.lines(unit(c(xl, xr), "npc"),
    #            unit(d.CI$lower[ind2], "native"), gp = gpar(col = col,lwd=unit(width,"npc")))
    # grid.lines(unit(0.5, "npc"),
    #            unit(c(d.CI$lower[ind2],d.CI$upper[ind2]), "native"), gp = gpar(col = col,lwd=unit(width/2,"npc")))
    # meanline=FALSE ## if FALSE mean point
    # if(meanline){
    #   grid.lines(unit(c(0.5 - width/2, 0.5+width/2), "npc"),
    #              unit(d.CI$mean[ind2], "native"), gp = gpar(col = col, lwd = 2))
    # }else{      grid.points(unit(0.5, "npc"),
    #                         unit(d.CI$mean[ind2], "native"), gp = gpar(col = col,lwd=1),size=unit(1,"lines"),pch=20)}
    # grid.lines(unit(c(xl, xr), "npc"), unit(d.CI$upper[ind2], "native"),
    #            gp = gpar(col = col,lwd=unit(width,"npc")))
    #
    #grid.yaxis(label=TRUE) ## TO DO only TRUE for left terminal
    #grid.rect(gp = gpar(fill = "transparent"))
    upViewport(2)
  }
  return(rval)
}
class(terminal_ETrees) <- "grapcon_generator"

#' Pruning of a Exact Tree
#'
#' Determines the optimally pruned size of the tree by applying the one
#' standard error rule to the results from the bias-corrected bootstrap procedure.
#'
#' @param tree fitted tree of the class \code{ETree}.
#' @param pp pruning parameter, the constant (\eqn{c}) to be used in the \eqn{c*}standard
#'   error rule. The default value is 1.
#' @param \dots optional additional arguments.
#'
#' @details The one standard error rule for \code{ETrees} uses the estimates of the bias-corrected
#'   criterion value (\eqn{C}) and its standard error for each value of \eqn{L}
#'   (= maximum number of leaves). The optimally pruned tree corresponds to the
#'   smallest tree with a bias-corrected \eqn{C} higher or equal to the maximum
#'   bias-corrected \eqn{C} minus its standard error.
#'
#' @return Returns an object of class \code{ETree}. The number of leaves of this object is
#'   equal to the optimally pruned size of the tree.
#'
#'
#'
#'
#'
#' @keywords tree
#'
#' @importFrom rpart prune
#' @export
prune.ETree<-function(tree,pp=1,...){

    #pp
    mindex <- which.min(tree$CVOutput$Rel_error)  # find the row of the minimum x-error
    crit.minse <- tree$CVOutput$Rel_error[mindex] + pp*tree$CVOutput$SE[mindex]  # the minimum x-error + c*SE
    crit.row <- min(which(tree$CVOutput$Rel_error<= crit.minse))  # find the smallest tree within the minimum x-error + c*SE
    size <- tree$CVOutput$Size[crit.row]  # get the cp value for the smallest tree


    CVOutput <- tree$CVOutput
    CVOutput$Size <- CVOutput$Size[size]
    CVOutput$Error <- CVOutput$Error[size]
    CVOutput$Rel_error <- CVOutput$Rel_error[size]
    CVOutput$SE <- CVOutput$SE[size]
    CVOutput$SE2 <- CVOutput$SE2[size]
    CVOutput$CHECK <- CVOutput$CHECK[size]

    pruned_tree<-list(Xvarnames = tree$Xvarnames, Yvarnames = tree$Yvarnames,
                      Desc = tree$Desc, Evaluation = tree$Evaluation[size],
                      ProbLeaves = tree$ProbLeaves, Prob = tree$Prob,
                      Sizes = tree$Sizes, Index = tree$Index[[size]],
                      CVOutput = CVOutput, Transf_Trees = list(tree$Transf_Trees[[size]]),
                      TAll = tree$TAll[[size]], hAll = tree$hAll[[size]],
                      Tree = tree$Tree, h = tree$hAll[size], Data = tree$Data)

    class(pruned_tree) <- "ETree"

    return(pruned_tree)
}

MODULE OptimalTrees

USE Timer
USE OptTreeCallBacks

IMPLICIT NONE

! Definition of TreeTable:
! each row a node, columns:
!     (predictor, split, left node index, right node index, heterogeneity in node)
! nodes are indexed with their row positions
! for MaxDepth trees, nodes have an absolute index
! for MaxSize  trees, nodes have a relative index: the number of rows from the current row


INTEGER,PARAMETER :: &
   cESS      = 0, &
   cClassErr = 1, &
   cMinRisk  = 3

REAL(dp),PARAMETER :: &
   SomethingLarge = 2.0E20

REAL(dp),ALLOCATABLE,PRIVATE :: &
   YR(:,:), &
   Prior(:), &
   LossM(:)

INTEGER,ALLOCATABLE,PRIVATE :: &
   MSpace(:,:), &
   YI(:,:), &
   MaxYI(:), &
   ClassFreqY(:), &
   MaxS(:), &
   nNL(:), &
   Path(:,:), &
   PredType(:), &   ! 0=ordinal/continuous, 1=categorical
   NCatX(:)         ! number of categories per predictor (categorical only)

! Heap structure def for level=2
TYPE TPPHeap
    REAL(dp), POINTER :: Seg(:,:,:)
END TYPE TPPHeap
TYPE (TPPHeap), PRIVATE, ALLOCATABLE :: PP(:,:)


INTEGER,PRIVATE :: &
   GOutUnit     = 0, &
   NDep         = 0, &
   NPred        = 0, &
   NObs         = 0, &
   Measure      = 0, &
   MinNodeSize  = 0, &
   MaxLevel     = 0, &
   MaxILevel    = 0, &
   LookAheadDepth =0,&
   Depth        = 0, &
   MaxMaxYI     = 0, &
   NTotLCells   = 0, &
   NTotClass    = 0
!, &
!   CurP         = 0

LOGICAL,PRIVATE :: &
   CallBack = .FALSE., &
   Level2Heap = .FALSE.

INTEGER, PRIVATE :: &
   nS           = 0, &
   nN           = 0, &
   nST          = 0, &
   nR           = 0

REAL(dp),PRIVATE :: &
   MinH         = 0.d0, &
   BoundH       = 0.d0, &
   normMSS      = 0.d0, &
   normRisk     = 0.d0

CONTAINS



SUBROUTINE SetupLevel2Heap(OutUnit,X,NRows,NColsX,MaxSize,AlgoType)
! Allocate Heap structure for Level=2
! Level 3 HeapSize: n=# predictoren, c cats, MaxSize= 8 => MaxSize 4 left
! ((2*c)^3*(nchoosek(11,3))*4)/1024^2 +  ! 3 pred combi
! (n*c^2*4)/1024^2                    ! 3 x same pred
! (2*c*c^2*n*(n-1)*4)/1024^2          ! 2 x same * pred combi
! =>((2*c)^3*(nchoosek(11,3))*4)/1024^2 + (n*c^2*4)/1024^2 + (2*c*c^2*n*(n-1)*4)/1024^2
! x 4 bytes!
! Arguments
INTEGER OutUnit,NRows,NColsX,MaxSize,AlgoType
REAL(dp) X(NRows,NColsX)
! Locals
INTEGER i, j, CumStatus,Status,MemCount
! Begin routine


   Level2Heap = .FALSE.
   IF (MaxSize/=0 .AND. MaxLevel>=2 .AND. (iand(AlgoType,4)>0) .AND. (iand(AlgoType, 3)==0)) THEN
      IF (.NOT. ALLOCATED(MaxS)) ALLOCATE(MaxS(NColsX), STAT=CumStatus)
      MaxS=INT(MAXVAL(X, DIM=1))-1 !CHANGED

      !Heap structure for Level=2
      Status=1
      IF (.NOT. ALLOCATED(PP)) ALLOCATE(PP(NColsX,NColsX), STAT=Status)
      CumStatus=CumStatus+Status
      IF (CumStatus==0) THEN
         MemCount = (NColsX+1)*NColsX
         DO i=1,NColsX
            DO j=i+1,NColsX
               ALLOCATE(PP(i,j)%Seg(2*MaxS(i),2*MaxS(j),MaxSize-3), STAT=Status) !2:MaxSize-1
               MemCount = MemCount + 2*MaxS(i)*2*MaxS(j)*(MaxSize-3)
               CumStatus=CumStatus+Status
               IF (Status==0) THEN
                  PP(i,j)%Seg=-1.0
                  PP(j,i)%Seg=>PP(i,j)%Seg
               END IF
            END DO
            ALLOCATE(PP(i,i)%Seg(MaxS(i)+1,MaxS(i)+1,MaxSize-3), STAT=Status) !2:MaxSize-1
            MemCount = MemCount + (MaxS(i)+1)*(MaxS(i)+1)*(MaxSize-3)
            CumStatus=CumStatus+Status
            IF (Status==0) PP(i,i)%Seg=-1.0
         END DO
         IF (CumStatus==0) Level2Heap = .TRUE.
         CALL WriteHeap2Allocation(OutUnit,CumStatus,MemCount)
      END IF
   END IF

END SUBROUTINE SetupLevel2Heap



SUBROUTINE DeleteLevel2Heap
! Locals
INTEGER i, j
! Begin routine

   IF (ALLOCATED(PP)) THEN
      DO i=1,NPred
         !CALL mexPrintFF(OutUnit,CHAR(10))
         !CALL mexWriteR4M(OutUnit,PP(i,i)%Seg,MaxS(i)+1,MaxS(i)+1)
         IF (ASSOCIATED(PP(i,i)%Seg)) DEALLOCATE(PP(i,i)%Seg)
         DO j=i+1,NPred
            !CALL mexPrintFF(OutUnit,CHAR(10))
            !CALL mexWriteR4M(OutUnit,PP(i,j)%Seg,2*MaxS(i),2*MaxS(j))
            NULLIFY(PP(j,i)%Seg)
            IF (ASSOCIATED(PP(i,j)%Seg)) DEALLOCATE(PP(i,j)%Seg)
         END DO
      END DO
      DEALLOCATE (PP)
   END IF
   IF (ALLOCATED(MaxS)) DEALLOCATE (MaxS)

END SUBROUTINE DeleteLevel2Heap





SUBROUTINE ComputeMaxTree (OutUnit,NRows, NColsY, Y, NColsX, X, XType, AMeasure, MaxSize, AMaxDepth, &
                           AMinNodeSize, AMinH, AlgoType, ABoundH, ALookAheadDepth, APrior, ALossM, &
                           Heterogeneity, NTrees, Tree, TreeTableSize, warn)
! Main routine that calls different variants of the recursive tree routines
! Setsup all private, module global variables to avoid argument passing overhead
! Arguments
INTEGER OutUnit,NRows,NColsY,NColsX,NTrees,TreeTableSize, warn
REAL(dp) Y(NRows,NColsY), X(NRows,NColsX)
INTEGER XType(NColsX)   ! 0=ordinal/continuous, 1=categorical
INTEGER AMeasure,MaxSize,AMaxDepth,AMinNodeSize,AlgoType,ALookAheadDepth
REAL(dp) AMinH,ABoundH,APrior(*),ALossM(*)
REAL(dp) Heterogeneity(NTrees), Tree(NTrees,TreeTableSize)
! Locals
INTEGER ObsIndex(NRows), i, TreeSize(NTrees)
!j,Pred(NColsY)
REAL(dp)  Time
LOGICAL, PARAMETER :: All = .TRUE.
! Begin routine


! Set Global Module Variables

   GOutUnit     = OutUnit
   NDep         = NColsY
   NPred        = NColsX
   NObs         = NRows
   Measure      = iand(AMeasure,127)
   CallBack     = iand(AMeasure,128)>0
   MinNodeSize  = AMinNodeSize
   MaxLevel     = AMaxDepth-1
   MaxILevel    = MAX(0,MaxLevel-3)
   LookAheadDepth = ALookAheadDepth
   MinH         = AMinH
   BoundH       = ABoundH
   Depth        = 0
   nS           = 0
   nN           = 0
   IF (.NOT. ALLOCATED(nNL)) ALLOCATE(nNL(0:MaxLevel))
   nNL          = 0
   nST          = 0
   nR           = 0
   IF (.NOT. ALLOCATED(Path)) ALLOCATE(Path(MaxLevel+1,2))
   Path = 0

   CALL SetupLevel2Heap(OutUnit,X,NRows,NColsX,MaxSize,AlgoType)

   CALL SetupCallBacks(NRows)

   ! Set Locals
   ! initialize output in case they are not defined
   Tree         = -1
   TreeSize     = 2**(1+MaxLevel)-1
   Heterogeneity= 0

   ObsIndex     = (/(i,i=1,NRows)/)

   !If the dll was aborted, variables may still be allocated
   IF (.NOT. ALLOCATED(MSpace)) ALLOCATE (MSpace(NRows,NColsX))
   MSpace = INT(X) !CHANGED

   ! Setup predictor type globals
   IF (.NOT. ALLOCATED(PredType)) ALLOCATE(PredType(NColsX))
   IF (.NOT. ALLOCATED(NCatX))    ALLOCATE(NCatX(NColsX))
   PredType = XType
   NCatX    = 0
   DO i=1,NColsX
      IF (XType(i) == 1) NCatX(i) = INT(MAXVAL(X(:,i)))
   END DO

   ! Setup Measure type globals
   IF (Measure == cESS) THEN
      IF (.NOT. ALLOCATED(YR)) ALLOCATE(YR(NRows,NColsY))
      YR = Y
      normMSS = MinH * GetHeterogeneity(NRows,ObsIndex)/NRows
   ELSE IF (Measure == cClassErr) THEN
      IF (.NOT. ALLOCATED(YI)) ALLOCATE(YI(NRows,NColsY))
      IF (.NOT. ALLOCATED(MaxYI)) ALLOCATE(MaxYI(NColsY))
      YI = INT(Y)
      MaxYI = MAXVAL(YI)
      MaxMaxYI = MAXVAL(MaxYI)
   ELSE IF (Measure == cMinRisk) THEN
      IF (.NOT. ALLOCATED(YI)) ALLOCATE(YI(NRows,NColsY))
      IF (.NOT. ALLOCATED(MaxYI)) ALLOCATE(MaxYI(NColsY))
      YI = INT(Y)
      MaxYI = MAXVAL(YI)
      MaxMaxYI = MAXVAL(MaxYI)
      NTotClass = SUM(MaxYI)
      NTotLCells = SUM(MaxYI**2)
      IF (.NOT. ALLOCATED(Prior)) ALLOCATE(Prior(NTotClass))
      IF (.NOT. ALLOCATED(LossM)) ALLOCATE(LossM(NTotLCells))
      IF (.NOT. ALLOCATED(ClassFreqY)) ALLOCATE(ClassFreqY(NTotClass))
      Prior(1:NTotClass) = APrior(1:NTotClass)
      LossM(1:NTotLCells) = ALossM(1:NTotLCells)
      CALL GetClassFreq(NRows,ObsIndex,ClassFreqY)
      normRisk = minH * GetHeterogeneity(NRows,ObsIndex)
   ELSE
      normMSS = MinH * GetHeterogeneity(NRows,ObsIndex)/NRows
   END IF



   CALL SetTimer(Time) ! Dummy argument here
   ! Choose and execute the correct algorithm variant
   IF (MaxSize == 0) THEN
      SELECT CASE (iand(AlgoType,3))
      CASE (0)
         IF (iand(AlgoType,8)>0) THEN
            CALL OptMaxDepthTreeSortP (0, NRows, ObsIndex, Heterogeneity(1))
         ELSE IF (iand(AlgoType,4)>0) THEN
            CALL OptMaxDepthTreeAll (0, NRows, ObsIndex, Heterogeneity)
         ELSE
            CALL OptMaxDepthTree (0, NRows, ObsIndex, Heterogeneity(1))
         END IF
      CASE (1)
         IF (iand(AlgoType,16)>0) THEN
            CALL OptMaxDepthTreeT (0, NRows, ObsIndex, Heterogeneity(1), Tree, 0)
            !CALL OptMaxDepthTreeTLAH (0, NRows, ObsIndex, Heterogeneity(1), Tree, 0)
         ELSE IF (iand(AlgoType,8)>0) THEN
            CALL OptMaxDepthTreeTSortP (0, NRows, ObsIndex, Heterogeneity(1), Tree, 0)
         ELSE IF (iand(AlgoType,4)>0) THEN
            CALL OptMaxDepthTreeTAll (0, NRows, ObsIndex, Heterogeneity, Tree, 0)
         ELSE
            CALL OptMaxDepthTreeT (0, NRows, ObsIndex, Heterogeneity(1), Tree, 0)
         END IF
      CASE (2)
         warn = 1 !CALL Warning('Branch and bound not implemented for MaxDepth tree, default used.')
         CALL OptMaxDepthTree (0, NRows, ObsIndex, Heterogeneity(1))
      END SELECT
   ELSE
      IF (AMaxDepth>MaxSize) MaxLevel = MaxSize-1
      SELECT CASE (iand(AlgoType,3))
      CASE (0)
         IF (iand(AlgoType,8)>0) THEN
            warn=2 !CALL Warning('predictor sorting not implemented for MaxSize tree.')
         ELSE IF (iand(AlgoType,4)>0) THEN
            IF (Level2Heap) THEN
               CALL OptMaxSizeTreeAllTest (0, NRows, ObsIndex, MaxSize, Heterogeneity)
            ELSE
               CALL OptMaxSizeTreeAll (0, NRows, ObsIndex, MaxSize, Heterogeneity)
            END IF
         ELSE
            CALL OptMaxSizeTree (0, NRows, ObsIndex, MaxSize, Heterogeneity(1))
         END IF
      CASE (1)
         IF (iand(AlgoType,8)>0) THEN
            warn = 3!CALL Warning('predictor sorting not implemented for MaxSize tree.')
         ELSE IF (iand(AlgoType, 4)>0) THEN
            CALL OptMaxSizeTreeTAll (0, NRows, ObsIndex, MaxSize, &
                                     Heterogeneity, Tree, TreeSize)
         ELSE
            CALL OptMaxSizeTreeT (0, NRows, ObsIndex, MaxSize, &
                                  Heterogeneity(1), Tree, TreeSize(1))
         END IF
      CASE (2)
         CALL OptMaxSizeTreeBB (0, NRows, ObsIndex, MaxSize, Heterogeneity(1))
      END SELECT
   END IF

   CALL GetTimer(Time) ! Dummy argument here
   !CALL OutputAlgorithmDetails (OutUnit,Heterogeneity, NTrees, Depth, nS, nN, nNL, nST)
   CALL OutputAlgorithmDetailsX (OutUnit,Heterogeneity, NTrees, Depth, nS, nN, nNL, nST, nR)

   DEALLOCATE (MSpace)
   DEALLOCATE (PredType, NCatX)
   CALL DeleteLevel2Heap
   DEALLOCATE (Path)
   DEALLOCATE (nNL)

   IF (Measure == cESS) THEN
      DEALLOCATE(YR)
   ELSE IF (Measure == cClassErr) THEN
      DEALLOCATE(YI,MaxYI)
   ELSE IF (Measure == cMinRisk) THEN
      DEALLOCATE(YI,MaxYI)
      DEALLOCATE(Prior,LossM,ClassFreqY)
   END IF

   CALL CloseCallBacks(NRows)

END SUBROUTINE ComputeMaxTree





SUBROUTINE TransposeCopyMaxSizeTreeX(Size,TreeTable,Tree)
! Arguments
INTEGER Size
REAL(dp) TreeTable(Size,5),Tree(5,Size)
! Begin routine

TreeTable=TRANSPOSE(Tree)

END SUBROUTINE TransposeCopyMaxSizeTreeX





RECURSIVE SUBROUTINE OptMaxDepthTree (Level, N, ObsIndex, h)
! Find optimal tree with maximal depth restrictions only (no size restriction).
! Return ONLY optimal objective function value.
! Arguments
INTEGER Level, N
INTEGER ObsIndex(N)
REAL(dp), INTENT(OUT) :: h
! Locals
INTEGER MinMSpace(NPred), MaxMSpace(NPred), Dif
INTEGER p, s, N1, N2, ObsIndex1(N), ObsIndex2(N)
REAL(dp)  h1, h2, hs
! Begin routine

nN = nN +1
CALL GetMinMaxMSpace(N, ObsIndex, MinMSpace, MaxMSpace)
Dif = SUM(MaxMSpace - MinMSpace)
h = GetHeterogeneity (N,ObsIndex)

IF ((Dif == 0) .OR. (level == MaxLevel) .OR. (N <= MinNodeSize) .OR. VeryHomogeneous(h,N)) THEN
   Depth = MAX(Depth,level+1)
   IF (N < MinNodeSize) h = SomethingLarge
ELSE
   DO p=1,NPred ! over all predictors
      !IF (Level == 0) CALL OutputProgressP (p)
      DO s=MinMSpace(p),MaxMSpace(p)-1 ! and all possible splits
         IF (Level == 0) CALL OutputProgress (p,s)
         IF ((Level <= MaxILevel) .AND. (IsInterrupted())) RETURN

         ! Excute Split
         IF (GetValidSplitAny(s, p, N, ObsIndex, N1, ObsIndex1, N2, ObsIndex2)) THEN
            nS = nS + 1
            IF (N2>N1) THEN !Reverse N2>N1= more nodes, less work
               CALL OptMaxDepthTree (Level+1, N1, ObsIndex1, h1)
               !IF (h1<h) THEN
                  CALL OptMaxDepthTree (Level+1, N2, ObsIndex2, h2)
                  IF (h1+h2<h) THEN ! Better split found
                     h = h1 + h2
                  END IF
               !END IF
            ELSE
               CALL OptMaxDepthTree (Level+1, N2, ObsIndex2, h2)
               !IF (h2<h) THEN
                  CALL OptMaxDepthTree (Level+1, N1, ObsIndex1, h1)
                  IF (h1+h2<h) THEN ! Better split found
                     h = h1 + h2
                  END IF
               !END IF
            END IF
            IF (Level == 0) THEN
               hs=h1+h2
               if (hs>BoundH) hs=BoundH
            END IF
         END IF
      END DO
   END DO
END IF

END SUBROUTINE OptMaxDepthTree




RECURSIVE SUBROUTINE OptMaxDepthTreeSortP (Level, N, ObsIndex, h)
! This version uses sorted predictors to find splits.
! So far, this a little bit slower, due to the sorting neccesary.
! Presumably, for continuous predictors, this is faster,
! because sorting is needed only once, and many splits occur for these predictors.
! Find optimal tree with maximal depth restrictions only (no size restriction).
! Return ONLY optimal objective function value.
! Arguments
INTEGER Level, N
INTEGER ObsIndex(N)
REAL(dp), INTENT(OUT) :: h
! Locals
INTEGER MinMSpace(NPred), MaxMSpace(NPred), Dif
INTEGER p, s, N1, N2, ObsIndex1(N), ObsIndex2(N)
REAL(dp)  h1, h2
!integer i
! Begin routine

nN = nN +1
CALL GetMinMaxMSpace(N, ObsIndex, MinMSpace, MaxMSpace)
Dif = SUM(MaxMSpace - MinMSpace)
h = GetHeterogeneity (N,ObsIndex)

IF ((Dif == 0) .OR. (level == MaxLevel) .OR. (N <= MinNodeSize) .OR. VeryHomogeneous(h,N)) THEN
   Depth = MAX(Depth,level+1)
   IF (N < MinNodeSize) h = SomethingLarge
ELSE
   DO p=1,NPred ! over all predictors
      !IF (Level == 0) CALL OutputProgressP (p)
      CALL SetupValidNextSplit(s,p,N,ObsIndex,N1,ObsIndex1,N2,ObsIndex2)
      DO WHILE (GetValidNextSplit(s, p, N, N1, N2, ObsIndex1))
         IF (Level == 0) CALL OutputProgress (p,s)
         IF ((Level <= MaxILevel) .AND. (IsInterrupted())) RETURN

         ! Excute Split
         nS = nS + 1
         IF (N2>N1) THEN !Reverse N2>N1= more nodes, less work
            CALL OptMaxDepthTreeSortP (Level+1, N1, ObsIndex1, h1)
            IF (h1<h) THEN
               CALL OptMaxDepthTreeSortP (Level+1, N2, ObsIndex2, h2)
               IF (h1+h2<h) THEN ! Better split found
                  h = h1 + h2
               END IF
            END IF
         ELSE
            CALL OptMaxDepthTreeSortP (Level+1, N2, ObsIndex2, h2)
            IF (h2<h) THEN
               CALL OptMaxDepthTreeSortP (Level+1, N1, ObsIndex1, h1)
               IF (h1+h2<h) THEN ! Better split found
                  h = h1 + h2
               END IF
            END IF
         END IF
      END DO
   END DO

END IF

END SUBROUTINE OptMaxDepthTreeSortP




RECURSIVE SUBROUTINE OptMaxDepthTreeT (Level, N, ObsIndex, h, Tree, LastNode)
! Find optimal tree with maximal depth restrictions only (no size restriction).
! Return both optimal objective function value AND optimal tree structure.
! Arguments
INTEGER Level, N
INTEGER ObsIndex(N)
INTEGER, INTENT(IN) :: LastNode
REAL(dp), INTENT(OUT) :: h, Tree(5,2**(1+MaxLevel-Level)-1)
! Locals
INTEGER MinMSpace(NPred), MaxMSpace(NPred), Dif
INTEGER p, s, N1, N2, ObsIndex1(N), ObsIndex2(N), ts, ts12
REAL(dp)  h1, h2
REAL(dp) Tree1(5,2**(MaxLevel-Level)-1),Tree2(5,2**(MaxLevel-Level)-1)
! Begin routine

nN = nN +1
ts=2**(1+MaxLevel-Level)-1
ts12=2**(MaxLevel-Level)-1

CALL GetMinMaxMSpace(N, ObsIndex, MinMSpace, MaxMSpace)
Dif = SUM(MaxMSpace - MinMSpace)
h = GetHeterogeneity (N,ObsIndex)
!in case none of the possible splits turn out to be successfull
!define endnode
Tree(1:5,1) = (/0.d0, 0.d0, 0.d0, 0.d0, h /)
Tree(1,2:ts) = -1 ! Signal unused nodes

IF ((Dif == 0) .OR. (level == MaxLevel) .OR. (N <= MinNodeSize) .OR. VeryHomogeneous(h,N)) THEN
   Depth = MAX(Depth,level+1)
   IF (N < MinNodeSize) THEN
      h = SomethingLarge
      Tree(5,1) = h
   END IF
ELSE
   DO p=1,NPred ! over all predictors
      !IF (Level == 0) CALL OutputProgressP (p)
      DO s=MinMSpace(p),MaxMSpace(p)-1 ! and all possible splits
         IF (Level == 0) CALL OutputProgress (p,s)
         IF ((Level <= MaxILevel) .AND. (IsInterrupted())) RETURN

         ! Excute Split
         IF (GetValidSplitAny(s, p, N, ObsIndex, N1, ObsIndex1, N2, ObsIndex2)) THEN
            nS = nS + 1
            IF (N2>N1) THEN !Reverse N2>N1= turns out to be less work
               CALL OptMaxDepthTreeT (Level+1, N1, ObsIndex1, h1, Tree1,LastNode+1)
            ELSE
               CALL OptMaxDepthTreeT (Level+1, N2, ObsIndex2, h1, Tree2,LastNode+1+ts12)
            END IF
            IF (h1<h) THEN
               IF (N2>N1) THEN
                  CALL OptMaxDepthTreeT (Level+1, N2, ObsIndex2, h2, Tree2, LastNode+1+ts12)
               ELSE
                  CALL OptMaxDepthTreeT (Level+1, N1, ObsIndex1, h2, Tree1, LastNode+1)
               END IF
               IF (h1+h2<h) THEN ! Better split found
                  h = h1 + h2
                  Tree(1:4,1) = (/p, s, 1, 1+ts12/)
                  Tree(5,1) =  h
                  Tree(1:5,2:1+ts12) = Tree1(1:5,1:ts12)
                  Tree(1:5,2+ts12:ts) = Tree2(1:5,1:ts12)
               END IF
            END IF
         END IF
      END DO
   END DO
END IF

END SUBROUTINE OptMaxDepthTreeT




RECURSIVE SUBROUTINE OptMaxDepthTreeTSortP (Level, N, ObsIndex, h, Tree, LastNode)
! This version uses sorted predictors to find splits.
! So far, this is a little bit slower, due to the sorting neccesary.
! Presumably, for continuous predictors, this is faster,
! because sorting is needed only once, and many splits occur for these predictors.
! Find optimal tree with maximal depth restrictions only (no size restriction).
! Return both optimal objective function value AND optimal tree structure.
! Arguments
INTEGER Level, N
INTEGER ObsIndex(N)
INTEGER, INTENT(IN) :: LastNode
REAL(dp), INTENT(OUT) :: h, Tree(5,2**(1+MaxLevel-Level)-1)
! Locals
INTEGER MinMSpace(NPred), MaxMSpace(NPred), Dif
INTEGER p, s, N1, N2, ObsIndex1(N), ObsIndex2(N), ts, ts12
REAL(dp)  h1, h2
REAL(dp) Tree1(5,2**(MaxLevel-Level)-1),Tree2(5,2**(MaxLevel-Level)-1)
! Begin routine

nN = nN +1
ts=2**(1+MaxLevel-Level)-1
ts12=2**(MaxLevel-Level)-1

CALL GetMinMaxMSpace(N, ObsIndex, MinMSpace, MaxMSpace)
Dif = SUM(MaxMSpace - MinMSpace)
h = GetHeterogeneity (N,ObsIndex)
!in case none of the possible splits turn out to be successfull
!define endnode
Tree(1:5,1) = (/0.d0, 0.d0, 0.d0, 0.d0, h /)
Tree(1,2:ts) = -1 ! Signal unused nodes

IF ((Dif == 0) .OR. (level == MaxLevel) .OR. (N <= MinNodeSize) .OR. VeryHomogeneous(h,N)) THEN
   Depth = MAX(Depth,level+1)
   IF (N < MinNodeSize)  THEN
      h = SomethingLarge
      Tree(5,1) = h
   END IF
ELSE

   DO p=1,NPred ! over all predictors
      !IF (Level == 0) CALL OutputProgressP (p)
      CALL SetupValidNextSplit(s,p,N,ObsIndex,N1,ObsIndex1,N2,ObsIndex2)
      DO WHILE (GetValidNextSplit(s, p, N, N1, N2, ObsIndex1))
         IF (Level == 0) CALL OutputProgress (p,s)
         IF ((Level <= MaxILevel) .AND. (IsInterrupted())) RETURN

         ! Excute Split
         nS = nS + 1
         IF (N2>N1) THEN !Reverse N2>N1= turns out to be less work
            CALL OptMaxDepthTreeTSortP (Level+1, N1, ObsIndex1, h1, Tree1,LastNode+1)
         ELSE
            CALL OptMaxDepthTreeTSortP (Level+1, N2, ObsIndex2, h1, Tree2,LastNode+1+ts12)
         END IF
         IF (h1<h) THEN
            IF (N2>N1) THEN
               CALL OptMaxDepthTreeTSortP (Level+1, N2, ObsIndex2, h2, Tree2, LastNode+1+ts12)
            ELSE
               CALL OptMaxDepthTreeTSortP (Level+1, N1, ObsIndex1, h2, Tree1, LastNode+1)
            END IF
            IF (h1+h2<h) THEN ! Better split found
               h = h1 + h2
               Tree(1:4,1) = (/p, s, 1, 1+ts12/)
               Tree(5,1) =  h
               Tree(1:5,2:1+ts12) = Tree1(1:5,1:ts12)
               Tree(1:5,2+ts12:ts) = Tree2(1:5,1:ts12)
            END IF
         END IF
      END DO
   END DO
END IF

END SUBROUTINE OptMaxDepthTreeTSortP



RECURSIVE SUBROUTINE OptMaxDepthTreeAll (Level, N, ObsIndex, h)
! Find optimal tree with maximal depth restrictions only (no size restriction).
! Return ONLY optimal objective function value.
! Arguments
INTEGER Level, N
INTEGER ObsIndex(N)
REAL(dp), INTENT(OUT) :: h(1+MaxLevel-Level)
! Locals
INTEGER MinMSpace(NPred), MaxMSpace(NPred), Dif
INTEGER p, s, Size, N1, N2, ObsIndex1(N), ObsIndex2(N)
REAL(dp)  h1(MaxLevel-Level), h2(MaxLevel-Level)
! Begin routine

nN = nN +1
CALL GetMinMaxMSpace(N, ObsIndex, MinMSpace, MaxMSpace)
Dif = SUM(MaxMSpace - MinMSpace)
h = GetHeterogeneity (N,ObsIndex) !<=all

IF ((Dif == 0) .OR. (level == MaxLevel) .OR. (N <= MinNodeSize) .OR. VeryHomogeneous(h(1),N)) THEN
   Depth = MAX(Depth,level+1)
   IF (N < MinNodeSize) h = SomethingLarge
ELSE
   DO p=1,NPred ! over all predictors
      !IF (Level == 0) CALL OutputProgressP (p)
      DO s=MinMSpace(p),MaxMSpace(p)-1 ! and all possible splits
         IF (Level == 0) CALL OutputProgress (p,s)
         IF ((Level <= MaxILevel) .AND. (IsInterrupted())) RETURN

         ! Excute Split
         IF (GetValidSplitAny(s, p, N, ObsIndex, N1, ObsIndex1, N2, ObsIndex2)) THEN
            nS = nS + 1
            CALL OptMaxDepthTreeAll (Level+1, N1, ObsIndex1, h1)
            CALL OptMaxDepthTreeAll (Level+1, N2, ObsIndex2, h2)
            DO Size=2,1+MaxLevel-Level
               IF (h1(Size-1)+h2(Size-1)<h(Size)) THEN ! Better split found
                  h(Size) = h1(Size-1) + h2(Size-1)
               END IF
            END DO
         END IF
      END DO
   END DO

   ! Propagate best values for smaller trees
   DO Size=2,1+MaxLevel-Level
      h(Size) = MINVAL(h(1:Size))
   END DO

END IF

END SUBROUTINE OptMaxDepthTreeAll




RECURSIVE SUBROUTINE OptMaxDepthTreeTAll (Level, N, ObsIndex, h, Tree, LastNode)
! Find optimal tree with maximal depth restrictions only (no size restriction).
! Return ONLY optimal objective function value.
! Arguments
INTEGER Level, N, LastNode
INTEGER ObsIndex(N)
REAL(dp), INTENT(OUT) :: h(1+MaxLevel-Level)
REAL(dp), INTENT(OUT) :: Tree(1+MaxLevel-Level,5,2**(1+MaxLevel-Level)-1)
! Locals
INTEGER MinMSpace(NPred), MaxMSpace(NPred), Dif
INTEGER p, s, Size, N1, N2, ObsIndex1(N), ObsIndex2(N)
INTEGER ts, ts12, NTrees
REAL(dp)  h1(MaxLevel-Level), h2(MaxLevel-Level)
REAL(dp) Tree1(MaxLevel-Level,5,2**(MaxLevel-Level)-1), &
       Tree2(MaxLevel-Level,5,2**(MaxLevel-Level)-1)
! Begin routine

nN = nN +1
NTrees = 1+MaxLevel-Level
ts=2**(1+MaxLevel-Level)-1
ts12=2**(MaxLevel-Level)-1


CALL GetMinMaxMSpace(N, ObsIndex, MinMSpace, MaxMSpace)
Dif = SUM(MaxMSpace - MinMSpace)
h = GetHeterogeneity (N,ObsIndex) !<=all
!in case none of the possible splits turn out to be successfull
!define endnode
Tree(1:NTrees,5,1) = h
Tree(1:NTrees,1:4,1) = 0.d0
Tree(1:NTrees,1,2:ts) = -1 ! Signal unused nodes

IF ((Dif == 0) .OR. (level == MaxLevel) .OR. (N <= MinNodeSize) .OR. VeryHomogeneous(h(1),N)) THEN
   Depth = MAX(Depth,level+1)
   IF (N < MinNodeSize) h = SomethingLarge
   Tree(1:NTrees,5,1) = h
ELSE
   DO p=1,NPred ! over all predictors
      !IF (Level == 0) CALL OutputProgressP (p)
      DO s=MinMSpace(p),MaxMSpace(p)-1 ! and all possible splits
         IF (Level == 0) CALL OutputProgress (p,s)
         IF ((Level <= MaxILevel) .AND. (IsInterrupted())) RETURN

         ! Excute Split
         IF (GetValidSplitAny(s, p, N, ObsIndex, N1, ObsIndex1, N2, ObsIndex2)) THEN
            nS = nS + 1
            CALL OptMaxDepthTreeTAll (Level+1, N1, ObsIndex1, h1, Tree1, LastNode+1)
            CALL OptMaxDepthTreeTAll (Level+1, N2, ObsIndex2, h2, Tree2, LastNode+1+ts12)
            DO Size=2,1+MaxLevel-Level
               IF (h1(Size-1)+h2(Size-1)<h(Size)) THEN ! Better split found
                  ts=2**Size-1
                  ts12=2**(Size-1)-1
                  h(Size) = h1(Size-1) + h2(Size-1)
                  Tree(Size,1:4,1) = (/p, s, 1, 1+ts12/)
                  Tree(Size,5,1) =  h(Size)
                  Tree(Size,1:5,2:1+ts12) = Tree1(Size-1,1:5,1:ts12)
                  Tree(Size,1:5,2+ts12:ts) = Tree2(Size-1,1:5,1:ts12)
               END IF
            END DO
         END IF
      END DO
   END DO

   ! Propagate best values for smaller trees
   DO Size=2,1+MaxLevel-Level
      IF (h(Size-1) <= h(Size)) THEN
         h(Size) = h(Size-1)
         Tree(Size,1:5,1:ts) = Tree(Size-1,1:5,1:ts)
      END IF
   END DO

END IF

END SUBROUTINE OptMaxDepthTreeTAll



RECURSIVE SUBROUTINE OptMaxSizeTree (Level, N, ObsIndex, MaxSize, h)
! Find optimal tree with maximal depth AND maximum size restrictions.
! Return ONLY optimal objective function value.
! Arguments
INTEGER Level, N
INTEGER ObsIndex(N)
INTEGER, INTENT(IN) :: MaxSize
REAL(dp), INTENT(OUT) :: h
! Locals
INTEGER MinMSpace(NPred), MaxMSpace(NPred), Dif
INTEGER p, s, N1, N2, ObsIndex1(N), ObsIndex2(N)
INTEGER SubMaxSize, MaxSubMaxSize, MaxMaxSize
REAL(dp)  h1, h2
LOGICAL SplitFound
! Begin routine

nN = nN +1

CALL GetMinMaxMSpace(N, ObsIndex, MinMSpace, MaxMSpace)
Dif = SUM(MaxMSpace - MinMSpace)
h = GetHeterogeneity (N,ObsIndex)

IF ((MaxSize <= 1) .OR. &
    (Dif == 0) .OR. (level == MaxLevel) .OR. (N <= MinNodeSize) .OR. &
    VeryHomogeneous(h,N)) THEN
   Depth = MAX(Depth,level+1)
   IF (N < MinNodeSize) h = SomethingLarge
ELSE
   SplitFound = .FALSE.
   IF (Level == 0) h = BoundH
   MaxSubMaxSize = MIN(MaxSize-1,MaxNLeaves(MaxLevel-Level))
   MaxMaxSize = MIN(MaxSize,MaxNLeaves(1+MaxLevel-Level))

   DO p=1,NPred ! over all predictors
      !IF (Level == 0) CALL OutputProgressP (p)
      DO s=MinMSpace(p),MaxMSpace(p)-1 ! and all possible splits
         IF (Level == 0) CALL OutputProgress (p,s)
         IF ((Level == 0) .AND. (IsInterrupted())) RETURN

         ! Excute Split
         IF (GetValidSplitAny(s, p, N, ObsIndex, N1, ObsIndex1, N2, ObsIndex2)) THEN
            nS = nS + 1
            DO SubMaxSize=MaxSubMaxSize,MaxMaxSize-MaxSubMaxSize,-1
               IF (N1>N2) THEN
                  CALL OptMaxSizeTree (Level+1, N1, ObsIndex1, SubMaxSize, &
                                         h1)
               ELSE
                  CALL OptMaxSizeTree (Level+1, N2, ObsIndex2, SubMaxSize, &
                                         h1)
               END IF
               IF (h1<h) THEN
                  nST = nST +1
                  IF (N1>N2) THEN
                     CALL OptMaxSizeTree (Level+1, N2, ObsIndex2, MaxSize-SubMaxSize, &
                                           h2)
                  ELSE
                     CALL OptMaxSizeTree (Level+1, N1, ObsIndex1, MaxSize-SubMaxSize, &
                                           h2)
                  END IF
                  IF (h1+h2<h) THEN ! Better split found
                     SplitFound = .TRUE.
                     h = h1 + h2
                  END IF
               ELSE
                  EXIT
               END IF
            END DO
         END IF
      END DO
   END DO
   IF ((level==0) .AND. .NOT. SplitFound) h = GetHeterogeneity (N,ObsIndex)
END IF

END SUBROUTINE OptMaxSizeTree




RECURSIVE SUBROUTINE OptMaxSizeTreeT (Level, N, ObsIndex, MaxSize, h, Tree, TreeSize)
! Find optimal tree with maximal depth AND maximum size restrictions.
! Return BOTH optimal objective function value AND optimal tree structure(s)
! Arguments
INTEGER Level, N
INTEGER ObsIndex(N)
INTEGER, INTENT(IN) :: MaxSize
REAL(dp), INTENT(OUT) :: h, Tree(5,1+2*(MaxSize-1))
INTEGER, INTENT(OUT) :: TreeSize
! Locals
INTEGER MinMSpace(NPred), MaxMSpace(NPred), Dif
INTEGER p, s, N1, N2, ObsIndex1(N), ObsIndex2(N)
INTEGER SubMaxSize, MaxSubMaxSize, MaxMaxSize, Tree1Size, Tree2Size
REAL(dp)  h1, h2
REAL(dp) Tree1(5,1+2*(MaxSize-2)),Tree2(5,1+2*(MaxSize-2))
LOGICAL SplitFound
!,IsI
! Begin routine

nN = nN +1

CALL GetMinMaxMSpace(N, ObsIndex, MinMSpace, MaxMSpace)
Dif = SUM(MaxMSpace - MinMSpace)
h = GetHeterogeneity (N,ObsIndex)
! Node needs to be defined allways in case none of the splits improves
Tree(1:5,1) = (/0.d0, 0.d0, 0.d0, 0.d0, h /)
Tree(1,2:1+2*(MaxSize-1)) = -1 ! Signal unused nodes
TreeSize = 1

IF ((MaxSize <= 1) .OR. &
    (Dif == 0) .OR. (level == MaxLevel) .OR. (N <= MinNodeSize) .OR. &
    VeryHomogeneous(h,N)) THEN
   Depth = MAX(Depth,level+1)
   IF (N < MinNodeSize) THEN
      h = SomethingLarge
      Tree(5,1) = h
   END IF
ELSE
   SplitFound = .FALSE.
   IF (Level == 0) h = BoundH
   MaxSubMaxSize = MIN(MaxSize-1,MaxNLeaves(MaxLevel-Level))
   MaxMaxSize = MIN(MaxSize,MaxNLeaves(1+MaxLevel-Level))

   DO p=1,NPred ! over all predictors
      !IF (Level == 0) CALL OutputProgressP (p)
      DO s=MinMSpace(p),MaxMSpace(p)-1 ! and all possible splits
         IF (Level == 0) CALL OutputProgress (p,s)
         IF ((Level <= MaxILevel) .AND. (IsInterrupted())) RETURN

         ! Excute Split
         IF (GetValidSplitAny(s, p, N, ObsIndex, N1, ObsIndex1, N2, ObsIndex2)) THEN
            nS = nS + 1
            DO SubMaxSize=MaxSubMaxSize,MaxMaxSize-MaxSubMaxSize,-1
            !DO SubMaxSize=MaxSize-1,1,-1
               IF (N1>N2) THEN
                  CALL OptMaxSizeTreeT (Level+1, N1, ObsIndex1, SubMaxSize, &
                                         h1, Tree1,Tree1Size)
               ELSE
                  CALL OptMaxSizeTreeT (Level+1, N2, ObsIndex2, SubMaxSize, &
                                         h1, Tree2,Tree2Size)
               END IF
               IF (h1<h) THEN
                  nST = nST +1
                  IF (N1>N2) THEN
                     CALL OptMaxSizeTreeT (Level+1, N2, ObsIndex2, MaxSize-SubMaxSize, &
                                           h2, Tree2, Tree2Size)
                  ELSE
                     CALL OptMaxSizeTreeT (Level+1, N1, ObsIndex1, MaxSize-SubMaxSize, &
                                           h2, Tree1, Tree1Size)
                  END IF
                  IF (h1+h2<h) THEN ! Better split found
                     SplitFound = .TRUE.
                     h = h1 + h2
                     TreeSize=1+Tree1Size+Tree2Size
                     Tree(1:4,1) = (/p, s, 1, 1+Tree1Size/)
                     Tree(5,1) =  h
                     Tree(1:5,2:1+Tree1Size) = Tree1(1:5,1:Tree1Size)
                     Tree(1:5,2+Tree1Size:TreeSize) = Tree2(1:5,1:Tree2Size)
                  END IF
               ELSE
                  EXIT
               END IF
            END DO
         END IF
      END DO
   END DO
   Tree(1,TreeSize+1:1+2*(MaxSize-1)) = -1 ! Signal unused nodes of worse but larger previously trees
   IF ((level==0) .AND. .NOT. SplitFound) h = GetHeterogeneity (N,ObsIndex)
END IF

END SUBROUTINE OptMaxSizeTreeT




RECURSIVE SUBROUTINE GetOptMaxDepthTreeBound (Level, N, ObsIndex, h)
! Find optimal (sub)tree with maximal depth restrictions only (no size restriction).
! Return the optimal objective function value,
! TO BE USED AS A BOUND for the OptMaxSizeTreeBB
! Arguments
INTEGER Level, N
INTEGER ObsIndex(N)
REAL(dp), INTENT(OUT) :: h
! Locals
INTEGER MinMSpace(NPred), MaxMSpace(NPred), Dif
INTEGER p, s, N1, N2, ObsIndex1(N), ObsIndex2(N)
REAL(dp)  h1, h2
! Begin routine

CALL GetMinMaxMSpace(N, ObsIndex, MinMSpace, MaxMSpace)
Dif = SUM(MaxMSpace - MinMSpace)
h = GetHeterogeneity (N,ObsIndex)

IF ((Dif == 0) .OR. (level == MaxLevel) .OR. (N <= MinNodeSize) .OR. VeryHomogeneous(h,N)) THEN
   Depth = MAX(Depth,level+1)
   IF (N < MinNodeSize) h = SomethingLarge
ELSE
   DO p=1,NPred ! over all predictors
      !IF (Level == 0) CALL OutputProgressP (p)
      DO s=MinMSpace(p),MaxMSpace(p)-1 ! and all possible splits
         IF (Level == 0) CALL OutputProgress (p,s)
         IF ((Level <= MaxILevel) .AND. (IsInterrupted())) RETURN

         ! Excute Split
         IF (GetValidSplitAny(s, p, N, ObsIndex, N1, ObsIndex1, N2, ObsIndex2)) THEN
            IF (N2>N1) THEN !Reverse N2>N1= more nodes, less work
               CALL GetOptMaxDepthTreeBound (Level+1, N1, ObsIndex1, h1)
               IF (h1<h) THEN
                  CALL GetOptMaxDepthTreeBound (Level+1, N2, ObsIndex2, h2)
                  IF (h1+h2<h) THEN ! Better split found
                     h = h1 + h2
                  END IF
               END IF
            ELSE
               CALL GetOptMaxDepthTreeBound (Level+1, N2, ObsIndex2, h2)
               IF (h2<h) THEN
                  CALL GetOptMaxDepthTreeBound (Level+1, N1, ObsIndex1, h1)
                  IF (h1+h2<h) THEN ! Better split found
                     h = h1 + h2
                  END IF
               END IF
            END IF
         END IF
      END DO
   END DO

END IF

END SUBROUTINE GetOptMaxDepthTreeBound



RECURSIVE SUBROUTINE OptMaxSizeTreeBB (Level, N, ObsIndex, MaxSize, h)
! Find optimal tree with maximal depth AND maximum size restrictions
! Include a Branch and Bound strategy:
!     the tree search is bounded by a search strategy using the maximum depth only.
! Return ONLY optimal objective function value.
! Arguments
INTEGER Level, N
INTEGER ObsIndex(N)
INTEGER, INTENT(IN) :: MaxSize
REAL(dp), INTENT(OUT) :: h
! Locals
INTEGER MinMSpace(NPred), MaxMSpace(NPred), Dif
INTEGER p, s, N1, N2, ObsIndex1(N), ObsIndex2(N)
INTEGER SubMaxSize, MaxSubMaxSize,MaxMaxSize
REAL(dp)  h1, h2
LOGICAL SplitFound
! Begin routine

nN = nN +1

CALL GetMinMaxMSpace(N, ObsIndex, MinMSpace, MaxMSpace)
Dif = SUM(MaxMSpace - MinMSpace)
h = GetHeterogeneity (N,ObsIndex)


IF ((MaxSize <= 1) .OR. &
    (Dif == 0) .OR. (level == MaxLevel) .OR. (N <= MinNodeSize) .OR. &
    VeryHomogeneous(h,N)) THEN
   Depth = MAX(Depth,level+1)
   IF (N < MinNodeSize) h = SomethingLarge
ELSE
   SplitFound = .FALSE.
   IF (Level == 0) THEN
      h = BoundH
!   ELSE
!      CALL GetOptMaxDepthTreeBound (Level+1, N, ObsIndex, h)
   END IF
   MaxSubMaxSize = MIN(MaxSize-1,MaxNLeaves(MaxLevel-Level))
   MaxMaxSize = MIN(MaxSize,MaxNLeaves(1+MaxLevel-Level))

   DO p=1,NPred ! over all predictors
      !IF (Level == 0) CALL OutputProgressP (p)
      DO s=MinMSpace(p),MaxMSpace(p)-1 ! and all possible splits
         IF (Level == 0) CALL OutputProgress (p,s)
         IF ((Level <= MaxILevel) .AND. (IsInterrupted())) RETURN

         ! Excute Split
         IF (GetValidSplitAny(s, p, N, ObsIndex, N1, ObsIndex1, N2, ObsIndex2)) THEN
            nS = nS + 1
!            IF (Level > MaxLevel-3) THEN
!            IF (Level == 0) THEN
            IF (Level < MaxLevel-1) THEN
               CALL GetOptMaxDepthTreeBound (Level+1, N1, ObsIndex1, h1)
            ELSE
               h1 = 0
            END IF
            IF (h1<h) THEN
!             IF (Level > MaxLevel-3) THEN
!             IF (Level == 0) THEN
            IF (Level < MaxLevel-1) THEN
                CALL GetOptMaxDepthTreeBound (Level+1, N2, ObsIndex2, h2)
             ELSE
                h2 = 0
             END IF
             IF (h1+h2<h) THEN
               DO SubMaxSize=MaxSubMaxSize,MaxMaxSize-MaxSubMaxSize,-1
                  IF (N1>N2) THEN
                     CALL OptMaxSizeTreeBB (Level+1, N1, ObsIndex1, SubMaxSize, &
                                            h1)
                  ELSE
                     CALL OptMaxSizeTreeBB (Level+1, N2, ObsIndex2, SubMaxSize, &
                                            h1)
                  END IF
                  IF (h1<h) THEN
                     nST = nST +1
                     IF (N1>N2) THEN
                        CALL OptMaxSizeTreeBB (Level+1, N2, ObsIndex2, MaxSize-SubMaxSize, &
                                              h2)
                     ELSE
                        CALL OptMaxSizeTreeBB (Level+1, N1, ObsIndex1, MaxSize-SubMaxSize, &
                                              h2)
                     END IF
                     IF (h1+h2<h) THEN ! Better split found
                        SplitFound = .TRUE.
                        h = h1 + h2
                     END IF
                  ELSE
                     EXIT
                  END IF
               END DO
             END IF
            END IF
         END IF
      END DO
   END DO
   IF ((level==0) .AND. .NOT. SplitFound) h = GetHeterogeneity (N,ObsIndex)
END IF

END SUBROUTINE OptMaxSizeTreeBB




RECURSIVE SUBROUTINE OptMaxSizeTreeAll (Level, N, ObsIndex, MaxSize, h)
! Find optimal tree with maximal depth AND maximum size restrictions
! Include a Branch and Bound strategy:
!     the tree search is bounded by a search strategy using the maximum depth only.
! Return optimal objective function for all sizes<MaxSize.
! Arguments
INTEGER Level, N
INTEGER ObsIndex(N)
INTEGER, INTENT(IN) :: MaxSize
REAL(dp), INTENT(OUT) :: h(MaxSize)
! Locals
INTEGER MinMSpace(NPred), MaxMSpace(NPred), Dif
INTEGER p, s, N1, N2, ObsIndex1(N), ObsIndex2(N)
INTEGER Size, SubMaxSize
REAL(dp)  h1(MaxSize-1), h2(MaxSize-1)
! Begin routine

nN = nN +1
nNL(Level) = nNL(Level)+1

CALL GetMinMaxMSpace(N, ObsIndex, MinMSpace, MaxMSpace)
Dif = SUM(MaxMSpace - MinMSpace)
h = GetHeterogeneity (N,ObsIndex) !<= all h

IF ((MaxSize <= 1) .OR. &
    (Dif == 0) .OR. (level == MaxLevel) .OR. (N <= MinNodeSize) .OR. &
    VeryHomogeneous(h(1),N)) THEN
   Depth = MAX(Depth,level+1)
   IF (N < MinNodeSize) h = SomethingLarge
ELSE
   DO p=1,NPred ! over all predictors
      !IF (Level == 0) CALL OutputProgressP (p)
      DO s=MinMSpace(p),MaxMSpace(p)-1 ! and all possible splits
         IF (Level == 0) CALL OutputProgress (p,s)
         IF ((Level <= MaxILevel) .AND. (IsInterrupted())) RETURN

         ! Excute Split
         IF (GetValidSplitAny(s, p, N, ObsIndex, N1, ObsIndex1, N2, ObsIndex2)) THEN
            nS = nS + 1
            CALL OptMaxSizeTreeAll (Level+1, N1, ObsIndex1, MaxSize-1, h1)
            CALL OptMaxSizeTreeAll (Level+1, N2, ObsIndex2, MaxSize-1, h2)
            DO Size=2,MaxSize
               DO SubMaxSize=Size-1,1,-1
                  IF (h1(SubMaxSize)+h2(Size-SubMaxSize)<h(Size)) THEN ! Better split found
                     h(Size) = h1(SubMaxSize) + h2(Size-SubMaxSize)
                  END IF
               END DO
            END DO
         END IF
      END DO
   END DO

   ! Propagate best values for smaller trees
   DO Size=2,MaxSize
      h(Size) = MINVAL(h(1:Size))
   END DO

END IF

END SUBROUTINE OptMaxSizeTreeAll




RECURSIVE SUBROUTINE OptMaxSizeTreeAllTest (Level, N, ObsIndex, MaxSize, h)
!USE MexLib
!
!Deze versie gebruikt 2 level2 heap. Dit is nog experimenteel, eerste resultaten waren niet bemoedigend.
!Dit is waarschijnlijk de enige variant die de level two heap gebruikt.
!
! Find optimal tree with maximal depth AND maximum size restrictions
! Include a Branch and Bound strategy:
!     the tree search is bounded by a search strategy using the maximum depth only.
! Return optimal objective function for all sizes<MaxSize.
! Arguments
INTEGER Level, N
INTEGER ObsIndex(N)
INTEGER, INTENT(IN) :: MaxSize
REAL(dp), INTENT(OUT) :: h(MaxSize)
! Locals
INTEGER MinMSpace(NPred), MaxMSpace(NPred), Dif
INTEGER p, s, N1, N2, ObsIndex1(N), ObsIndex2(N)
INTEGER Size, SubMaxSize
REAL(dp)  h1(MaxSize-1), h2(MaxSize-1)
INTEGER s1, s2
!KeepS,
TYPE (TPPHeap) :: PPHeap
LOGICAL NoSkip
! Begin routine

nN = nN +1
nNL(Level) = nNL(Level)+1
IF (Level==0) Path=0

CALL GetMinMaxMSpace(N, ObsIndex, MinMSpace, MaxMSpace)
Dif = SUM(MaxMSpace - MinMSpace)
h = GetHeterogeneity (N,ObsIndex) !<= all h

IF ((MaxSize <= 1) .OR. &
    (Dif == 0) .OR. (level == MaxLevel) .OR. (N <= MinNodeSize) .OR. &
    VeryHomogeneous(h(1),N)) THEN
   Depth = MAX(Depth,level+1)
   IF (N < MinNodeSize) h = SomethingLarge
ELSE
   NoSkip=.TRUE.
   IF (Level==2) THEN
      PPHeap%Seg=>PP(Path(1,1),Path(2,1))%Seg
      IF (Path(1,1)/=Path(2,1)) THEN
         IF (Path(1,1)<Path(2,1)) THEN
            s1=Path(1,2)
            s2=Path(2,2)
            IF (s1<MaxS(Path(1,1))) s1=MIN(s1,MaxMSpace(Path(1,1)))
            IF (s2<MaxS(Path(2,1))) s2=MIN(s2,MaxMSpace(Path(2,1)))
         ELSE
            s2=Path(1,2)
            s1=Path(2,2)
            IF (s2<MaxS(Path(1,1))) s2=MIN(s2,MaxMSpace(Path(1,1)))
            IF (s1<MaxS(Path(2,1))) s1=MIN(s1,MaxMSpace(Path(2,1)))
         END IF
      ELSE
         PPHeap%Seg=>PP(Path(1,1),Path(1,1))%Seg
         s1=MinMSpace(Path(1,1))
         s2=MaxMSpace(Path(1,1))
      END IF
      IF (PPHeap%Seg(s1,s2,1)>-1.0) THEN
         h(2:MaxSize)=PPHeap%Seg(s1,s2,1:(MaxSize-1))
         nR=nR+1
         NoSkip=.FALSE.
      END IF
   END IF

   IF (NoSkip) THEN
      DO p=1,NPred ! over all predictors
         Path(Level+1,1)=p
         !IF (Level == 0) CALL OutputProgressP (p)
         DO s=MinMSpace(p),MaxMSpace(p)-1 ! and all possible splits
            IF (Level == 0) CALL OutputProgress (p,s)
            IF ((Level <= MaxILevel) .AND. (IsInterrupted())) RETURN

            ! Excute Split
            IF (GetValidSplitAny(s, p, N, ObsIndex, N1, ObsIndex1, N2, ObsIndex2)) THEN
               nS = nS + 1
               Path(Level+1,2)=s
               CALL OptMaxSizeTreeAllTest (Level+1, N1, ObsIndex1, MaxSize-1, h1)
               Path(Level+1,2)=MaxS(p)+s
               CALL OptMaxSizeTreeAllTest (Level+1, N2, ObsIndex2, MaxSize-1, h2)
               DO Size=2,MaxSize
                  DO SubMaxSize=Size-1,1,-1
                     IF (h1(SubMaxSize)+h2(Size-SubMaxSize)<h(Size)) THEN ! Better split found
                        h(Size) = h1(SubMaxSize) + h2(Size-SubMaxSize)
                     END IF
                  END DO
               END DO
            END IF
         END DO
      END DO

      ! Propagate best values for smaller trees
      DO Size=2,MaxSize
         h(Size) = MINVAL(h(1:Size))
      END DO

      !IF (Level==2) PPHeap%Seg(s1,s2,1:MaxSize-1)=h(2:MaxSize)
      !IF (Level==2) THEN
      !   IF (PPHeap%Seg(s1,s2,1)>-1.0) THEN
      !      IF (ABS(PPHeap%Seg(s1,s2,1)-h(1))>0.1) THEN
      !         DO Size=1,1 !MaxSize
      !            WRITE (GOutUnit, '(5I3,2F15.4," VIOLATED")') Size,Path(1,1),Path(2,1),s1,s2,h1(Size),PPHeap%Seg(s1,s2,Size)
      !         END DO
      !      END IF
      !   END IF
      !   PPHeap%Seg(s1,s2,1:MaxSize)=h(1:MaxSize)
      !END IF

   END IF
END IF

END SUBROUTINE OptMaxSizeTreeAllTest




RECURSIVE SUBROUTINE OptMaxSizeTreeTAll (Level, N, ObsIndex, MaxSize, h, Tree, TreeSize)
! Find optimal tree with maximal depth AND maximum size restrictions
! Include a Branch and Bound strategy:
!     the tree search is bounded by a search strategy using the maximum depth only.
! Return optimal objective function AND tree structures for all sizes<MaxSize.
! Arguments

INTEGER, INTENT(IN)  :: Level, N
INTEGER, INTENT(IN)  :: ObsIndex(N)
INTEGER, INTENT(IN)  :: MaxSize
REAL(dp), INTENT(OUT)  :: h(MaxSize), Tree(MaxSize,5,MaxNNodes(MaxSize))
INTEGER, INTENT(OUT) :: TreeSize(MaxSize)
! Locals
INTEGER MinMSpace(NPred), MaxMSpace(NPred), Dif, MaxNodes
!, MaxSubNodes
INTEGER p, s, N1, N2, ObsIndex1(N), ObsIndex2(N)
!,nnp
INTEGER Size, SubMaxSize, Tree1Size(MaxSize-1), Tree2Size(MaxSize-1)
REAL(dp)  h1(MaxSize-1), h2(MaxSize-1)
REAL(dp)  Tree1(MaxSize-1,5,MaxNNodes(MaxSize-1)),Tree2(MaxSize-1,5,MaxNNodes(MaxSize-1))
!REAL(dp)  ht(MaxSize)
! Begin routine

nN = nN +1

CALL GetMinMaxMSpace(N, ObsIndex, MinMSpace, MaxMSpace)
Dif = SUM(MaxMSpace - MinMSpace)
h = GetHeterogeneity (N,ObsIndex) !<= all h
! Node needs to be defined allways in case none of the splits improves
TreeSize = 1
Tree(1:MaxSize,5,1) = h
Tree(1:MaxSize,1:4,1) = 0.d0
MaxNodes = MaxNNodes(MaxSize)

IF ((MaxSize <= 1) .OR. &
    (Dif == 0) .OR. (level == MaxLevel) .OR. (N <= MinNodeSize) .OR. &
    VeryHomogeneous(h(1),N)) THEN
   Depth = MAX(Depth,level+1)
   IF (N < MinNodeSize) THEN
      h = SomethingLarge
      Tree(1:MaxSize,5,1) = h
   END IF
ELSE
   DO p=1,NPred ! over all predictors
   !IF (Level == 0) THEN
   !     nnp=1
   !ELSE
   !     nnp=NPred
   !END IF
   !DO p=1,nnp ! over all predictors
      !IF (Level == 0) CALL OutputProgressP (p)
      DO s=MinMSpace(p),MaxMSpace(p)-1 ! and all possible splits
         IF (Level == 0) CALL OutputProgress (p,s)
         IF ((Level <= MaxILevel) .AND. (IsInterrupted())) RETURN

         ! Excute Split
         IF (GetValidSplitAny(s, p, N, ObsIndex, N1, ObsIndex1, N2, ObsIndex2)) THEN
            nS = nS + 1
            CALL OptMaxSizeTreeTAll (Level+1, N1, ObsIndex1, MaxSize-1, h1, Tree1, Tree1Size)
            CALL OptMaxSizeTreeTAll (Level+1, N2, ObsIndex2, MaxSize-1, h2, Tree2, Tree2Size)
            DO Size=2,MaxSize
    !           ht(Size)= SomethingLarge
               DO SubMaxSize=Size-1,1,-1
    !              IF ((Level == 0) .AND. (h1(SubMaxSize)+h2(Size-SubMaxSize)<ht(Size))) THEN
    !                 ht(Size) = h1(SubMaxSize) + h2(Size-SubMaxSize)
    !              END IF
                  IF (h1(SubMaxSize)+h2(Size-SubMaxSize)<h(Size)) THEN ! Better split found
                     h(Size) = h1(SubMaxSize) + h2(Size-SubMaxSize)
                     TreeSize(Size)=1+Tree1Size(SubMaxSize)+Tree2Size(Size-SubMaxSize)
                     Tree(Size,1:4,1) = (/p, s, 1, 1+Tree1Size(SubMaxSize)/)
                     Tree(Size,5,1) =  h(Size)
                     Tree(Size,1:5,2:1+Tree1Size(SubMaxSize)) = &
                        Tree1(SubMaxSize,1:5,1:Tree1Size(SubMaxSize))
                     Tree(Size,1:5,2+Tree1Size(SubMaxSize):TreeSize(Size)) = &
                        Tree2(Size-SubMaxSize,1:5,1:Tree2Size(Size-SubMaxSize))
                  END IF
               END DO
            END DO
     !       IF (Level == 0) CALL mexWriteRV(0,ht,MaxSize)
         END IF
      END DO
     ! IF (Level == 0) CALL mexPrintFF(0,CHAR(10))
   END DO

   ! Propagate best values for smaller trees
   DO Size=2,MaxSize
      IF (h(Size-1)<= h(Size)) THEN
         h(Size) = h(Size-1)
         TreeSize(Size) = TreeSize(Size-1)
         Tree(Size,1:5,1:MaxNodes) = Tree(Size-1,1:5,1:MaxNodes)
      END IF
      IF (Level == 0) Tree(Size,1:5,TreeSize(Size)+1:MaxNodes) = -1 ! Signal unused nodes
   END DO
END IF

END SUBROUTINE OptMaxSizeTreeTAll






PURE INTEGER FUNCTION MaxNLeaves(NLevels)
! Returns the maximum number of leaves (endnodes) in a (sub)tree
! with depth of NLevels
! Arguments
INTEGER, INTENT(IN) :: NLevels
! Begin routine

   IF (NLevels>0) THEN
      MaxNLeaves = 2**(NLevels-1)
   ELSE
      MaxNLeaves = 0
   END IF

END FUNCTION MaxNLeaves



PURE INTEGER FUNCTION MaxNNodes(MaxSize)
! Returns the maximum number of leaves (endnodes) in a (sub)tree
! with depth of NLevels
! Arguments
INTEGER, INTENT(IN) :: MaxSize
! Begin routine

MaxNNodes = 1+2*(MaxSize-1)

END FUNCTION MaxNNodes



LOGICAL FUNCTION GetValidSplitAny(s,p,N,ObsIndex,N1,ObsIndex1,N2,ObsIndex2)
! Dispatches to GetValidSplit or GetValidSplitCat based on PredType(p)
INTEGER  s,p,N,N1,N2
INTEGER  ObsIndex(N),ObsIndex1(N),ObsIndex2(N)
! Begin routine
IF (PredType(p) == 1) THEN
   GetValidSplitAny = GetValidSplitCat(s,p,N,ObsIndex,N1,ObsIndex1,N2,ObsIndex2)
ELSE
   GetValidSplitAny = GetValidSplit(s,p,N,ObsIndex,N1,ObsIndex1,N2,ObsIndex2)
END IF
END FUNCTION GetValidSplitAny



LOGICAL FUNCTION GetValidSplit(s,p,N,ObsIndex,N1,ObsIndex1,N2,ObsIndex2)
! returns the splitted measurement space in ObsIndex1 and 2
! and whether the split s is valid:
!     a. no empty nodes on either side
!     b. the split is not congruent i.e. not redundant w.r.t. other splits
!        (two splits s1,s2 are congruent if {group1}<s1<s2<{group2}
! Arguments
INTEGER  s,p,N,N1,N2
INTEGER  ObsIndex(N),ObsIndex1(N),ObsIndex2(N)
! Locals
INTEGER i
! Begin routine

GetValidSplit = .FALSE.
N1 = 0
N2 = 0
DO i=1,N
   IF (MSpace(ObsIndex(i),p) <= s) THEN
      !The next condition skips redundant splits (factually equal to other splits)
      IF (MSpace(ObsIndex(i),p) == s) GetValidSplit = .TRUE.
      N1 = N1+1
      ObsIndex1(N1)=ObsIndex(i)
   ELSE
      N2 = N2+1
      ObsIndex2(N2)=ObsIndex(i)
   END IF
END DO
IF ((N1 == 0) .OR. (N2 == 0)) GetValidSplit = .FALSE.

END FUNCTION GetValidSplit



LOGICAL FUNCTION GetValidSplitCat(s,p,N,ObsIndex,N1,ObsIndex1,N2,ObsIndex2)
! Categorical variant of GetValidSplit.
! s is a bitmask: bit k-1 set means category k goes left.
! Enumerates 1..2^(NCatX(p)-1)-1 bitmasks; complement symmetry handled by caller.
! Valid if both sides non-empty.
! Arguments
INTEGER  s,p,N,N1,N2
INTEGER  ObsIndex(N),ObsIndex1(N),ObsIndex2(N)
! Locals
INTEGER i, cat
! Begin routine

GetValidSplitCat = .FALSE.
N1 = 0
N2 = 0
DO i=1,N
   cat = MSpace(ObsIndex(i),p)              ! category value (1..NCatX(p))
   IF (IAND(s, ISHFT(1,cat-1)) /= 0) THEN  ! bit cat-1 set => goes left
      N1 = N1+1
      ObsIndex1(N1) = ObsIndex(i)
   ELSE
      N2 = N2+1
      ObsIndex2(N2) = ObsIndex(i)
   END IF
END DO
IF ((N1 /= 0) .AND. (N2 /= 0)) GetValidSplitCat = .TRUE.

END FUNCTION GetValidSplitCat



RECURSIVE SUBROUTINE QSortObjIndex(O,l,r,N,p)
! Function    : Quick sort object indices according to predictor p
! Version     : 1.0, November 1996
! Author      : Bart Jan van Os
! Used by     :
! Called by   :
! Parameter block.
INTEGER     l,r,N,p
INTEGER     O(N)
! Local variables.
INTEGER     i,j
INTEGER     x,y
! Program start

  i = l
  j = r
  x = O((l+r)/ 2);
  DO WHILE (i<=j)
    DO WHILE (MSpace(O(i),p)<MSpace(x,p))
       i = i+1
    END DO
    DO WHILE (MSpace(x,p)<MSpace(O(j),p))
       j = j-1
    END DO
    IF (i<=j) THEN !swap i and j
      y = O(i)
      O(i) = O(j)
      O(j) = y
      i = i+1
      j = j-1
    END IF
  END DO
  IF (l<j) CALL QSortObjIndex(O,l,j,N,p)
  IF (i<r) CALL QSortObjIndex(O,i,r,N,p)

END SUBROUTINE QSortObjIndex



SUBROUTINE SetupValidNextSplit(s,p,N,ObsIndex,N1,ObsIndex1,N2,ObsIndex2)
! sets up the function GetValidNextSplit by
! - sorting the ObsIndex according to the predictor
! - set ObsIndex1 to this order
! - set ObsIndex2 to reverse of this order
! - set N1=0
! Arguments
INTEGER  s,p,N,N1,N2
INTEGER  ObsIndex(N),ObsIndex1(N),ObsIndex2(N)
! Begin routine

N2=N2

ObsIndex1=ObsIndex ! crucial, because otherwise recursive routine fails one level back
!CurP=p
CALL QSortObjIndex(ObsIndex1,1,N,N,p)
ObsIndex2=ObsIndex1(N:1:-1)
N1=0
s=0


END SUBROUTINE SetupValidNextSplit



LOGICAL FUNCTION GetValidNextSplit(s,p,N,N1,N2,ObsIndex1)
! returns the splitted measurement space in ObsIndex1 and 2
! this version assumes that observation index is sorted with respect to the predictor
! and whether the split s is valid:
!     a. no empty nodes on either side
!     b. the split is not congruent i.e. not redundant w.r.t. other splits
!        (two splits s1,s2 are congruent if {group1}<s1<s2<{group2}
! Arguments
INTEGER  s,p,N,N1,N2
INTEGER  ObsIndex1(N)
! Locals
!INTEGER i
! Begin routine

GetValidNextSplit = .FALSE.
N1=N1+1
s=MSpace(ObsIndex1(N1),p)
DO WHILE ((N1<N) .AND. (MSpace(ObsIndex1(N1+1),p) == s))
   N1=N1+1
END DO
IF (N1<N) THEN
   GetValidNextSplit = .TRUE.
   N2=N-N1
END IF

END FUNCTION GetValidNextSplit




SUBROUTINE GetMinMaxMSpace(N, ObsIndex, MinMSpace, MaxMSpace)
! Arguments
INTEGER N
INTEGER ObsIndex(N), MinMSpace(NPred), MaxMSpace(NPred)
! Locals
INTEGER p
! Begin routine

MinMSpace = MINVAL(MSpace(ObsIndex,1:NPred),1)
MaxMSpace = MAXVAL(MSpace(ObsIndex,1:NPred),1)

! Override range for categorical predictors: enumerate bitmasks 1..2^(k-1)
! so that the standard DO s=MinMSpace(p),MaxMSpace(p)-1 loop covers 1..2^(k-1)-1
DO p=1,NPred
   IF (PredType(p) == 1) THEN
      MinMSpace(p) = 1
      MaxMSpace(p) = ISHFT(1, NCatX(p)-1)  ! 2^(k-1), loop goes to MaxMSpace-1
   END IF
END DO

END SUBROUTINE GetMinMaxMSpace



SUBROUTINE GetMinMaxMSpaceOri(N, ObsIndex, MinMSpace, MaxMSpace)
! Arguments
INTEGER N
INTEGER ObsIndex(N), MinMSpace(NPred), MaxMSpace(NPred)
! Locals
INTEGER i,p
! Begin routine

MinMSpace = MSpace(ObsIndex(1),1:NPred)
MaxMSpace = MSpace(ObsIndex(1),1:NPred)

DO p=1,NPred
   DO i=2,N
      IF (MSpace(ObsIndex(i),p)>MaxMSpace(p)) THEN
         MaxMSpace(p) = MSpace(ObsIndex(i),p)
      ELSE IF (MSpace(ObsIndex(i),p)<MinMSpace(p)) THEN
         MinMSpace(p) = MSpace(ObsIndex(i),p)
      END IF
   END DO
END DO

END SUBROUTINE GetMinMaxMSpaceOri




LOGICAL FUNCTION VeryHomogeneous(h,N)
! Arguments
REAL(dp)   h
INTEGER  N
! Begin routine

!avoid this external call if possible, extremely slow
!IF (CallBack) THEN
!   VeryHomogeneous = GetExternVeryHomogeneous(h,N)
!ELSE
   SELECT CASE (Measure)
   CASE (cESS)
      VeryHomogeneous = (h/N)<normMSS;
   CASE (cClassErr)
      VeryHomogeneous = (h/N)<minH;
   CASE (cMinRisk)
      VeryHomogeneous = h<normRisk;
   CASE DEFAULT
      VeryHomogeneous = (h/N)<normMSS;
   END SELECT
!END IF

END FUNCTION VeryHomogeneous



REAL(dp) FUNCTION GetHeterogeneity(N,ObsIndex)
! Arguments
INTEGER  N
INTEGER  ObsIndex(N)
! Begin routine

IF (CallBack) THEN
   GetHeterogeneity = GetExternHeterogeneity(N,ObsIndex)
ELSE
   SELECT CASE (Measure)
   CASE (cESS)
      GetHeterogeneity = GetSSE(N,ObsIndex)
   CASE (cClassErr)
      GetHeterogeneity = GetMisclassification(N,ObsIndex)
   CASE (cMinRisk)
      GetHeterogeneity = GetExpectedRisk(N,ObsIndex)
   CASE DEFAULT
      GetHeterogeneity = 0.0
   END SELECT
END IF

END FUNCTION GetHeterogeneity




REAL(dp) FUNCTION GetSSE(N,ObsIndex)
! Arguments
INTEGER  N
INTEGER  ObsIndex(N)
! Locals
INTEGER d,i
REAL(dp)  S,SS,SSError
! Begin routine

SSError=0.0
DO d=1,NDep
   SS = 0.d0
   S=0.d0
   DO i=1,N
      S=S+YR(ObsIndex(i),d)
      SS=SS+YR(ObsIndex(i),d)**2
   END DO
   SSError=SSError+SS-(S**2)/N
END DO

GetSSE = SSError

END FUNCTION GetSSE



INTEGER FUNCTION GetMisclassification(N,ObsIndex)
! Arguments
INTEGER  N
INTEGER  ObsIndex(N)
! Locals
INTEGER d,i, Freq(MaxMaxYI), Error

! Begin routine

Error = 0
DO d=1,NDep
   Freq(1:MaxYI(d)) = 0
   DO i=1,N
      Freq(YI(ObsIndex(i),d)) = Freq(YI(ObsIndex(i),d)) +1
   END DO
   Error = Error + SUM(Freq(1:MaxYI(d))) - MAXVAL(Freq(1:MaxYI(d)))
END DO

GetMisclassification = Error

END FUNCTION GetMisclassification



SUBROUTINE GetClassFreq(N,ObsIndex,AClassFreq)
! Arguments
INTEGER  N
!,MaxClass
INTEGER  ObsIndex(N),AClassFreq(NTotClass)
! Locals
INTEGER d,i,s

! Begin routine
s=0
DO d=1,NDep
   AClassFreq = 0
   DO i=1,N
      AClassFreq(s+YI(ObsIndex(i),d)) = AClassFreq(s+YI(ObsIndex(i),d)) +1
   END DO
   s=s+MaxYI(d)
END DO

END SUBROUTINE GetClassFreq



REAL(dp) FUNCTION GetExpectedRisk(N,ObsIndex)
! Arguments
INTEGER  N
INTEGER  ObsIndex(N)
! Locals
INTEGER d,s,ss
INTEGER  F(NTotClass)
REAL(dp)  ExpectedRisk
! Begin routine

ExpectedRisk=0.0
CALL GetClassFreq(N,ObsIndex,F)

s=1
ss=1
DO d=1,NDep
   ExpectedRisk = ExpectedRisk + &
                 CalculateRiskA(MaxYI(d),F(s),Prior(s),LossM(ss))
END DO

GetExpectedRisk = ExpectedRisk

END FUNCTION GetExpectedRisk



REAL(dp) FUNCTION CalculateRiskA(NCat,F,Prior,LossM)
! Arguments
INTEGER  NCat
INTEGER  F(NCat)
REAL(dp)   Prior(NCat),LossM(NCat,NCat)
! Locals
!INTEGER d,i
REAL(dp)  E(NCat),piA(NCat),pa,ra
! Begin routine

   E=Prior*F/ClassFreqY
   pa=SUM(E);
   piA=E/pa;
   E=MATMUL(piA,LossM)
   ra=MINVAL(E);
   CalculateRiskA=pa*ra;

END FUNCTION CalculateRiskA



END MODULE OptimalTrees

SetupMeasure<-function(defMeasure){ #defMeasure is supposed to be a cell which is like a 2-d list
#%passes measure and checks whether defMeasure includes a Prior and Loss (Measure=3)
Prior<-list()
LossM<-list()

if(is.list(defMeasure)){
  measureCB <- defMeasure[[1]] # measure + callback bit
  measure <- bitwAnd(measureCB,127) # strip callback bit
  if(measure!=3){
    stop('Measure parameter is a list, but measure type is not 3')
  }
  if(length(defMeasure)>1){#MISSING IN R MAKES NO SENSE A 2-D LIST
    if(!is.list(defMeasure[[2]])){ #%ensure it has the correct cell structure
      Prior[[1]]<-matrix(defMeasure[[2]],ncol=1)
    }else{
      Prior<-defMeasure[[2]]
    }

    if(length(defMeasure)>2){
      if (!is.list(defMeasure[[3]])){ #%ensure it has the correct cell structure
        LossM[[1]]<-defMeasure[[3]]
      }else{
        LossM<-defMeasure[[3]]
      }
    }
  }


}else{
  measureCB <- defMeasure # measure + callback bit
  measure <- bitwAnd(measureCB,127) # strip callback bit
}

return(list(measure=measure, measureCB=measureCB, Prior=Prior, LossM=LossM))
}

ProcessSubTree<-function(Y,X,node,Tree, check, measure, Prior, LossM, ClassSizes, Indices){
  #library(Rfast)
  #global Tree check MISSING

  #Tree<-globalenv()$Tree

  n<-NROW(X)
  if(n==0){
    stop('empty node!')
  }

  h<-GetHeterogeneity(Y=Y, measure=measure, Prior=Prior, LossM=LossM, ClassSizes=ClassSizes)
  if(h>0){
    if(check && (Tree[node,1]<=0) && (abs(Tree[node,5]-h)/h)>0.01){
      warning(paste0('Heterogeneity different for node ',node, Tree[node,5],h))
      Tree[node,5]<-h
    }
  }

  Tree[node,6]<-NROW(Y)
  Tree[node,7]<-GetPrediction(Y, measure, Prior, LossM, ClassSizes)
  if(NROW(X)==1){
    Cat<-rbind(X, X)
  }else{
    Cat<-rbind(apply(X,2,min), apply(X,2,max))
  }
  Tree[node,8:NCOL(Tree)]<-matrix(as.vector(Cat),nrow=1)

  Indices[[2*node-1]]<-0
  Indices[[2*node]]<-0

  # Sizes[[2*node-1]]<-list()
  # Sizes[[2*node]]<-list()
  #
  # Prob[[2*node-1]]<-list()
  # Prob[[2*node]]<-list()


  if(Tree[node,1]>0){
    p<-Tree[node,1]
    s<-Tree[node,2]
    SpSpout<-SplitSpace(Y,X,p,s)
    Y1<-SpSpout$Y1
    if(NCOL(Y1)==1 || NROW(Y1)==1){
      Y1<-matrix(Y1, ncol = 1)
    }
    X1<-SpSpout$X1
    if(NCOL(X1)==1 || NROW(X1)==1){
      X1<-matrix(X1, ncol = 1)
    }
    Y2<-SpSpout$Y2
    if(NCOL(Y2)==1 || NROW(Y2)==1){
      Y2<-matrix(Y2, ncol = 1)
    }
    X2<-SpSpout$X2
    if(NCOL(X2)==1 || NROW(X2)==1){
      X2<-matrix(X2, ncol = 1)
    }

    Indices[[2*node-1]]<-SpSpout$I1
    Indices[[2*node]]<-SpSpout$I2


    # if(measure!=0){
    #   Yleft<-as.factor(Y[Indices[[2*node-1]],])
    #   if(length(Yleft)!=0){
    #     Sizes[[2*node-1]]<-table(Yleft)
    #     names(Sizes[[2*node-1]])<-NULL
    #     Prob[[2*node-1]]<-Sizes[[2*node-1]]/length(Yleft)
    #   }
    #
    #   Yright<-Y[Indices[[2*node]],]
    #   if(length(Yright)!=0){
    #     Sizes[[2*node]]<-table(Yright)
    #     names(Sizes[[2*node]])<-NULL
    #     Prob[[2*node]]<-Sizes[[2*node]]/length(Yright)
    #   }
    # }


    ProcessST1<-ProcessSubTree(Y1,X1,node+Tree[node,3], Tree, check, measure, Prior, LossM, ClassSizes,Indices)
    Tree<-ProcessST1$Tree
    Indices<-ProcessST1$Indices

    ProcessST2<-ProcessSubTree(Y2,X2,node+Tree[node,4], Tree, check, measure, Prior, LossM, ClassSizes,Indices)
    Tree<-ProcessST2$Tree
    Indices<-ProcessST2$Indices

    Tree[node,3]<-node+Tree[node,3]
    Tree[node,4]<-node+Tree[node,4]
  }


  return(list(Tree=Tree,Indices=Indices))


}

# BootsTrapYX<-function(Y,X,BootSize,WithDraw){
# SampleSize<-dim(X)[1]
# if(nargs()<4){
#   I<-BootsTrapIndex(BootSize,SampleSize)
# }else{
#   I<-BootsTrapIndex(BootSize,SampleSize,WithDraw)
# }
# X<-X[I,]
# Y<-Y[I,]
#
# return(list(Y=Y, X=X))
# }

#' Summarizing Exact Trees Information
#'
#' Summary method for an object of class \code{ETree}.
#'
#' @param object a \code{ETree} object. This can be the output of \code{\link{ETree}}.
#' @param TerminalNodes Number of terminal nodes of the tree to summarize.
#' @param \dots optional additional arguments.
#' @param digits specified number of decimal places (default is 3).
#'
#' @return prints a summarized version of the \code{ETree} output.
#'
#' @details This function is a method for the generic function summary for class
#'   \code{ETree}. It extracts the following essential components from a \code{ETree}
#'   object: 1) Split information;
#'   2) Leaf information, and 3) CV information.
#'
#' @examples
#' \donttest{
#'   data(iris)
#'   # Fit an Exact Tree model
#'   controlEtree <- ETree.control(measure=0, maxsize = 4, maxdepth = 3,
#'   minbucket = 5, ncv=5, alltreesizes = FALSE)
#'   tree<-ETree(Species ~ Sepal.Length + Sepal.Width + Petal.Length + Petal.Width,
#'   control= controlEtree, data = iris)
#'   summary(tree, TerminalNodes=4)
#' }
#'
#' @keywords summary
#'
#' @export
summary.ETree<-function(object,TerminalNodes=NULL,digits=3,...){


  object<-object$Transf_Trees


  if(is.null(TerminalNodes)||(length(object)==1)){

    if(length(object)!=1){
      stop("Terminal nodes is NULL. Please, indicate the number of terminal nodes.")
    }

    object<-object[[1]]

  }else{

    if(length(object)<TerminalNodes){
      stop("TerminalNodes is larger than the maximum number of terminal nodes.")
    }

    object<-object[[TerminalNodes]]

  }

  if(is.null(dim(object$si)[2])){
    #cat("\n")
  }else{
    cat("Split information:","\n")
    if(is.numeric(object$si[,5])){
      object$si[,5]<-round(object$si[,5],digits = digits)
    }
    print(object$si,row.names=TRUE)
  }
  cat("\n")
  cat("Leaf information:","\n")
  #options(warn=-1)
  print(round(object$li[,c(2:4)],digits=digits))

  #options(warn=-1)
  if(!is.null(object$CVOutput)){
    cat("\n")
    cat("CV information:","\n")
    print(round(object$CVOutput,digits=digits))
  }

}

SetupY<-function(defY,measure){
#init Y: autorecode for discrete data (measure>1)

if((measure==1) || (measure==3)){
  Y<-defY
  for(y in 1:(NCOL(defY))){
    Y[,y]<-AutoRecode(defY[,y])$O
  }
}else{
  Y<-defY
}

return(Y)
}

CVOptTreeF<-function(CVVec,Y,X,XType,measureCB,defMaxSize,defMaxDepth,defMinNodeSize,defMinH,AlgoType,BoundH,UP,UL,IsInterrupted, LookAheadDepth, measure, Prior, LossM, ClassSizes){
#library(pracma)
#CVOptTreeF(fid,CVVec,Y,X,measureCB,defMaxSize,defMaxDepth,defMinNodeSize,defMinH,AlgoType,BoundH,UP,UL)
#  global CurrentCV LogFileName LookAheadDepth IsInterrupted measure;

#measure<-globalenv()$measure
#IsInterrupted<-globalenv()$IsInterrupted
#LookAheadDepth<-globalenv()$LookAheadDepth
#LogFileName<-globalenv()$LogFileName


if(bitwAnd(AlgoType,1)==0){
  AlgoType<-AlgoType+1 #force tree table output
}

if(bitwAnd(AlgoType,4)>0){ #all trees
  if(defMaxSize>0){
    NTrees<-defMaxSize
  }else{
    NTrees<-defMaxDepth
  }

}else{
    NTrees<-1
}


n<-NROW(X)
NCV<-max(CVVec) #if CVVec is a vector then max, otherwise apply(CVVec,2, max)
Ypred<-list() #cell(NTrees);

for (t in 1:NTrees){
  Ypred[[t]]<-matrix(0, nrow = n, ncol = 1)
}

hTotal<-rep(0,NTrees)
for(c in 1:NCV){#(c in 1:NCV){
  CurrentCV <- c

  Pt<-which(CVVec==c)
  Pl<-which(CVVec!=c)
  Yt<-Y[Pt,]
  Xt<-X[Pt,]
  Yl<-Y[Pl,]
  Xl<-X[Pl,]


  optTreeFOut<-OptTreeF(Yl,Xl,XType,measureCB,defMaxSize,defMaxDepth,defMinNodeSize,defMinH,AlgoType,
                           BoundH,UP,UL,LookAheadDepth)

  h<-optTreeFOut$h
  Tree<-optTreeFOut$Tv
  hAll<-optTreeFOut$hAll
  TAll<-optTreeFOut$TAll



  for(t in 1:NTrees){
    #do postprocessing stuff in MATLAB
    ProcessT<-ProcessTree(Y=Y,X=X,Tv=TAll[,,t],node=1,measure=measure, Prior=Prior, LossM=LossM, ClassSizes=ClassSizes)
    Tree<-ProcessT$Tv
    GLSout<-GetLearningStuff(Tree,Yl,Xl,Yt,Xt, measure=measure, Prior=Prior, LossM=LossM, ClassSizes=ClassSizes)
    Ypredl<-GLSout$Ypred
    h<-GLSout$h
    for(id in 1:length(Pt)){
      idPt<-Pt[id]
      Ypred[[t]][idPt]<-Ypredl[id]
    }
    hTotal[t]<-hTotal[t]+h
  }
}

  if(measure!=0){
    stop('noCV. CROSSVALIDATED ERROR IS NOT YET CALCULATED FOR CLASSIFICATION TREES (MEASURE=1 or 3)')
  }

  hAll<-matrix(hAll, nrow = 1)
  NTrees <- dim(hAll)[2]
  # fprintf(file=fid,fmt='\n************************************************************\n')
  # fprintf(file=fid,fmt='\n%i-FOLD CROSS VALIDATED ERROR:\n',NCV)
  # fprintf(file=fid,fmt='%s\n','Size       Error     Rel.Error            SE           SE2         CHECK')
  maxH<-GetHeterogeneity(Y=Y, measure=measure, Prior=Prior, LossM=LossM, ClassSizes=ClassSizes)

  Size<-rep(NA,NTrees)
  Error<-rep(NA,NTrees)
  SE<-rep(NA,NTrees)
  SE2<-rep(NA,NTrees)
  RelError<-rep(NA,NTrees)#Error/maxH

  for(t in 1:NTrees){
    Size[t] <- MaxEndNodes(t,NTrees,defMaxSize,defMaxDepth)
    #[Error,SE]=GetCVErrorAndSE(Y,Ypred{t});
    if(measure==0){
      GCVEASEout<-GetCVErrorAndSE(Y=Y,Ypred=Ypred[[t]], measure=measure, Prior=Prior, LossM=LossM, ClassSizes=ClassSizes)
      Error[t]<-GCVEASEout$Error
      SE[t]<-GCVEASEout$SE
      SE2[t]<-GCVEASEout$SE2
      RelError[t]<-Error[t]/maxH
      #fprintf(file=fid,fmt='%2i %13.4f %13.4f %13.4f %13.4f %13.4f\n',Size,Error,Error/maxH,SE,SE2,hTotal[t])
    }else{
      #fprintf(file=fid,fmt='%2i %13.4f %13.4f %13.4f %13.4f\n',Size,Error,Error/maxH,SE,hTotal[t])
    }
  }


  CVoutput<-list(Size=Size, Error=Error, Rel_error=RelError, SE=SE, SE2=SE2, CHECK=hTotal)

  return(CVoutput)


}

MODULE OptTreeCallBacks

!Supplementary routines for OptimalTrees modules
!This module contains all routines that need Matlab callbacks,
!including output routines

INTEGER, PARAMETER :: dp = SELECTED_REAL_KIND(15, 307)

!USE Mex
!USE MexLib

!INTEGER prhsGEH(1),prhsGEHptr

!INTEGER, PRIVATE :: &
!   prhsGEH(1), &
!   prhsGEHptr, &
!   IsInterruptedPtr = 0

CONTAINS

SUBROUTINE SetupCallBacks(N)
! Arguments
INTEGER N, IsInterruptedPtr
!INTEGER IsInterruptedPtr
! Locals
! Begin routine

!prhsGEH(1) = mxCreateNumericMatrix(1, N, mxClassIDFromClassName('int32'), 0)
!prhsGEHptr = mxGetPr(prhsGEH(1))
!IsInterruptedPtr = mexGetVariablePtr('global', 'IsInterrupted')
N=N
IsInterruptedPtr = 0

END SUBROUTINE SetupCallBacks



SUBROUTINE CloseCallBacks(N)
! Arguments
INTEGER N
! Locals
! Begin routine

!CALL mxSetN(prhsGEH(1),N)
!CALL mxSetPr(prhsGEH(1),prhsGEHptr)
!CALL mxDestroyArray(prhsGEH(1))

N=N

END SUBROUTINE CloseCallBacks



REAL(dp) FUNCTION GetExternHeterogeneity(N,ObsIndex)
!USE Mex
! Arguments
INTEGER  N
INTEGER  ObsIndex(N)
! Locals
!INTEGER plhs(1), Status
! Begin routine

!CALL mxSetN(prhsGEH(1),N)
!CALL mxCopyInteger4ToPtr(ObsIndex, prhsGEHptr, N)

!Status = mexCallMATLAB(1, plhs, 1, prhsGEH, 'HeterogeneityCallBack')
!GetExternHeterogeneity = mxGetScalar(plhs(1))

! CLEAN UP
!CALL mxDestroyArray(plhs(1))

N=N
ObsIndex=ObsIndex
GetExternHeterogeneity = 0.0

END FUNCTION GetExternHeterogeneity



LOGICAL FUNCTION GetExternVeryHomogeneous(h,N)
! Arguments
REAL(dp)   h
INTEGER  N
!locals
!INTEGER plhs(1), prhs(2), Status
! Begin routine

!prhs(1) = mxCreateNumericMatrix(1, 1, mxClassIDFromClassName('double'), 0)
!CALL mxCopyReal8ToPtr(h, mxGetPr(prhs(1)), 1)
!prhs(2) = mxCreateNumericMatrix(1, 1, mxClassIDFromClassName('int32'), 0)
!CALL mxCopyInteger4ToPtr(N, mxGetPr(prhs(2)), 1)

!Status = mexCallMATLAB(1, plhs, 2, prhs, 'VeryHomogeneousCallBack')

!IF (mxGetScalar(plhs(1)) > 0) THEN
!   GetExternVeryHomogeneous = .TRUE.
!ELSE
   GetExternVeryHomogeneous = .FALSE.
!END IF

! CLEAN UP
!CALL mxDestroyArray(plhs(1))
!CALL mxDestroyArray(prhs(1))
!CALL mxDestroyArray(prhs(2))

h=h
N=N

END FUNCTION GetExternVeryHomogeneous



SUBROUTINE Warning(Msg)
!USE MexLib
CHARACTER(LEN=*) Msg

!CALL mexWarning(Msg)

Msg=Msg

END SUBROUTINE Warning



SUBROUTINE WriteHeader(OutUnit)
!USE MexLib
!Arguments
INTEGER OutUnit
!Locals
!INTEGER i,s,ss
! Begin routine

!CALL mexPrintFF (OutUnit,'***                O P T I M A L  T R E E                ***'//CHAR(10))
!CALL mexPrintFF (OutUnit,'***GLOBALLY OPTIMIZED CLASSIFICATION AND REGRESSION TREES***'//CHAR(10))
!CALL mexPrintFF (OutUnit,'***               USING DYNAMIC PROGRAMMING              ***'//CHAR(10))
!CALL mexPrintFF (OutUnit,'***               AUTHOR: B.J. Van Os @2005              ***'//CHAR(10))
!CALL mexPrintFF (OutUnit,'************************************************************'//CHAR(10))
OutUnit=OutUnit
END SUBROUTINE WriteHeader



SUBROUTINE WriteInterruptWarning(OutUnit)
!USE MexLib
!Arguments
INTEGER OutUnit
!Locals
! Begin routine

!IF (OutUnit/=0) THEN
!   CALL mexWarning('Computational Process Interrupted and Aborted'//CHAR(10))
!END IF

!CALL mexPrintFF (OutUnit,CHAR(10)//'WARNING: COMPUTATIONAL PROCESS INTERRUPTED AND ABORTED'//CHAR(10))
!CALL mexPrintFF (OutUnit,'SUBSEQUENT OUTPUT CAN NOT BE TRUSTED.'//CHAR(10))
!CALL mexPrintFF (OutUnit,'FOUND TREES HAVE NO GLOBAL OPTIMALITY GUARANTEE:'//CHAR(10))
!CALL mexPrintFF (OutUnit,'TREES REPORTED ARE THE BEST TREES FOUND IN AN INCOMPLETE SEARCH.'//CHAR(10))
!CALL mexPrintFF (OutUnit,''//CHAR(10))
!CALL mexPrintFF (OutUnit,''//CHAR(10))

OutUnit=OutUnit
END SUBROUTINE WriteInterruptWarning



SUBROUTINE EchoInput(NColsY,Measure,NCat)
!USE MexLib
IMPLICIT NONE
!Arguments
!INTEGER OutUnit,NRows,NColsY,NColsX
!REAL(dp) Y(NRows,NColsY), X(NRows,NColsX)
!INTEGER Measure,MaxSize,MaxDepth,MinNodeSize,AlgoType,LookAheadDepth
!REAL(dp) MinH, BoundH, Prior(*), LossM(*)
!Locals
!INTEGER
INTEGER Measure, NColsY
REAL(dp) NCat(NColsY)
INTEGER i,s,ss
!CHARACTER(40) GetAlgoTypeStr
! Begin routine

IF (Measure ==3) THEN
        s=1
        ss=1
        DO i=1,NColsY
          s=s+INT(NCat(i))
          ss=ss+INT(NCat(i)**2)
        END DO
END IF


END SUBROUTINE EchoInput



SUBROUTINE WriteHeap2Allocation(OutUnit,Status,MemCount)
!USE MexLib
!Arguments
INTEGER OutUnit,Status,MemCount
!Locals
! Begin routine

!IF (Status>0) THEN
!   IF (OutUnit/=0) CALL mexWarning('Level 2 Heap allocation failed!'//CHAR(10))
!   CALL mexPrintFF (OutUnit,CHAR(10)//'WARNING: LEVEL 2 HEAP ALLOCATION FAILED'//CHAR(10))
!   CALL mexPrintFF (OutUnit,'LIKELY CAUSE: NOT ENOUGH (CONTIGUOUS) MEMORY'//CHAR(10))
!   CALL mexPrintFF (OutUnit,'LEVEL 2 HEAP NEEDS '//LongI2Str(CEILING(MemCount/(1024**2/4.0)))//' MB OF MEMORY'//CHAR(10))
!   CALL mexPrintFF (OutUnit,'=>PROCESS WILL RUN WITHOUT HEAP'//CHAR(10))
!ELSE
!   CALL mexPrintFF (OutUnit,CHAR(10)//'LEVEL 2 HEAP NEEDED '//LongI2Str(CEILING(MemCount/(1024**2/4.0)))//' MB OF MEMORY'//CHAR(10))
!END IF

OutUnit=OutUnit
Status=Status
MemCount=MemCount

END SUBROUTINE WriteHeap2Allocation



CHARACTER(30) FUNCTION GetMeasureStr(Measure)
! Returns string with algorithm type description
!        0  = regression tree
!        1  = simple misclassification rate
!        2  = Risk function
! Parameter block.
INTEGER Measure
! Local variables.
!CHARACTER(30) Str
! BEGIN FUNCTION

SELECT CASE (iand(Measure,127))
CASE (0)
   GetMeasureStr = ' Regression: SS'
CASE (1)
   GetMeasureStr = ' Classification:misclass.rate'
CASE (2)
   GetMeasureStr = ' Classification:risk function'
END SELECT

END FUNCTION GetMeasureStr





CHARACTER(20) FUNCTION GetMeasureCalculationStr(Measure)
! Returns string with algorithm type description
!        0  = regression tree
!        1  = simple misclassification rate
!        2  = Risk function
! Parameter block.
INTEGER Measure
! BEGIN FUNCTION

IF (iand(Measure,128)>0) THEN
   GetMeasureCalculationStr = ' MATLAB CALLBACK'
ELSE
   GetMeasureCalculationStr = ' DIRECT '
END IF

END FUNCTION GetMeasureCalculationStr



CHARACTER(40) FUNCTION GetAlgoTypeStr(AlgoType)
! Returns string with algorithm type description
!        0  = h only;
!        1  = h+tree table
!        2  = branch and bound, MaxSize only
!        +4 = results for all trees <= size restrictions
!        +8 = sorted predictor splitting (faster for continous predictors?)
!        default = 1
! Parameter block.
INTEGER AlgoType
! Local variables.
CHARACTER(40) Str
! BEGIN FUNCTION

Str = ''
IF (iand(AlgoType,16)>0) THEN
   Str = 'lookahead heuristic'
END IF
IF (iand(AlgoType,8)>0) THEN
   Str = 'sorted'
   !//Str
END IF
IF (iand(AlgoType,4)>0) THEN
   Str = 'all trees, '
   !//Str
END IF
SELECT CASE (iand(AlgoType,3))
CASE (0)
   Str = 'h only, '
   !//Str
CASE (1)
   Str = 'h + tree, '
   !//Str
CASE (2)
   Str = 'h (branch and bound), '
   !//Str
END SELECT

GetAlgoTypeStr = Str

END FUNCTION GetAlgoTypeStr






SUBROUTINE OutputAlgorithmDetails ()
!USE MexLib
IMPLICIT NONE
!Arguments
!INTEGER OutUnit,NTrees,Depth,nS,nN,nST
!INTEGER nNL(0:Depth-1)
!REAL(dp)  hAll(NTrees)
!Locals
!INTEGER i
! Begin routine

   !CALL mexPrintFF(OutUnit,CHAR(10)//CHAR(10)//'ALGORITHM DETAILS'//CHAR(10))
   !CALL mexWriteSpecifiedElapsedTime(OutUnit,'OptTreeF')
   !CALL mexPrintFF(OutUnit,'Remaining heterogeneity check: '//CHAR(10))
   !DO i=1,NTrees
  !    CALL mexPrintFF(OutUnit,'Tree '//I2Str(i)//' = '//R2Str(hAll(i)))
   !END DO

   !CALL mexPrintFF(OutUnit,CHAR(10))
   !CALL mexPrintFF(OutUnit,'Actual Depth:                  '//LongI2Str(Depth)//CHAR(10))
   !CALL mexPrintFF(OutUnit,'Number of splits:              '//LongI2Str(nS)//CHAR(10))
   !CALL mexPrintFF(OutUnit,'Number of nodes:               '//LongI2Str(nN)//CHAR(10))
   !DO i=1,Depth
  !    CALL mexPrintFF(OutUnit,'   Level '//I2Str(i-1)//'                 '//LongI2Str(nNL(i-1))//CHAR(10))
   !END DO
   !CALL mexPrintFF(OutUnit,'Number of subTree searches:    '//LongI2Str(nST)//CHAR(10))


END SUBROUTINE OutputAlgorithmDetails



SUBROUTINE OutputAlgorithmDetailsX (OutUnit,hAll, NTrees, Depth, nS, nN, nNL, nST, nR)
!USE MexLib
IMPLICIT NONE
!Arguments
INTEGER OutUnit,NTrees,Depth,nS,nN,nST,nR
INTEGER nNL(0:Depth-1)
REAL(dp)  hAll(NTrees)
!Locals
!INTEGER i
! Begin routine

   !CALL mexPrintFF(OutUnit,CHAR(10)//CHAR(10)//'ALGORITHM DETAILS'//CHAR(10))
   !CALL mexWriteSpecifiedElapsedTime(OutUnit,'OptTreeF')
   !CALL mexPrintFF(OutUnit,'Remaining heterogeneity check: '//CHAR(10))
   !DO i=1,NTrees
    !  CALL mexPrintFF(OutUnit,'Tree '//I2Str(i)//' = '//R2Str(hAll(i)))
   !END DO

   !CALL mexPrintFF(OutUnit,CHAR(10))
   !CALL mexPrintFF(OutUnit,'Actual Depth:                  '//LongI2Str(Depth)//CHAR(10))
   !CALL mexPrintFF(OutUnit,'Number of splits:              '//LongI2Str(nS)//CHAR(10))
   !CALL mexPrintFF(OutUnit,'Number of nodes:               '//LongI2Str(nN)//CHAR(10))
   !DO i=1,Depth
    !  CALL mexPrintFF(OutUnit,'   Level '//I2Str(i-1)//'                 '//LongI2Str(nNL(i-1))//CHAR(10))
   !END DO
   !CALL mexPrintFF(OutUnit,'Number of subTree searches:    '//LongI2Str(nST)//CHAR(10))
   !CALL mexPrintFF(OutUnit,'Number of redundant level2:    '//LongI2Str(nR)//CHAR(10))

   OutUnit=OutUnit
   hAll=hAll
   NTrees=NTrees
   Depth=Depth
   nS=nS
   nN= nN
   nNL=nNL
   nST=nST
   nR=nR

END SUBROUTINE OutputAlgorithmDetailsX




SUBROUTINE OutputProgressP (OutUnit, Predictor)
!USE MexLib
!USE IFPORT
!Arguments
INTEGER OutUnit, Predictor
! Begin routine

!CALL mexPrintFF (OutUnit, CHAR(10)//'Doing p: '//I2Str(Predictor)//', s:')
IF (OutUnit/=0) CALL FLUSH(OutUnit)
Predictor=Predictor

END SUBROUTINE OutputProgressP



SUBROUTINE OutputProgressS (OutUnit, Split)
!USE MexLib
!USE IFPORT
!Arguments
INTEGER OutUnit, Split
! Begin routine

!CALL mexPrintFF (OutUnit, I2Str(Split))
IF (OutUnit/=0) CALL FLUSH(OutUnit)
Split=Split

END SUBROUTINE OutputProgressS



SUBROUTINE OutputProgress (Predictor,Split)
!USE MexLib
! Arguments
INTEGER  Predictor,Split
!locals
!INTEGER plhs(0), prhs(2), Status
! Begin routine

!prhs(1) = mxCreateNumericMatrix(1, 1, mxClassIDFromClassName('int32'), 0)
!CALL mxCopyInteger4ToPtr(Predictor, mxGetPr(prhs(1)), 1)
!prhs(2) = mxCreateNumericMatrix(1, 1, mxClassIDFromClassName('int32'), 0)
!CALL mxCopyInteger4ToPtr(Split, mxGetPr(prhs(2)), 1)

!Status = mexCallMATLAB(0, plhs, 2, prhs, 'ProgressCallBack')


Predictor=Predictor
Split=Split


! CLEAN UP
!CALL mxDestroyArray(prhs(1))
!CALL mxDestroyArray(prhs(2))
END SUBROUTINE OutputProgress



LOGICAL FUNCTION IsInterrupted()
! Returns whether or not in MATLAB an process interruption flag has been set.
!USE MexLib
! Arguments
!locals
! Begin routine

!IF (-1 > 0) THEN
!   IsInterrupted = .TRUE.
!ELSE
  IsInterrupted = .FALSE.
!END IF

END FUNCTION IsInterrupted



END MODULE OptTreeCallBacks

AutoRecode<-function(V){
# recode in ordered categories
C<-sort(V)
I<-order(V)
n<-NROW(V)
O<-I
a<-1
O[I[1]]<-a
M<-c()
M[1]<-V[I[1]]
for(i in 2:n){
  if(V[I[i]]!=V[I[i-1]]){
    #if ((V(I(i))-V(I(i-1)))>0.49)
      a<-a+1
      M[a]<-V[I[i]]
  }
  O[I[i]]<-a
}

return(list(O=O, M=M))

}

SUBROUTINE mainFunction(Y,X,Measure,MaxSize,MaxDepth,MinNodeSize,MinH, &
                     AlgoType,BoundH,LookAheadDepth, Prior, LossM, MY, NY, MX, NX, &
                     TAll,hAll,h,Tv, TreeTableSize, NTrees, warn, NMaxNodes, XType)

USE OptTreeCallBacks, ONLY: dp
IMPLICIT NONE
! Arguments
INTEGER measure,MaxSize,MaxDepth,MinNodeSize, &
                     AlgoType,LookAheadDepth,warn, &
                     MY, NY, MX, NX, NTrees, TreeTableSize, NMaxNodes
INTEGER XType(NX)   ! 0=ordinal/continuous, 1=categorical
! Locals
INTEGER,PARAMETER:: LogFileNameLen=255
CHARACTER(LEN=LogFileNameLen) LogFileName
!INTEGER TAllPtr
REAL(dp) TAll(NMaxNodes,5,NTrees)
REAL(dp) MinH, BoundH, Prior(*), LossM(*), h, hAll(NTrees), Tv(NMaxNodes,5), X(MX,NX), Y(MY,NY)
!INTEGER i

LogFileName='default.log'


CALL OptTree (TRIM(LogFileName),MY, NY, Y, NX, X, XType, &
              Measure, MaxSize, MaxDepth, MinNodeSize, MinH, &
              AlgoType, BoundH, LookAheadDepth, Prior, LossM, &
              h, hAll, NTrees, TAll, TreeTableSize, NMaxNodes, warn)


Tv = TAll(1:NMaxNodes,1:5,NTrees) ! T = copy of the (largest) tree


END SUBROUTINE mainFunction


SUBROUTINE TransposeCopyMaxSizeTree(Size,TreeTable,Tree)
!If moved to OptimalTrees module, %VAL statement fails
USE OptTreeCallBacks, ONLY: dp
IMPLICIT NONE
! Arguments
INTEGER Size
REAL(dp) TreeTable(Size,5),Tree(5,Size)
! Begin routine

TreeTable=TRANSPOSE(Tree)

END SUBROUTINE TransposeCopyMaxSizeTree



SUBROUTINE OptTree (LogFileName,NRows,NColsY,Y,NColsX,X,XType,Measure,MaxSize,MaxDepth,MinNodeSize,MinH, &
                    AlgoType, BoundH, LookAheadDepth, Prior, LossM,  &
                    Heterogeneity, hAll, NTrees, TAll, TreeTableSize, NMaxNodes, warn)
USE OptTreeCallBacks
USE OptimalTrees
!USE IFPORT
IMPLICIT NONE
!! Arguments
CHARACTER(LEN=*) LogFileName
INTEGER NRows,NColsY,NColsX,NTrees,TreeTableSize, warn
REAL(dp) Y(NRows,NColsY), X(NRows,NColsX), NCat(NColsY)
INTEGER XType(NColsX)   ! 0=ordinal/continuous, 1=categorical
INTEGER Measure,MaxSize,MaxDepth,MinNodeSize,AlgoType,LookAheadDepth
REAL(dp) MinH, BoundH,Prior(*),LossM(*)
REAL(dp) Heterogeneity, h, hAll(NTrees)
!!Locals
INTEGER OutUnit,i, NMaxNodes
REAL(dp)  Tree(NTrees,TreeTableSize)
REAL(dp) TAll(NMaxNodes,5,NTrees)
!! Begin routine

NCat = MAXVAL(Y,DIM=1)

!Setup output (MATLAB display or file)
IF (LogFileName/='') THEN
   OutUnit = 2
   !OPEN(OutUnit,FILE=LogFileName,IOSTAT=Status,SHARED)
   !OPEN(OutUnit,FILE=LogFileName,ACCESS='APPEND',IOSTAT=Status)
   !IF (Status /= 0) THEN
   !  warn = 10 !CALL mexWarning('Unable to open output file '//LogFileName//CHAR(10))
   !   OutUnit = 0 !MATLAB display
   !END IF
ELSE
   OutUnit = 0 !MATLAB display
END IF



!! start output and show input parameters
!!CALL WriteHeader(OutUnit)
CALL EchoInput(NColsY,Measure,NCat)


IF (OutUnit/=0) CALL FLUSH(OutUnit)


! execute optimal tree algorithm
CALL ComputeMaxTree(OutUnit,NRows,NColsY,Y,NColsX,X,XType,Measure,MaxSize,MaxDepth,MinNodeSize,MinH, &
                    AlgoType, BoundH, LookAheadDepth, Prior,LossM, &
                    hAll, NTrees, Tree, TreeTableSize, warn)

!IF (IsInterrupted()) CALL WriteInterruptWarning(OutUnit)

! process returned data
Heterogeneity = hAll(1)
h = Heterogeneity

!      !Matlab expects TreeTable in transposed form
DO i=1,NTrees
   CALL TransposeCopyMaxSizeTree(TreeTableSize/5,TAll(1:NMaxNodes,1:5,i),Tree(i,1:TreeTableSize))
END DO

IF (OutUnit/=0) warn=45!CLOSE(OutUnit) !close external file

END SUBROUTINE OptTree

MaxEndNodes<-function(t,NTrees,defMaxSize,defMaxDepth){
if(NTrees==1){
  if(defMaxSize>0){
    Size<-min(defMaxSize,2^(defMaxDepth-1))
  }else{
    Size<-2^(defMaxDepth-1)
  }
}else{
  if(defMaxSize>0){
    Size<-t
  }else{
    Size<-2^(t-1)
  }
}

return(Size)
}

# calculateNominalSplit<-function(cat){
#
#   ncat<-length(cat)
#   split<-list()
#   cat1<-cat[1]
#   split[[1]]<-cat1
#   listIndex<-1
#   for (i in 2:(ncat-1)) {
#     index <- which(combn(cat,i)[1,]==cat1)
#     for (j in index) {
#       listIndex<-listIndex + 1
#       split[[listIndex]]<-combn(cat,i)[,j]
#     }
#
#
#   }
#
#   return(split)
#
#
# }
#
#
#
#
# calculateAllNominalSplit<-function(cat){
#
#   ncat<-length(cat)
#   split<-list()
#   cat1<-cat[1]
#   split[[1]]<-cat1
#   split[[2]]<-cat[-1]
#   listIndex<-2
#   for (i in 2:(ncat-1)) {
#     index <- which(combn(cat,i)[1,]==cat1)
#     for (j in index) {
#       listIndex<-listIndex + 1
#       split[[listIndex]]<-combn(cat,i)[,j]
#       listIndex<-listIndex + 1
#       split[[listIndex]]<-cat[! cat %in% combn(cat,i)[,j]]
#     }
#
#
#   }
#
#   return(split)
#
#
# }
#

GetHeterogeneity<-function(Y,aMeasure=NULL, measure, Prior, LossM, ClassSizes){
#global measure Prior LossM ClassSizes;

# measure<-globalenv()$measure
# Prior<-globalenv()$Prior
# LossM<-globalenv()$LossM
# ClassSizes<-globalenv()$ClassSizes

if(!is.null(aMeasure)){
  ActiveMeasure <- aMeasure
}else{
  ActiveMeasure <- measure
}

if(ActiveMeasure==0){
  h<-GetSSE(Y)
}else if(ActiveMeasure==1){
  #h=GetMisclassification(Y);
  h<-GetMisclassification2(Y) #faster in M6.5
}else if(ActiveMeasure==3){
  h<-GetExpectedRisk(Y, Prior, LossM, ClassSizes)
}else{
  stop('Undefined Heterogeneity Measure')
}

return(h)
}





GetMisclassification<-function(Y){
h<-0
Y<-matrix(Y, nrow = NROW(Y), ncol = NCOL(Y))
for(d in 1:(NCOL(Y))){
  hf<-0
  O<-sort(Y[,d])
  while(dim(O)[1]>0){
    c<-min(O)
    I<-which(O>c) #[I]=find(O>c);
    f<-dim(O)[1]-dim(I)[1]
    hf<-max(hf,f)
    O<-O[I]
  }
  h<-h+(dim(Y)[1]-hf)
}

return(h)
}




GetMisclassification2<-function(Y){
h<-0
yn<-NROW(Y)
ym<-NCOL(Y)
Y<-matrix(Y, nrow = yn, ncol = ym)
if(yn==1){
  return()
}

for(d in 1:ym){
  O<-sort(Y[,d])
  hf<-0
  s<-1
  for(i in 2:yn){
    if(O[i]!=O[i-1]){
      hf<-max(hf,(i-s))
      s<-i
    }
  }
  if(O[yn]==O[yn-1]){
    hf<-max(hf,(yn+1-s))
  }
  h<-h+(yn-hf)
}


return(h)
}



GetSSE<-function(Y){
#library(pracma)
#Give Sum of Squared error in prediction
E<-Y-repmat(a=mean(Y),n=NROW(Y),m=1)

h<-sum(E*E)

return(h)
}




GetExpectedRisk<-function(Y, Prior, LossM, ClassSizes){
#global Prior LossM ClassSizes;

# Prior<-globalenv()$Prior
# LossM<-globalenv()$LossM
# ClassSizes<-globalenv()$ClassSizes
Y<-matrix(Y, nrow = NROW(Y), ncol = NCOL(Y))
h<-0
for(d in 1:NCOL(Y)){
  dimCS<-c(NROW(ClassSizes[[d]]), NCOL(ClassSizes[[d]]))
  nCat <- max(dimCS)
  L<-LossM[[d]]
  f<-histc(x=Y[,d],edges=1:nCat)$cnt
  f<-as.vector(f) #f has different shape if Y has one row
  E<-as.vector(Prior[[d]])*f/ClassSizes[[d]]
  if(NCOL(E)==1){
    pa<-sum(E)
  }else{
    pa<-colSums(E)
  }
  piA<-E/pa
  ra<-min(t(L)%*%piA)

  h=h+pa%*%ra
}

return(h)
}

GetOP<-function(I,Splits){
# returns the ordered partition assignment for an ordered partitioning
# defined by Splits
dimensions<-c()
dimensions[1]<-NROW(Splits)
dimensions[2]<-NCOL(Splits)
k_1<-max(dimensions) #check if dim or length
s<-1
A<-c()
for(ki in 1:k_1){
  A[I[s:Splits[ki]]]<-ki
  s<-Splits[ki]+1
}

A[I[s:length(I)]]<-k_1+1

return(A)
}

CombineTrees<-function(H1,TT1,H2,TT2,maxSize,FirstSplit){
# %[H,TT]=CombineTrees2(H1,TT1,H2,TT2,maxSize);
# %return
# %and enforce split
H<-rep(2.3E10,maxSize)
Tl<-list() #=cell(3,1);
TT<-list() #cell(maxSize,1);
availSize<- dim(H1)[2]+1

for(ms in 2:maxSize){
  msm<-ms-min(ms,availSize)
  # H1(1+msm:ms-(1+msm));
  # H2(ms-(1+msm):-1:1+msm);
  HN<-H1[(1+msm):(ms-(1+msm))]+H2[seq(from=ms-(1+msm),to=1+msm,by=-1)]#there might be a dimensional problem if availSize<ms
  #%HN=H1(1:ms-1)+H2(ms-1:-1:1);
  mH<-min(HN)
  mi<-which.min(HN)

  if(mH<H[ms]){
    mi<-msm+mi
    H[ms:maxSize]<-mH
    Tl[[1]]<-FirstSplit
    Tl[[2]]<-TT1[mi]
    Tl[[3]]<-TT2[ms-mi]
    TT[[ms]]<-Tl
  }
}
H[1]<-0

return(H=H,TT=TT)

}

SelectCluster<-function(X,pc){

nr<-NROW(X)
ind<-matrix(rep(1,nr),nrow = nr, ncol = 1)
expand<-matrix(rep(1,nr),nrow = nr, ncol = 1)
for(jj in 1:NCOL(X)){
  ind <- (ind & (X[,jj] >= pc[expand,jj])) &  (X[,jj] <= pc[2*expand,jj])
}
ind<-which(ind)

return(ind)

}

GetPrediction<-function(Y, measure, Prior, LossM, ClassSizes){
#global measure;

#measure<-globalenv()$measure

if(measure==0){
  y<-GetPSSE(Y)
}else if(measure==1){
  y<-GetPMisclassification(Y)
}else if(measure==3){
y<-GetPExpectedRisk(Y, Prior, LossM, ClassSizes)
}else{
  stop('Undefined Heterogeneity Measure')
}

return(y)

}

OptTreeF <- function(Y,X,XType,measure,MaxSize,MaxDepth,MinNodeSize,MinH,
                     AlgoType,BoundH, Prior, LossM,LookAheadDepth){

  warn<-0
  MY<-NROW(Y)
  NY<-NCOL(Y)
  MX<-NROW(X)
  NX<-NCOL(X)


  if(MaxSize == 0){
    NMaxNodes = (2^MaxDepth) -1
    if(bitwAnd(AlgoType,4)>0){ #THEN #Trees for all sizes<=restrictions
      NTrees <- MaxDepth
    }else{
      NTrees <- 1
    }
  }else{
    MaxSize<-min(MaxSize,2^(MaxDepth-1)) #! Prevent redundant heap
    NMaxNodes <- 1+2*(MaxSize-1)
    if(bitwAnd(AlgoType,4)>0){ #THEN #!Trees for all sizes<=restrictions
      NTrees <- MaxSize
    }else{
      NTrees <- 1
    }
  }
  TreeTableSize <- NMaxNodes*5


  hAll<-numeric(NTrees)
  h<-0
  Tv<-matrix(0,nrow = NMaxNodes, ncol=5)
  TAll<-array(data=0,dim=c(NMaxNodes,5,NTrees))#mxCreateFull(NMaxNodes, 5, 0) iteration over Ntrees

  Tree<-.Fortran("mainFunction",Y=as.matrix(Y),X=as.matrix(X),
                 Measure=as.integer(measure),
                 MaxSize=as.integer(MaxSize), MaxDepth=as.integer(MaxDepth),
                 MinNodeSize=as.integer(MinNodeSize), MinH=as.double(MinH),
                 AlgoType=as.integer(AlgoType), BoundH=as.double(BoundH),
                 LookAheadDepth=as.integer(LookAheadDepth),Prior=as.double(Prior),
                 LossM=as.double(LossM), MY=as.integer(MY), NY=as.integer(NY),
                 MX=as.integer(MX), NX= as.integer(NX),
                 #output variables TAll, hAll, h, Tv
                 TAll=as.array(TAll), hAll=as.double(hAll),
                 h=as.double(h),Tv=as.matrix(Tv),
                 TreeTableSize=as.integer(TreeTableSize), NTrees=as.integer(NTrees),
                 warn=as.integer(warn), NMaxNodes=as.integer(NMaxNodes),
                 XType=as.integer(XType),
                 PACKAGE="ExactTree")


  return(Tree)
}

# BootsTrapIndex<-function(BootSize,SampleSize,WithDraw){
#
# if(nargs()<3){
#   #met teruglegging
#   I<-1+trunc(runif(BootSize)*SampleSize)
# }else{
#   #zonder teruglegging
#   #library(pracma)
#   P<-randperm(SampleSize)
#   I<-P[1:BootSize]
# }
#
# #I<-matrix(I, ncol=1)
#
# return(I)
# }

GetHeterogeneityTest<-function(Yt,Prediction,measure){
#global measure

#globalenv()$measure

if(measure==0){
  h<-GetSSETest(Yt,Prediction)
}else if(measure==1){
  h<-GetMisclassificationTest(Yt,Prediction)
}else{
  stop('Undefined Heterogeneity Measure')
}

return(h)
}





GetMisclassificationTest<-function(Yt,Prediction){
h<-0
for(d in 1:(dim(Yt)[2])){
I<-which(Yt[,d]!=Prediction[d])
h<-h+dim(I)[1]
}

return(h)
}


GetSSETest<-function(Yt,Prediction){
# %Give Sum of Squared error in prediction
# %E=Y-repmat(mean(Y),size(Y,1),1);
#library(pracma)
E<-Yt-repmat(Prediction,NROW(Yt),1)
h<-sum(E*E)
}

GetPExpectedRisk<-function(Y, Prior, LossM, ClassSizes){
#global Prior LossM ClassSizes;
#library(pracma)


# Prior<-globalenv()$Prior
# LossM<-globalenv()$LossM
# ClassSizes<-globalenv()$ClassSizes

nc<-NCOL(Y)
y<-matrix(0,nrow = nc, ncol = 1)

for(d in 1:nc){
  nCat <- max(NROW(ClassSizes[[d]]),NCOL(ClassSizes[[d]]))
  L<-LossM[[d]]
  f<-histc(Y,1:nCat)$cnt
  E<-(as.vector(Prior[[d]])*f)/ClassSizes[[d]]
  if(NCOL(E)==1){
    pa<-sum(E)
  }else{
    pa<-colSums(E)
  }

  piA<-E/pa
  LpiA<-t(L)%*%piA #matrix product
  ra <- min(LpiA)
  y[d]<-which.min(LpiA)
}

return(y)

}

GetLearningStuff<-function(Tree,Yl,Xl,Yt,Xt, measure, Prior, LossM, ClassSizes){
#library(pracma)

Yl<-as.matrix(Yl)
Yt<-as.matrix(Yt)

m<-dim(Xl)[2]
Ypred<-matrix(rep(0,NROW(Yt)),nrow = NROW(Yt), ncol=1)#dim(Yt)[1]
I<-which(Tree[,1]==0)#=find(Tree(:,1)==0);
Tv<-Tree[I,] #Terminal Nodes
Tv<-matrix(Tv, ncol = NCOL(Tree))
nT<-NROW(Tv)
h<-0
for(t in 1:nT){ #all terminal nodes
  pc<-Reshape(Tv[t,8:(NCOL(Tv))],n=2,m=m) #Use Reshape
  Cl<-SelectCluster(Xl,pc)
  Ct<-SelectCluster(Xt,pc)
  CYl<-Yl[Cl,]
  CYt<-Yt[Ct,]
  s<-max(NROW(Ct))
  #The CV problem is here: With the s and the Ypred


  if (s==0){ #in the case of doing it as in matlab it should be s!=0
    next
  }

  Ypred[Ct,]<-repmat(GetPrediction(CYl, measure, Prior, LossM, ClassSizes),n=s,m=1)
  y<-GetPrediction(CYl,measure, Prior, LossM, ClassSizes)
  h<-h+GetHeterogeneityTest(CYt,y, measure)

}


return(list(Ypred=Ypred, h=h))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ETree.R
\name{ETree}
\alias{ETree}
\title{Main function of the package. It performs the Exact Tree method.}
\usage{
ETree(
  formula = NULL,
  data = NULL,
  map = NULL,
  original = NULL,
  round = NULL,
  discretize = NULL,
  selV = NULL,
  control = NULL,
  verbose = TRUE
)
}
\arguments{
\item{formula}{a description of the model to be fit. The format is \code{Y ~  X1 + \dots + Xn},
where the variable before ~ represents the dependent variable and the variables after
the ~ are the independent variables.}

\item{data}{Dataset to be analyzed. Note: If data contains ordinal variables, transform them to numerical
before using this function. Otherwise, they will be treated as nominal variables.}

\item{map}{[ori vars] sets for each variable whether the tree will depict
thresholds based upon original values (var number), or recoded values (0).}

\item{original}{[TRUE/FALSE] sets for each variable whether the tree will depict
thresholds based upon original values (true), or recoded values (false).}

\item{round}{[factors] autorecoding based upon rounded categories; the categories will
be round after multiplying with the factor, and devided by the
factor after rounding}

\item{discretize}{[ncat] optimal discretization while minimizing SS within
a group, indicate with a number how many categories;set ncat=0 for each variable that is to
be left unaltered}

\item{selV}{list containing the output obtained using \code{\link{SelectVar}}.
Use this list instead of the previous inputs if you want to perform SelectVars outside ETree.}

\item{control}{a list with control parameters as returned by \code{\link{ETree.control}}.}

\item{verbose}{logical, if TRUE, prints information about the progress of the algorithm.}
}
\value{
Returns the following 6 elements:
 \item{h}{contains the objective function value (fit) of the best tree.}
 \item{Tree}{contains the largest Tree table.}
 \item{hAll}{(OPTIONAL) contains the objective function values for all trees, if more than one is requested}
 \item{TAll}{(OPTIONAL) contains Tree tables for all trees, if more than one is requested.}
 \item{Transf_Trees}{contains an object of class ExactTree that can be used in \code{\link{plot.ETree}} and \code{\link{summary.ETree}}}
 \item{CVOutput}{Cross validation results for all the requested trees.}
}
\description{
Main function of the package. It performs the Exact Tree method.
}
\details{
The function results in a global tree (for given maxsize and maxdepth) and transforms the output to obtain summary information and plot the tree.
}
\examples{
\donttest{
  data(iris)
  # Fit an Exact Tree model
  controlEtree <- ETree.control(measure=0, maxsize = 4, maxdepth = 3,
  minbucket = 5, ncv=5, alltreesizes = FALSE)
  ETree(Species ~ Sepal.Length + Sepal.Width + Petal.Length + Petal.Width,
  control= controlEtree, data = iris)
}


}
\seealso{
\code{\link{summary.ETree}}, \code{\link{SelectVar}},
  \code{\link{plot.ETree}},\code{\link{ETree.control}},
  \code{\link{predict.ETree}},\code{\link{prune.ETree}}
}
\keyword{exact}
\keyword{tree}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_support.R
\name{as.party.ETree}
\alias{as.party.ETree}
\title{Transformation of a Exact Tree object to party object}
\usage{
\method{as.party}{ETree}(obj, nodeID = 1L, ...)
}
\arguments{
\item{obj}{tree of class \code{ETree}.}

\item{nodeID}{Node identification.}

\item{\dots}{additional arguments to be passed.}
}
\value{
object transformed to a \code{constparty} object.
}
\description{
Transformation of a Exact Tree object to party object
}
\keyword{as.party}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.ETree.R
\name{predict.ETree}
\alias{predict.ETree}
\title{Predictions for new data with a ETree object}
\usage{
\method{predict}{ETree}(object, newdata = NULL, type = "pred", depth = NULL, ...)
}
\arguments{
\item{object}{an object of the class \dQuote{ETree}.}

\item{newdata}{a data frame with data on new subjects for whom predictions should be made.
The data frame should contain at least the variables used in the splits of the fitted tree.
It is not necessary to include the treatment variable.}

\item{type}{character string denoting the type of predicted object to be returned. The default is
set to \code{type="pred"}: a vector with predicted treatment subgroup classes per subject
is returned. If set to \code{"matrix"}, a matrix is returned with the leaf and
corresponding node of the tree to which a subject is assigned.}

\item{depth}{If alltreesizes was set to TRUE in ETree, you need to specify the depth of the tree
you want to use in the predict function. This parameter should be equal to the number of the tree
you want to use in the TAll output of ETree. If NULL, the largest tree is used.}

\item{\dots}{optional additional arguments.}
}
\value{
One of the following objects is returned depending on output type specified
  in the function:

  If \code{type="pred"}:
  vector of the predicted outcome for every individual in the data set. Returns NA
  for subjects with missing values on one or more of the splitting variables.

  If \code{type="matrix"}:
  a matrix with predicted locations of subjects within the fitted tree. The leaf numbers are
  in the first column and the corresponding node numbers in the second column. Returns NA
  for subjects with missing values on one or more of the splitting variables.

  If \code{type="prob"}:
  a matrix with probabilities.
}
\description{
Predicts for (new) subjects the outcome variable based on a fitted
  \code{ETree} object.
}
\examples{
\donttest{
  data(iris)
  trainingIris<-iris[1:100,]
  # Fit an Exact Tree model
  controlEtree <- ETree.control(measure=0, maxsize = 4, maxdepth = 3,
  minbucket = 5, ncv=5, alltreesizes = FALSE)
  tree<-ETree(Species ~ Sepal.Length + Sepal.Width + Petal.Length + Petal.Width,
  control= controlEtree, data = trainingIris)
  testIris<-iris[101:150,]
  predictions<-predict(tree, newdata=testIris, type="pred", depth=1)
}


}
\seealso{
\code{\link{ETree}}, \code{\link{ETree.control}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.ETree.R
\name{summary.ETree}
\alias{summary.ETree}
\title{Summarizing Exact Trees Information}
\usage{
\method{summary}{ETree}(object, TerminalNodes = NULL, digits = 3, ...)
}
\arguments{
\item{object}{a \code{ETree} object. This can be the output of \code{\link{ETree}}.}

\item{TerminalNodes}{Number of terminal nodes of the tree to summarize.}

\item{digits}{specified number of decimal places (default is 3).}

\item{\dots}{optional additional arguments.}
}
\value{
prints a summarized version of the \code{ETree} output.
}
\description{
Summary method for an object of class \code{ETree}.
}
\details{
This function is a method for the generic function summary for class
  \code{ETree}. It extracts the following essential components from a \code{ETree}
  object: 1) Split information;
  2) Leaf information, and 3) CV information.
}
\examples{
\donttest{
  data(iris)
  # Fit an Exact Tree model
  controlEtree <- ETree.control(measure=0, maxsize = 4, maxdepth = 3,
  minbucket = 5, ncv=5, alltreesizes = FALSE)
  tree<-ETree(Species ~ Sepal.Length + Sepal.Width + Petal.Length + Petal.Width,
  control= controlEtree, data = iris)
  summary(tree, TerminalNodes=4)
}

}
\keyword{summary}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.ETree.R
\name{plot.ETree}
\alias{plot.ETree}
\title{Visualisation of a Exact Tree}
\usage{
\method{plot}{ETree}(x, TerminalNodes = NULL, digits = 2, ...)
}
\arguments{
\item{x}{transformed tree of class \code{ExactTree}.}

\item{TerminalNodes}{Number of terminal nodes of the tree to plot.}

\item{digits}{specified number of decimal places of the splitpoints in the graph
(default is 2).}

\item{\dots}{additional arguments to be passed.}
}
\value{
A plot of the tree with the specified number of terminal nodes.
}
\description{
Visualisation of a Exact Tree
}
\examples{
\donttest{
  data(iris)
  # Fit an Exact Tree model
  controlEtree <- ETree.control(measure=0, maxsize = 4, maxdepth = 3,
  minbucket = 5, ncv=5, alltreesizes = FALSE)
  tree<-ETree(Species ~ Sepal.Length + Sepal.Width + Petal.Length + Petal.Width,
  control= controlEtree, data = iris)
  plot(tree, TerminalNodes=4)
}

}
\references{
Torsten Hothorn and Achim Zeileis (2013). partykit: A Toolkit for
  Recursive Partytioning. R package version 0.1-5.
}
\seealso{
\code{\link{ETree}},\code{\link{summary.ETree}}
}
\keyword{as.party}
\keyword{plot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SelectVar.R
\name{SelectVar}
\alias{SelectVar}
\title{This function preprocess the data for the Optimal Trees function.}
\usage{
SelectVar(
  formula = NULL,
  data,
  Names = NULL,
  YSelected = NULL,
  XSelected = NULL,
  map = NULL,
  original = NULL,
  round = NULL,
  discretize = NULL,
  verbose = TRUE
)
}
\arguments{
\item{formula}{a description of the model to be fit. The format is \code{Y1 + \dots + Yn ~  X1 + \dots + Xn},
where the variable before ~ represents the dependent variables and the variables after
the ~ are the independent variables.}

\item{data}{Dataset to be analyzed. If data contains ordinal variables, order the factor levels
before using this function. Otherwise, the factors will be ordered alphabetically and the results will not be correct.}

\item{Names}{names of the variables in the data. If empty, it uses the column names of Data}

\item{YSelected}{index corresponding to the first Selected column.}

\item{XSelected}{Indices corresponding to the selected columns that will form X}

\item{map}{[ori vars] sets for each variable whether the tree will depict
thresholds based upon original values (var number), or recoded values (0).}

\item{original}{[TRUE/FALSE] sets for each variable whether the tree will depict
thresholds based upon original values (true), or recoded values (false).}

\item{round}{[factors] autorecoding based upon rounded categories; the categories will
be round after multiplying with the factor, and divided by the
factor after rounding. Set round=0 for variables that are not
rounded (e.g. discretized variables).}

\item{discretize}{[ncat] optimal discretization while minimizing SS within
a group, set ncat=0 for each variable that is to
be left unaltered}

\item{verbose}{logical, if TRUE, prints information about the progress of the algorithm.}
}
\value{
Returns the following 3 elements:
 \item{Y}{dataset containing the Y variables.}
 \item{X}{dataset containing the X variables.}
 \item{Desc}{List containing the variable names and their corresponding cutting points.}
}
\description{
This function preprocess the data for the Optimal Trees function.
}
\details{
The function selects the Y and X variables according to the inputs and also returns
Desc, a list containing variable cutting points and variable names. These 3 elements are
required by the OptimalTrees function.
}
\examples{
data(mtcars)
dataSelection<-SelectVar( mpg ~ cyl + hp + wt, data = mtcars, discretize= c(0, 10, 10))

}
\seealso{
\code{\link{summary.ETree}}, \code{\link{ETree}},
  \code{\link{plot.ETree}}, \code{\link{predict.ETree}}
}
\keyword{exact}
\keyword{tree}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prune.ETree.R
\name{prune.ETree}
\alias{prune.ETree}
\title{Pruning of a Exact Tree}
\usage{
\method{prune}{ETree}(tree, pp = 1, ...)
}
\arguments{
\item{tree}{fitted tree of the class \code{ETree}.}

\item{pp}{pruning parameter, the constant (\eqn{c}) to be used in the \eqn{c*}standard
error rule. The default value is 1.}

\item{\dots}{optional additional arguments.}
}
\value{
Returns an object of class \code{ETree}. The number of leaves of this object is
  equal to the optimally pruned size of the tree.
}
\description{
Determines the optimally pruned size of the tree by applying the one
standard error rule to the results from the bias-corrected bootstrap procedure.
}
\details{
The one standard error rule for \code{ETrees} uses the estimates of the bias-corrected
  criterion value (\eqn{C}) and its standard error for each value of \eqn{L}
  (= maximum number of leaves). The optimally pruned tree corresponds to the
  smallest tree with a bias-corrected \eqn{C} higher or equal to the maximum
  bias-corrected \eqn{C} minus its standard error.
}
\keyword{tree}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ETree.control.R
\name{ETree.control}
\alias{ETree.control}
\title{Control Parameters for ETree Algorithm}
\usage{
ETree.control(
  measure = NULL,
  maxsize = 4,
  maxdepth = 2,
  minbucket = 10,
  minheterogeneity = 0.05,
  ncv = 10,
  cvVector = NULL,
  heterogeneityonly = NULL,
  BoundH = NULL,
  branchandbound = NULL,
  alltreesizes = FALSE,
  sortedpsplits = NULL,
  lookaheadheuristic = NULL
)
}
\arguments{
\item{measure}{0 = minimize residual Sum of Squares (continuous Y)
1 = minimize Misclassification Rate        (discrete Y)
3 = minimize the Risk Sum P(A)*R(A)        (discrete Y)
    Measure = scalar or Measure=c(3,c(Prior,LossM))}

\item{maxsize}{if null: only maxdepth restrictions are applied
if defined: the maximum number of terminal nodes for the best tree}

\item{maxdepth}{The maximum number of layers in a tree, NOT INCLUDING THE LAYER 0.
maxdepth=0 means a tree with just one node (the root node),
maxdepth=1 means one split and two terminal nodes, etc.}

\item{minbucket}{defines the minimum number of observations in a terminal node
if the number of observations in a node is equal to this minimum
the node becomes a possible terminal node; if the number of observations in a node
is smaller, the node is set illegal and disregarded.}

\item{minheterogeneity}{defines the situation when a node A is considered Very Homogeneous
and is not allowed to be split further, i.e. the node A becomes a possible terminal node
(|A| = number of objects in node)}

\item{ncv}{specify N for N-fold cross-validation
if NCV = 0, no N-fold cross-validation
if NCV<0, abs(N)-fold cross-validation ONLY (no tree estimation)
if NULL (default), NCV=0}

\item{cvVector}{<vector> process N-fold cross-validation according to the classes defined by <vector>
vector is assumed to assign each observational unit to one of
N-classes, number 1,...,N}

\item{heterogeneityonly}{(TRUE,FALSE)
Algorithm will only give Heterogeneity/Impurity not a tree table
=> setting to false will give treetable as well}

\item{BoundH}{optional bound for best tree value, for BranchandBound}

\item{branchandbound}{(TRUE,FALSE)
Algorithm will use branch and bound rules to speed up
=> for maxsize tree only, h only
=> make sure you specify boundH}

\item{alltreesizes}{(FALSE,TRUE)
algorithm will compute multiple trees of size<=size restriction}

\item{sortedpsplits}{(TRUE,FALSE)
Algorithm will try to speed up use a sorting logic.
=> suitable for continuous predictors with many categories}

\item{lookaheadheuristic}{Specify treedepth=the number of levels the heuristic will lookahead.
When growing a tree with lookahead search (heurist approach),
this specifies the maximum depth with respect to which
any local split is optimized. When setting this depth = 1, this
ammounts to convential tree growing.}
}
\value{
A list containing the options for the function ETree.
}
\description{
Various parameters that control aspects of the \dQuote{ETree} algorithm.
}
\examples{
ETree.control(measure=0, maxsize = 6, maxdepth = 4, minbucket = 5, ncv=10, alltreesizes = TRUE)

}
\seealso{
\code{\link{ETree}}
}
